#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "lyapunov.h"

// --------------------------------------------------------------------------
//                      FIND_POLY
// --------------------------------------------------------------------------

struct poly * find_poly( struct hom_poly * f, int i )
{
	struct poly *g = NULL;

	while ( ( f = f->next_hom_poly ) != NULL )
		if ( i > f->total_degree )
			break;
		else if ( f->total_degree == i )
		{
			g = f->p;
			break;
		}
	return ( g );
}

// --------------------------------------------------------------------------
//                      DIFF
// --------------------------------------------------------------------------

void diff( struct poly * f )
{
	struct poly *w = f->next_poly;

	while ( w != NULL )
	{
		if ( w->degz )
		{
			w->re *= w->degz;
			w->im *= w->degz;
			w->degz--;
			f = w;
			w = w->next_poly;
		}
		else
		{
			w = w->next_poly;
			delete_coeff( f );
		}
	}
}

// --------------------------------------------------------------------------
//                      G
// --------------------------------------------------------------------------

void G( struct poly * f )
{
	int deg;
	double a;
	struct poly *w = f->next_poly;

	if ( w )
		deg = w->degz + w->degzb;
	while ( w != NULL )
	{
		if ( deg - 2 * w->degzb )
		{
			a = 2.0 / ( deg - 2 * w->degzb );
			w->re *= a;
			w->im *= a;
			f = w;
			w = w->next_poly;
		}
		else
		{
			w = w->next_poly;
			delete_coeff( f );
		}
	}
}

// --------------------------------------------------------------------------
//                      IMGZ
// --------------------------------------------------------------------------

void Imgz( struct poly * f, struct poly * g )
{
	struct poly * h;

	prod_poly( f, g );
	diff( f );
	G( f );
	h = conj_poly( f );
	sub_poly( f, h );
	delete_poly( h );
	multc_poly( f, 0.0, -0.5 );
}

// --------------------------------------------------------------------------
//                      REGZ
// --------------------------------------------------------------------------

void Regz( struct poly * f, struct poly * g )
{
	struct poly * h;

	prod_poly( f, g );
	multc_poly( f, 0.0, -1.0 );
	diff( f );
	G( f );
	h = conj_poly( f );
	add_poly( f, h );
	delete_poly( h );
	multc_poly( f, 0.5, 0.0 );
}

// --------------------------------------------------------------------------
//                      LL
// --------------------------------------------------------------------------

void LL( struct poly * f, struct poly * g, int i, double * v )
{
	v[ 0 ] = 0.0;
	v[ 1 ] = 0.0;
	prod_poly( f, g );
	diff( f );
	while ( ( f = f->next_poly ) != NULL )
	{
		if ( i > f->degz )
			break;
		if ( i == f->degz && i > f->degzb )
			break;
		if ( i == f->degz && i == f->degzb )
		{
			v[ 0 ] = f->re;
			v[ 1 ] = f->im;
			break;
		}
	}
}

// --------------------------------------------------------------------------
//                      PART_LYAPUNOV_COEFF
// --------------------------------------------------------------------------

double part_lyapunov_coeff( char * s, int k )
{
	struct poly * R[ DIM1 ], *f = NULL;
	int i, t = 0, ok = 1, j = 0;
	char a[ DIM1 ];
	double v[ 2 ], w;

	for ( i = 0;s[ i ];i++ )
	{
		if ( s[ i ] == ',' )
		{
			R[ t ] = find_poly( vec_field, atoi( a ) + 1 );  /* find the homogeneus part
			                                                    of degree a+1 */
			if ( R[ t ] == NULL )
			{
				ok = 0;
				break;
			}
			t++;
			j = 0;
		}
		else
		{
			a[ j ] = s[ i ];
			j++;
		}
	}
	if ( ok )
	{
		R[ t ] = find_poly( vec_field, atoi( a ) + 1 );
		if ( R[ t ] == NULL )
			ok = 0;
	}
	if ( ok )
	{
		f = ( struct poly * ) malloc( sizeof( struct poly ) );
		f->next_poly = NULL;
		ins_poly( f, 0, 0, -1.0, 0.0 ); /* f=-1 */
		for ( i = 0;i < t;i++ )
			if ( i % 2 )
				Regz( f, R[ i ] );
			else
				Imgz( f, R[ i ] );
		if ( t % 2 )
		{
			multc_poly( f, 0.0, -1.0 );
			LL( f, R[ t ], ( k - 1 ) / 2, v );
			w = v[ 1 ];
		}
		else
		{
			LL( f, R[ t ], ( k - 1 ) / 2, v );
			w = v[ 0 ];
		}
		delete_poly( f );
	}
	else
		w = 0.0;
	return w;
}
