procedure assoc_points(vec_field);
begin scalar points,f,g,r;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                   %
%  points is the set of points associated to the given vector field %
%  in the (x,y) plane                                               % 
%  !! load assist                                                   %
%                                                                   %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	points:={};
        r:=symbolic !*rounded;
        on rounded;
	f:=first vec_field;
        while f neq 0 do
         << g:=nlterm(f,x,y);
            points:=union(points,{{deg(g,x)-1,deg(g,y)}});
            f:=f-g; >>;
        f:=second vec_field;
	   while f neq 0 do
         << g:=nlterm(f,x,y);
            points:=union(points,{{deg(g,x),deg(g,y)-1}});
            f:=f-g; >>;
	if not r then off rounded;
	return points
end;

procedure newton_diagram(vec_field);
begin scalar points,hset,a,l,j,k,pl,pk,pj;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                            %
%  hset is the set of points which define the newton diagram %
%  of the given vector field in the (x,y) plane              %
%  !! load assist                                            %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

hset:=assoc_points(vec_field);
points:=hset;
a:=length points;
for l:=1:a-1 do
 << pl:=part(points,l);
    if member(pl,hset) then
     for j:=l+1:a do
       << pj:=part(points,j);
          if member(pj,hset) then
           if (first pl<=first pj) and (second pl<= second pj)
             then hset:=diffset(hset,{pj})
             else
              if (first pj<=first pl) and (second pj<=second pl)
               then hset:=diffset(hset,{pl})
               else 
                for k:=1:a do 
                 << pk:=part(points,k);
                    if member(pk,hset) and (k neq l) and (k neq j) then
                     if ((first pl<=first pk) and (second pl<=second pk))
                                              or
                        ((first pj<=first pk) and (second pj<=second pk))
                      then hset:=diffset(hset,{pk})
                      else
                       if ((first pl<first pk) and (first pk<first pj))
                                               or
                          ((first pj<first pk) and (first pk<first pl))
                       then  
                        if ((second pj-second pl)/(first pj-first pl)*
                           (first pk-first pj)+second pj) <= second pk
                         then  hset:=diffset(hset,{pk});
                   >>; >>; >>;

return hset;
end;

procedure gamma1(s);
begin scalar k,l,pt1,pt2,pt3;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                         %
% GAMMA_1 is the first line in the NEWTON DIAGRAM         %
%                                                         %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

l:=length(s);
if first first s<first second s 
 then << pt1:=first s; pt2:=second s>>
 else << pt1:=second s; pt2:=first s>>;
if l=2 then return {pt1,pt2}
       else  
  if first third s<first pt1 then << pt3:=pt2; pt2:=pt1; pt1:=third s >>
                               else
   if first third s<first pt2 then << pt3:=pt2; pt2:=third s >> 
                                else pt3:=third s;
for k:=4:l do
 if first part(s,k)<first pt1 then << pt3:=pt2; pt2:=pt1; pt1:=part(s,k) >>
                              else
  if first part(s,k)<first pt2 then << pt3:=pt2; pt2:=part(s,k) >> 
                               else
   if first part(s,k)<first pt3 then pt3:=part(s,k);
if first pt2>0 then return {pt1,pt2}
               else return {pt2,pt3};
end;

procedure quasi_degree(s);
begin scalar alpha,beta,h;
 h:=(second second s-second first s)/(first first s-first second s);
 alpha:=num h;
 beta:=den h ;
 return {alpha,beta,alpha*first first s+beta*second first s};
end;


procedure blow_upx(vec_field,q_degree,direction,x,y);
begin scalar alpha,beta,d,vx,vy,p_g,q_g,f,g,r,s,delta;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                           %
% Blowing up in the x-direction                                             %
% if direction=1 then positive x-direction, if direction=-1 then negative   %
% direction                                                                 %
%                                                                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

on ratarg;
alpha:=first q_degree;
beta:=second q_degree;
d:=part(q_degree,3);
vx:=0; vy:=0;
p_g:=0; q_g:=0;
t_g:=0;
f:=first vec_field;
while f neq 0 do
  << g:=nlterm(f,x,y);
     r:=deg(g,x)-1;
     s:=deg(g,y);
     delta:=alpha*r+beta*s;
     vx:=vx+direction*sub(x=direction,g)*x^(delta+1-d);
     vy:=vy-direction*x^(delta-d)*beta*y*sub(x=direction,g);
     if delta=d then << p_g:=p_g+direction*sub(x=direction,g);
                        t_g:=t_g-direction*beta*y*sub(x=direction,g) >>;
     f:=f-g;
   >>;
f:=second vec_field;
while f neq 0 do
  << g:=nlterm(f,x,y);
     r:=deg(g,x);
     s:=deg(g,y)-1;
     delta:=alpha*r+beta*s;
     vy:=vy+alpha*x^(delta-d)*sub(x=direction,g);
     if delta=d then << q_g:=q_g+sub(x=direction,g);
                        t_g:=t_g+alpha*sub(x=direction,g) >>;
     f:=f-g;
   >>;
off ratarg;
return {{vx,vy},{p_g,q_g},t_g};
end;


procedure center_type2(f,x,l,direction);
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% return type of center on x=0        %
% f(0)=0   y-direction                %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
begin scalar s,g,m,b;
g:=first_term(f,x);
on ratarg; 
m:=deg(g,x); 
b:=sign(lcof(g,x));
off ratarg;
if evenp(m) then s:={b,1,sign(evalf(l))}
            else s:={b,-1,sign(evalf(l))};
if direction=1 then 
 <<if s={-1,1,1} then return 9;
   if s={1,1,1} then return 10; 
   if s={1,-1,1} then return 3;
   if s={-1,-1,1} then return 1; 
   if s={-1,1,-1} then return 11;
   if s={1,1,-1} then return 12;
   if s={1,-1,-1} then return 2 else return 4;>>
               else
 <<if s={-1,1,1} then return 10;
   if s={1,1,1} then return 9;
   if s={1,-1,1} then return 3;
   if s={-1,-1,1} then return 1; 
   if s={-1,1,-1} then return 12;
   if s={1,1,-1} then return 11;
   if s={1,-1,-1} then return 2 else return 4>>;
end;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% blowing up in the y direction and %%%
%%% determing the type of singularity %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure blow_upy(vec_field,q_degree,direction,transformations);
begin scalar z,vec_field2,lambda1,lambda2,vec,g,man;

vec_field2:=transformation(vec_field,mat((0,1),(1,0)));
z:=blow_upx(vec_field2,{second q_degree,first q_degree,third q_degree},
    direction,x,y);
vec_field2:=transformation(first z,mat((0,1),(1,0)));
write "blowing up vector field ",vec_field2;
lambda1:=sub(y=0,first second z);
lambda2:=sub(y=0,df(third z,y));
write "{0,0} is a singular point with eigenvalues ",lambda2,",",lambda1;
if save_all then
<< openfile(result_file);
   write "blowing up vector field ",vec_field2;
   write "{0,0} is a singular point with eigenvalues ",lambda2,",",lambda1;
   out 't;
>>;
% which type of singularity ? %
if eeq(lambda1,0) or eeq(lambda2,0) or gteq(0,lambda1*lambda2) then
 << vec:=sub(x=0,y=0,df(second first z,x))/(lambda1-lambda2);
    g:=transformation(first z,mat((1,0),(vec,1)));
    write "Transform this vector field so that the linear part is diagonal";
    write "The new vector field is ",transformation(g,mat((0,1),(1,0)));
    if save_all then
    <<openfile(result_file);
      write "Transform this vector field so that the linear part is diagonal";
      write "The new vector field is ",transformation(g,mat((0,1),(1,0)));
      out 't;
    >>;
    if gt(lambda1,0) then 
      <<sep:=manifold_blow_up(g,taylor_level);
        write "unstable separatrice x=",sub(t=y,sep);
        if save_all then
        << openfile(result_file);
           write "unstable separatrice x=",sub(t=y,sep);
           out 't;
        >>;
        if nneq(lambda2,0) then
        man:={transformations,{0,0},{vec,1,1,0},g,sep,1,symbolic !*rounded}
                           else
        man:={transformations,{0,0},{vec,1,1,0},g,sep,center_type2(
           sub(y=0,first vec_field2),x,lambda1,direction),symbolic !*rounded};
      >>
    else if gt(0,lambda1) then
      <<sep:=manifold_blow_up(g,taylor_level);
        write "stable separatrice x=",sub(t=y,sep);
        if save_all then
        << openfile(result_file);
           write "stable separatrice x=",sub(t=y,sep);
           out 't;
        >>;   
        if nneq(lambda2,0) then
         man:={transformations,{0,0},{vec,1,1,0},g,sep,2,symbolic !*rounded}
                           else
         man:={transformations,{0,0},{vec,1,1,0},g,sep,center_type2(
           sub(y=0,first vec_field2),x,lambda1,direction),symbolic !*rounded};  
       >>
    else
     <<sep:=center_manifold_blow_up(g,taylor_level);
       write "center manifold x=",sub(x=y,first sep);
       write "on center manifold y'=",sub(x=y,second sep);
       if save_all then
       <<openfile(result_file);
         write "center manifold x=",sub(x=y,first sep);
         write "on center manifold y'=",sub(x=y,second sep);
         out 't;
       >>;
       if gt(lambda2,0) then
        if sign(sub(x=1,last sep))=1 then
         man:={transformations,{0,0},{vec,1,1,0},g,sub(x='t,first sep),5,
               symbolic !*rounded}
                                    else
         man:={transformations,{0,0},{vec,1,1,0},g,sub(x='t,first sep),6,
               symbolic !*rounded}
                        else
        if sign(sub(x=1,last sep))=1 then
         man:={transformations,{0,0},{vec,1,1,0},g,sub(x='t,first sep),7,
               symbolic !*rounded}
                                    else
         man:={transformations,{0,0},{vec,1,1,0},g,sub(x='t,first sep),8,
               symbolic !*rounded}

     >>;
  >>
  else
  if gt(lambda1,0) then
   if nneq(lambda1,lambda2) then
    <<vec:=sub(x=0,y=0,df(second first z,x))/(lambda1-lambda2); 
      g:=transformation(first z,mat((1,0),(vec,1))); 
      sep:=manifold2_blow_up(g,taylor_level); 
      man:={transformations,{0,0},{vec,1,1,0},g,sep,3,symbolic !*rounded}; 
    >>
                            else
      man:={transformations,{0,0},{0,1,1,0},first z,0,3,symbolic !*rounded}      
                   else
   if nneq(lambda1,lambda2) then
    <<vec:=sub(x=0,y=0,df(second first z,x))/(lambda1-lambda2);
      g:=transformation(first z,mat((1,0),(vec,1)));
      sep:=manifold2_blow_up(g,taylor_level);
      man:={transformations,{0,0},{vec,1,1,0},g,sep,4,symbolic !*rounded}
    >>
                            else
    man:={transformations,{0,0},{0,1,1,0},first z,0,4,symbolic !*rounded};
   
return man
end;

procedure sort_set_pos(s,y0,r);
begin scalar a,ok,l;
l:=0;
ok:=NIL;
while (l<length s) and (not ok) do
 << l:=l+1; 
    if llt(y0,first part(s,l)) then << a:=insert({y0,r},s,l);
                                ok:='t >>
 >>;
if ok then return a
      else return append(s,{{y0,r}});
end;

procedure sort_set_neg(s,y0,r);
begin scalar a,ok,l;
l:=0;
ok:=NIL;
while (l<length s) and (not ok) do
 << l:=l+1;
    if gt(y0,first part(s,l)) then << a:=insert({y0,r},s,l);
                                ok:='t >>
 >>;
if ok then return a
      else return append(s,{{y0,r}});
end;

%CALCULATE THE SEP THAT IS TANGENT TO THE X-AXES, BUT WE
%MAKE NOT THE TEST 
procedure manifold_blow_up(g,n);
begin scalar l1,l2,f,dfx,l,h,p;
l1:=coeffn(coeffn(first g,x,1),y,0);
l2:=coeffn(coeffn(second g,y,1),x,0); 
f:=<< for l:=2:n sum mkid(sep_coeff!*,l)*x^l >>; 
dfx:=<< for l:=2:n sum mkid(sep_coeff!*,l)*l*x^(l-1) >>;
weight x=1,y=1; wtlevel n;   %x^l*y^j=0 <=> l+j>n
h:=sub(y=f,second g-l2*y)-dfx*sub(y=f,first g-l1*x);
for l:=2:n do set(mkid(sep_coeff!*,l),coeffn(h,x,l)/(l*l1-l2));
clear x,y; % clear weigthlevel
p:=<< for l:=2:n sum mkid(sep_coeff!*,l)*t^l >>;
for l:=2:n do clear mkid('sep_coeff!*,l); 
return p;
end;

procedure manifold2_blow_up(g,n);
begin scalar l1,l2,f,dfx,l,h,p;
l1:=coeffn(coeffn(first g,x,1),y,0);
l2:=coeffn(coeffn(second g,y,1),x,0);
f:=<< for l:=2:n sum mkid(sep_coeff!*,l)*x^l >>;
dfx:=<< for l:=2:n sum mkid(sep_coeff!*,l)*l*x^(l-1) >>;
weight x=1,y=1; wtlevel n;   %x^l*y^j=0 <=> l+j>n
h:=sub(y=f,second g-l2*y)-dfx*sub(y=f,first g-l1*x);
l:=2;
n1:=n;
while l<n+1 do
 if nneq(l*l1-l2,0) then
  << set(mkid(sep_coeff!*,l),coeffn(h,x,l)/(l*l1-l2));
     l:=l+1 >>
                    else
  << n1:=l-1; l:=n+1 >>;
clear x,y; % clear weigthlevel
p:=<< for l:=2:n1 sum mkid(sep_coeff!*,l)*t^l >>;
for l:=2:n do clear mkid('sep_coeff!*,l);
return p;
end;

procedure center_manifold_blow_up(g,n);
begin scalar r,l,f,dfx,h,k,p,dx,n1;
r:=symbolic !*rounded;
l:=coeffn(coeffn(second g,y,1),x,0);
f:=<< for k:=2:n sum mkid(sep_coeff!*,k)*x^k >>;
dfx:=<< for k:=2:n sum mkid(sep_coeff!*,k)*k*x^(k-1)>>;
weight x=1,y=1; wtlevel n;   %x^k*y^j=0 <=> k+j>n
h:=-(sub(y=f,second g-l*y)-dfx*sub(y=f,first g))/l;
for k:=2:n do set(mkid(sep_coeff!*,k),coeffn(h,x,k));
p:=<< for k:=2:n sum mkid(sep_coeff!*,k)*x^k >>;
dx:=sub(y=p,first g);
clear x,y;
n1:=n;
while (dx=0) do
<< n1:=n1+1;
   weight x=1,y=1; wtlevel n1;
   f:=f+mkid(sep_coeff!*,n1)*x^n1;
   dfx:=dfx+n1*mkid(sep_coeff!*,n1)*x^(n1-1);
   h:=-(sub(y=f,second g-l*y)-dfx*sub(y=f,first g))/l;
   set(mkid(sep_coeff!*,n1),coeffn(h,x,n1));
   p:=p+mkid(sep_coeff!*,n1)*x^n1;
   dx:=sub(y=p,first g);
   clear x,y  
>>;
for l:=2:n1 do clear mkid('sep_coeff!*,l);
return {p,dx,first_term(dx,x)};
end;


%PUT THE SEPARATRICE IN TABLE
procedure  
 put_separatrice(direction,type,transformations,pt,trans,vec_field,sep);
begin 
if direction=1 then
 <<!*NUMBER_X_POS!*:=!*NUMBER_X_POS!*+1;
   set(mkid(!*MAN_X_POS!*,!*NUMBER_X_POS!*),{transformations,pt,trans,
     vec_field,sep,type,symbolic !*rounded}) >>
               else
 <<!*NUMBER_X_NEG!*:=!*NUMBER_X_NEG!*+1;
   set(mkid(!*MAN_X_NEG!*,!*NUMBER_X_NEG!*),{transformations,pt,trans,
    vec_field,sep,type,symbolic !*rounded}) >>;
end;
       
procedure center_type1(f,x,l,direction);
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% return type of center on x=0        %
% f(0)=0   x-direction                %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
begin scalar s,g,m,b;
g:=first_term(f,x); % write f," ",x;
on ratarg; 
m:=deg(g,x);
b:=sign(lcof(g,x));
off ratarg;
if evenp(m) then s:={b,1,sign(evalf(l))}
            else s:={b,-1,sign(evalf(l))};
% write s;            
if direction=1 then 
 <<if s={-1,1,1} then return 10;
   if s={1,1,1} then return 9; 
   if s={1,-1,1} then return 3;
   if s={-1,-1,1} then return 1; 
   if s={-1,1,-1} then return 12;
   if s={1,1,-1} then return 11;
   if s={1,-1,-1} then return 2 else return 4;>>
               else
 <<if s={-1,1,1} then return 9;
   if s={1,1,1} then return 10;
   if s={1,-1,1} then return 3;
   if s={-1,-1,1} then return 1; 
   if s={-1,1,-1} then return 11;
   if s={1,1,-1} then return 12;
   if s={1,-1,-1} then return 2 else return 4>>;
end;

procedure sing_eigenv(z,direction,level,transformations,xpos);
begin scalar r,s,l,lambda1,lambda2,h,pt,vec,ff,trans,g,sep;
r:=symbolic !*rounded;
if third z neq 0 then
 << h:=find_real_roots1(third z,y); 
    s:={}; % set of sing points %
    for l:=1:length(first h) do
     if xpos=1 then s:=sort_set_pos(s,part(first h,l),NIL)
                    else s:=sort_set_neg(s,part(first h,l),NIL);
    for l:=1:length(second h) do
     if xpos=1 then s:=sort_set_pos(s,part(second h,l),'t)
                    else s:=sort_set_neg(s,part(second h,l),'t); 
    for l:=1:length(s) do
     << pt:=first part(s,l); 
        if second part(s,l) then on rounded;
        lambda1:=sub(y=pt,first second z);
        lambda2:=sub(y=pt,df(third z,y));
        %type of singularity%
        if eeq(lambda1,0) and eeq(lambda2,0) then
         <<write "(0,",pt,") is a non-elementary point";
           if save_all then
           << openfile(result_file);
              write "(0,",pt,") is a non-elementary point";
              out 't;
           >>;   
           blowing_up2(first z,pt,level+1,transformations,xpos,0);>>
                                             else
        if eeq(lambda1,0) or eeq(lambda2,0) or gteq(0,lambda1*lambda2) then
         <<vec:=sub(x=0,y=pt,df(second first z,x))/(lambda1-lambda2);
           ff:=translation(first z,0,pt);
           trans:={1,0,vec,1};
           g:=transformation(ff,mat((1,0),(vec,1)));
           write "(0,",pt,") is a singular point with eigenvalues ",
            lambda1," ",lambda2;

           write "moving this point to the origin";
           write "the new vector field is ",ff;
           write 
            "Transform this vector field so that the linear part is diagonal";
           write "The new vector field is ",g; 
           if save_all then
           <<openfile(result_file);
             write "(0,",pt,") is a singular point with eigenvalues ",
              lambda1," ",lambda2;

             write "moving this point to the origin";
             write "the new vector field is ",ff;
             write 
              "Transform this vector field so that the linear part is diagonal";
             write "The new vector field is ",g;
             out 't;
           >>;  
           if gt(lambda1,0) then 
            <<sep:=manifold_blow_up(g,taylor_level);
              write "unstable separatrice y=",sub(t=x,sep);
              if save_all then
              <<openfile(result_file);
                write "unstable separatrice y=",sub(t=x,sep);
                out 't;
              >>;  
              if nneq(lambda2,0) then
               put_separatrice(xpos,1,transformations,
                {0,pt},trans,g,sep)
                                 else
               put_separatrice(xpos,
                center_type1(sub(x=0,second g),y,lambda1,direction),
                transformations,{0,pt},trans,g,sep)
            >>
                           else
            if gt(0,lambda1) then
            <<sep:=manifold_blow_up(g,taylor_level);
              write "stable separatrice y=",sub(t=x,sep);
              if save_all then
              <<openfile(result_file);
                write "stable separatrice y=",sub(t=x,sep);
                out 't;
              >>;  
              if nneq(lambda2,0) then   
               put_separatrice(xpos,2,transformations,
                {0,pt},trans,g,sep) 
                                 else
               put_separatrice(xpos,
                center_type1(sub(x=0,second g),y,lambda1,direction),
                transformations,{0,pt},trans,g,sep)
            >>
                             else
           
           <<sep:=center_manifold_blow_up(g,taylor_level);
             write "center manifold y=",first sep;
             write "on center manifold is x'=",second sep;
             if save_all then
             <<openfile(result_file);
               write "center manifold y=",first sep;
               write "on center manifold is x'=",second sep;
               out 't;
             >>;  
             if gt(lambda2,0) then
              if sign(sub(x=1,last sep))=1 then
               put_separatrice(xpos,5,transformations,{0,pt},
                trans,g,sub(x='t,first sep))
                                           else
                put_separatrice(xpos,6,transformations,{0,pt},
                trans,g,sub(x='t,first sep))
                              else
              if sign(sub(x=1,last sep))=1 then
               put_separatrice(xpos,7,transformations,{0,pt},
                trans,g,sub(x='t,first sep))
                                           else
                put_separatrice(xpos,8,transformations,{0,pt},
                trans,g,sub(x='t,first sep))            
            >>         
          >>
                    else
          if gt(lambda1,0) then
           <<write "(0,",pt,") is a unstable node with eigenvalues ",
              lambda1," ",lambda2;
             if save_all then
             <<openfile(result_file);
               write "(0,",pt,") is a unstable node with eigenvalues ",
                lambda1," ",lambda2;
               out 't;
             >>;
           if nneq(lambda1,lambda2) then
           <<vec:=sub(x=0,y=pt,df(second first z,x))/(lambda1-lambda2);
             ff:=translation(first z,0,pt);
             trans:={1,0,vec,1};
             g:=transformation(ff,mat((1,0),(vec,1)));           
             sep:=manifold2_blow_up(g,taylor_level);
             put_separatrice(xpos,3,transformations,
               {0,pt},trans,g,sep) 
            >>
                                      else
             put_separatrice(xpos,3,transformations,
                {0,pt},{1,0,0,1},translation(first z,0,pt),0)       
           >>
                           else
           << write "(0,",pt,") is a stable node with eigenvalues ",
                lambda1," ",lambda2;
               if save_all then
               <<openfile(result_file);
                 write "(0,",pt,") is a stable node with eigenvalues ",
                  lambda1," ",lambda2;
                 out 't;
               >>;                
            if nneq(lambda1,lambda2) then
             <<vec:=sub(x=0,y=pt,df(second first z,x))/(lambda1-lambda2);
               ff:=translation(first z,0,pt);
               trans:={1,0,vec,1};
               g:=transformation(ff,mat((1,0),(vec,1)));         
               sep:=manifold2_blow_up(g,taylor_level);
               put_separatrice(xpos,4,transformations,
                  {0,pt},trans,g,sep) >>
                                     else
               put_separatrice(xpos,4,transformations,
                  {0,pt},{1,0,0,1},translation(first z,0,pt),0)
         >>         
     >>;
     if not r then off rounded;
 >>
else 
<<  %X=0 IS A LINE OF SINGULARITIES!!!!!
    write "x=0 is a line of singularities";
    h:=find_real_roots1(first second z,y);
    s:={}; 
    for l:=1:length(first h) do
     if xpos=1 then s:=sort_set_pos(s,part(first h,l),NIL)
                    else s:=sort_set_neg(s,part(first h,l),NIL);
    for l:=1:length(second h) do
     if xpos=1 then s:=sort_set_pos(s,part(second h,l),'t)
                    else s:=sort_set_neg(s,part(second h,l),'t);
    openfile(result_file);
    write "X=0 IS A LINE OF SINGULARITIES";
    write "the non-elementary singular points are :";
    for l:=1:length(first h) do write "(0,",part(first h,l),")";
    on rounded;
    for l:=1:length(second h) do write "(0,",part(second h,l),")";
    if not r then off rounded;
    out 't;
    for l:=1:length(s) do
     << if (not r) and (second part(s,l)) then on rounded;
        blowing_up2(first z,first part(s,l),level+1,transformations,xpos,1);
        if not r then off rounded;
     >>;
  
>>;  
end;

% need otherwise in transformation something like 0^0
procedure power(a,b);
begin
 if b=0 then return 1
 else return a^b;
end;

procedure make_transformation(sep);
begin scalar x0,y0,x1,y1,trans;
trans:=part(sep,3);
x0:=(first second sep)+(first trans)*t+(second trans)*part(sep,5);
y0:=(second second sep)+(third trans)*t+(last trans)*part(sep,5);
trans:=first sep;
l:=length(trans); 
while l>0 do
<< a:=part(trans,l);
   x1:=(first a)+part(a,3)*power(x0,part(a,5))*power(y0,part(a,6));
   y1:=(second a)+part(a,4)*power(x0,part(a,7))*power(y0,part(a,8));    
   x0:=x1; y0:=y1;
   if (x0=0) and (y0=0) then l:=0 else l:=l-1;
 >>;
return {x0,y0};
end;


procedure append_trajectors();
begin scalar man1,man2,l,ok,w1,w2,s1,s2,s3,j;
man1:={};
if !*NUMBER_Y!* neq 0 then
   << man2:=make_transformation(!*MAN_Y_NEG!*);
      %if man2 neq {0,0} then 
      man1:=append(man1,{append(!*MAN_Y_NEG!*,{man2})});
   >>;
for l:=1:!*NUMBER_X_POS!* do
 <<man2:=make_transformation(mkid(!*MAN_X_POS!*,l));
  % if man2 neq {0,0} then 
   man1:=append(man1,{append(mkid(!*MAN_X_POS!*,l),{man2})});
 >>;
if !*NUMBER_Y!* neq 0 then
 << man2:=make_transformation(!*MAN_Y_POS!*);
   % if man2 neq {0,0} then
    man1:=append(man1,{append(!*MAN_Y_POS!*,{man2})});
  >>;
for l:=1:!*NUMBER_X_NEG!* do
 <<man2:=make_transformation(mkid(!*MAN_X_NEG!*,l));
   %if man2 neq {0,0} then 
   man1:=append(man1,{append(mkid(!*MAN_X_NEG!*,l),{man2})});
 >>;

%ok:='nil;
%w1:={1,3,5,7,9,10};
%w2:={2,4,6,8,11,12};
%while not ok do
% << l:=length(man1);
%    j:=0; ok:='t;
%    while (j<l) and ok do
%     << j:=j+1;
%        s2:=part(part(man1,j),6);
%        if j neq 1 then s1:=part(part(man1,j-1),6)
%                   else s1:=part(part(man1,l),6);
%        if j neq l then s3:=part(part(man1,j+1),6)
%                   else s3:=part(part(man1,1),6);
%        s2:=part(part(man1,j),6);
%        if member(s1,w1) and member(s2,w1) and member(s3,w1) 
%         then
%         << ok:='nil; 
%            man1:=remove(man1,j); 
%          >>
%         else
%        if member(s1,w2) and member(s2,w2) and member(s3,w2) 
%          then
%         << ok:='nil; 
%            man1:=remove(man1,j); 
%          >>;
%      >>;
%  >>;
return man1;
end;

% if point is infinite then split the trajectors in a positive ( g'(t)>0) and
% negative ( g'(t)<0) part when sep=(f(t),g(t))
procedure split_trajectors(sep);
begin scalar sep_pos,sep_neg,j,s,g;
sep_pos:={}; sep_neg:={};
for j:=1:length(sep) do
 <<s:=second last part(sep,j); 
   if s neq 0 then
   << g:=sign(evalf(sub(t=1,first_term(df(s,t),t))));
      if g=1 then sep_pos:=append(sep_pos,{part(sep,j)})
             else sep_neg:=append(sep_neg,{part(sep,j)});
   >>
 >>;
return({sep_pos,sep_neg});
end;
 
      

procedure write_trajectors(sep,x0,y0);
begin scalar l,w1,w2,s1,s2,j,el,hyp;
if sep neq {} then <<
write " " ; write " ";
write "INVARIANT SEPARATRICES (t>0)";
openfile(result_file);
write " " ; write " ";
write "The invariant separatrices are (t>0)";
out 't;
l:=length(sep);
w1:={1,3,5,7,9,10};
w2:={2,4,6,8,11,12};
s1:=part(first sep,6); i % write s1; 
if member(s1,w1) then <<write last first sep, ": unstable";
                        openfile(result_file);
                        write last first sep, ": unstable";
                        out 't;
                      >>
                 else <<write last first sep, ": stable";
                        openfile(result_file);
                        write last first sep, ": stable";
                        out 't;
                      >>;
for j:=2:l do
 << s2:=part(part(sep,j),6); % write s2;
    if (member(s1,{1,7,10}) and member(s2,{2,6,12}))
                            or
       (member(s1,{2,6,11}) and member(s2,{1,7,9}))
       then <<write "hyperbolic sector";
              openfile(result_file);
              write "hyperbolic sector";
              out 't;
              hyp:=hyp+1;
            >>
       else
    if (member(s1,{3,5,9}) and member(s2,{4,8,11}))
                            or
       (member(s1,{4,8,12}) and member(s2,{3,5,10}))
       then <<write "elliptic sector";
              openfile(result_file);
              write "elliptic sector";
              out 't;
              el:=el+1;
            >>
       else <<write "parabolic sector";
              openfile(result_file);
              write "parabolic sector";
              out 't;
            >>;
    if member(s2,w1) then <<write last part(sep,j), ": unstable";
                            openfile(result_file);
                            write last part(sep,j), ": unstable";
                            out 't;
                          >>
                     else <<write last part(sep,j), ": stable";   
                            openfile(result_file);
                            write last part(sep,j), ": stable";
                            out 't;
                          >>;
    s1:=s2 
  >>;
s1:=part(last sep,6);
s2:=part(first sep,6);
if (member(s1,{1,7,10}) and member(s2,{2,6,12}))
                         or
   (member(s1,{2,6,11}) and member(s2,{1,7,9}))
 then <<write "hyperbolic sector";
        openfile(result_file);
        write "hyperbolic sector";
        out 't;
        hyp:=hyp+1;
      >>
 else
if (member(s1,{3,5,9}) and member(s2,{4,8,11}))
                       or
   (member(s1,{4,8,12}) and member(s2,{3,5,10}))
 then <<write "elliptic sector";
        openfile(result_file);
        write "elliptic sector";
        out 't;
        el:=el+1;
      >>
 else <<write "parabolic sector";
        openfile(result_file);
        write "parabolic sector";
        out 't;
      >>;
>>;
write "The singularity (",x0,",",y0,") has index ", 1+(el-hyp)/2;
openfile(result_file);
write "The singularity (",x0,",",y0,") has index ", 1+(el-hyp)/2;
end;

procedure blowing_upy2(vec_field,q_degree,direction,transformations,xpos);
begin scalar vec_field2,z,sx,sy,jac,vec;
vec_field2:=transformation(vec_field,mat((0,1),(1,0)));
z:=blow_upx(vec_field2,{second q_degree,first q_degree,third q_degree},
  direction,x,y);
vec_field2:=transformation(first z,mat((0,1),(1,0)));  
if direction=1 then
 << write "blowing up in the positive y-direction";
    write "blowing up vector field",vec_field2;
    if save_all then
    << openfile(result_file);
       write "blowing up in the positive y-direction";
       write "blowing up vector field",vec_field2;
       out 't;
    >>
  >>
else
 << write "blowing up in the negative y-direction";
    write "blowing up vector field",vec_field2;
    if save_all then
    << openfile(result_file);
       write "blowing up in the negative y-direction";
       write "blowing up vector field",vec_field2;
       out 't;
    >>
 >>;
% behaviour on the pos x-axis and pos y-axis
on ratarg;
sx:=sign(lcof(first_term(sub(y=0,first vec_field2),x),x));
sy:=sign(lcof(first_term(sub(x=0,second vec_field2),y),y));
off ratarg;
if sx=sy then  
 if sx=1 then put_separatrice(xpos,3,transformations,{0,0},{1,0,0,1}
                ,vec_field2,t)
         else put_separatrice(xpos,4,transformations,{0,0},{1,0,0,1},
                 vec_field2,t);
end;                      

procedure blowing_up2(vec_field,pt,level,transformations,xpos,sing);
begin scalar vec_field2,n_d,g1,q_degree,z,r,trajectors;
r:=symbolic !*rounded;
write "blowing up level ",level;
write "------------------";
write "blowing up ",vec_field," in point ",{0,pt};
if save_all then
<<openfile(result_file);
  write "blowing up level ",level;
  write "------------------";
  write "blowing up ",vec_field," in point ",{0,pt};
  out 't;
>>;  


if evalf(pt) neq 0 then
 << vec_field2:=translation(vec_field,0,pt);
    write "moving ",{0,pt}," to the origin";
    write "new vector field ",vec_field2;
    if save_all then
    <<openfile(result_file);
      write "moving ",{0,pt}," to the origin";
      write "new vector field ",vec_field2;
      out 't;
    >>;  
 >>
else vec_field2:=vec_field;
off rounded;
n_d:=newton_diagram(vec_field2);
write "newton diagram ",n_d;
g1:=gamma1(n_d);
q_degree:=quasi_degree(g1);
if r then on rounded;
write "{alpha,beta,degree}=",q_degree;
if save_all then
<<openfile(result_file);
  write "newton diagram ",n_d;
  write "{alpha,beta,degree}=",q_degree;
  out 't;
>>;  

% First we blow up in the negative y-direction if xpos=1
% else we blow up in the positive y-direction
% We have to know whats happen in the corner
% Only in the case that the X=0 is not a line of singularities
if sing=0 then
<<if xpos=1 then blowing_upy2(vec_field2,q_degree,-1,
   append(transformations,{{0,pt,1,-1,1,first q_degree,0,
   second q_degree,last q_degree}}),xpos)
  else blowing_upy2(vec_field2,q_degree,1,
    append(transformations,{{0,pt,1,1,1,first q_degree,0,
    second q_degree,last q_degree}}),xpos);
>>;


write "blowing up in the positive x direction";
z:=blow_upx(vec_field2,q_degree,1,x,y);
write "blowing up vector field ",first z;
if save_all then
<<openfile(result_file);
  write "blowing up in the positive x direction";
  write "blowing up vector field ",first z;
  out 't;
>>;  
sing_eigenv(z,1,level,append(transformations,
 {{0,pt,1,1,first q_degree,0,second q_degree,1,last q_degree}}),xpos);
 
% Now we blow up in the positive y-direction if xpos=1
% else we blow up in the negative y-direction
if sing=0 then
<<if xpos=1 then blowing_upy2(vec_field2,q_degree,1,
    append(transformations,{{0,pt,1,1,1,first q_degree,0,
    second q_degree,last q_degree}}),xpos)
  else blowing_upy2(vec_field2,q_degree,-1,
    append(transformations,{{0,pt,1,-1,1,first q_degree,0,
    second q_degree,last q_degree}}),xpos);
>>;
   
write "---------------------------------------";
write "Back to level ",level-1;
if save_all then
<<openfile(result_file);
  write "---------------------------------------";
  write "Back to level ",level-1;
  out 't;
>>;
end;



procedure blowing_up(vec_field,x0,y0,finite_point);
begin scalar r,vec_field2,n_d,g1,q_degree,z;
r:=symbolic !*rounded;
write "blowing up ",vec_field," in point (",x0,",",y0,")";
if save_all then
<<openfile(result_file);
  write "blowing up ",vec_field," in point (",x0,",",y0,")";
  out 't;
>>;  
if (x0 neq 0) or (y0 neq 0) then
 << vec_field2:=translation(vec_field,x0,y0);
    write "moving (",x0,",",y0,") to the origin";
    write "new vector field ",vec_field2;
    if save_all then
    <<openfile(result_file);
      write "moving (",x0,",",y0,") to the origin";
      write "new vector field ",vec_field2;
      out 't;
    >>;  
 >>
               else vec_field2:=vec_field;
off rounded;
n_d:=newton_diagram(vec_field2); 
g1:=gamma1(n_d); 
q_degree:=quasi_degree(g1);
if r then on rounded;
write "newton diagram ",n_d;
write "g1=",g1;
write "{alpha,beta,degree}=",q_degree;
write;
write "1. blowing up in the positive x-direction";
write "-----------------------------------------";
write;
write "blowing up level 1";
write "------------------";
write "blowing up ",vec_field2," in point (0,0)";
if save_all then
<<openfile(result_file);
  write "newton diagram ",n_d; 
  write "{alpha,beta,degree}=",q_degree;
  write;
  write "1. blowing up in the positive x-direction";
  write "-----------------------------------------";
  write;
  write "blowing up level 1";
  write "------------------";
  write "blowing up ",vec_field2," in point (0,0)";
  out 't;
>>;  
z:=blow_upx(vec_field2,q_degree,1,x,y);
write "blowing up vectorfield ",first z;
if save_all then
<<openfile(result_file);
  write "blowing up vectorfield ",first z;
  out 't;
>>; 
!*NUMBER_X_POS!*:=0; %number of invariant sep in the pos x-direction%
sing_eigenv(z,1,1,{{x0,y0,1,1,first q_degree,0,second q_degree,1,last q_degree}},1);

%for l:=1:!*NUMBER_X_POS!* do
% write mkid(!*MAN_X_POS!*,l);

write "---------------------------------------------------------";
write;
write "2. blowing up in the negative x-direction";
write "-----------------------------------------";
write;
write "blowing up level 1";
write "------------------";
write "blowing up ",vec_field2," in point (0,0)";
if save_all then
<<openfile(result_file);
  write "---------------------------------------------------------";
  write;
  write "2. blowing up in the negative x-direction";
  write "-----------------------------------------";
  write;
  write "blowing up level 1";
  write "------------------";
  write "blowing up ",vec_field2," in point (0,0)";
  out 't;
>>;
z:=blow_upx(vec_field2,q_degree,-1,x,y);
write "blowing up vectorfield ",first z;
if save_all then
<<openfile(result_file);
  write "blowing up vectorfield ",first z;
  out 't;
>>;  
!*NUMBER_X_NEG!*:=0; %number of invariant sep in the pos x-direction%
sing_eigenv(z,-1,1,{{x0,y0,-1,1,first q_degree,0,second q_degree,1,last q_degree}},-1);

%for l:=1:!*NUMBER_X_NEG!* do
% write mkid(!*MAN_X_NEG!*,l);

%%%%%% (0,pi/2) a singular point ?
% BLOWING UP IN THE Y DIRECTION ?
!*NUMBER_Y!*:=0;
if first first g1 >-1 then
 << write "3. blowing up in the positive y-direction";
    write "-----------------------------------------";
    if save_all then
    <<openfile(result_file);
      write "3. blowing up in the positive y-direction";
      write "-----------------------------------------";
      out 't;
    >>;
    !*MAN_Y_POS!*:=blow_upy(vec_field2,q_degree,1,{{x0,y0,1,1,1,first 
                            q_degree,0,second q_degree,last q_degree}});
    !*NUMBER_Y!*:=1;
    write;
    write "4. blowing up in the negative y-direction";
    write "-----------------------------------------";
    if save_all then
    <<openfile(result_file);
      write;
      write "4. blowing up in the negative y-direction";
      write "-----------------------------------------";
      out 't;
    >>;
    !*MAN_Y_NEG!*:=blow_upy(vec_field2,q_degree,-1,{{x0,y0,1,-1,1,first 
                            q_degree,0,second q_degree,last q_degree}})
 >>;

trajectors:=append_trajectors();
write_trajectors(trajectors,x0,y0);
if finite_point=0 then trajectors:=split_trajectors(trajectors);

%clear some variables
for l:=1:!*NUMBER_X_POS!* do clear mkid('!*MAN_X_POS!*,l);
for l:=1:!*NUMBER_X_NEG!* do clear mkid('!*MAN_X_NEG!*,l);

return(trajectors);
end;


