procedure jacobian(f,x0,y0);
begin scalar jac;
 jac:={sub({x=x0,y=y0},df(first f,x)),sub({x=x0,y=y0},df(first f,y)),
      sub({x=x0,y=y0},df(second f,x)),sub({x=x0,y=y0},df(second f,y))};
return jac;
end;

procedure eigenvector(a,b,c,d);
begin scalar c;
if nneq(b,0) then c:={1,-a/b}
else if nneq(a,0) then c:={0,1}
else if nneq(d,0) then c:={1,-c/d}
else c:={0,1};
if symbolic !*rounded then <<d:=sqrt(first c^2+second c^2); 
	                     c:={first c/d,second c/d};
	                   >>;
return c;
end;

procedure translation(v,x0,y0);
begin scalar f1,f2;
f1:=sub(x=x+x0,y=y+y0,first v);
f2:=sub(x=x+x0,y=y+y0,second v);
if (symbolic !*rounded) then <<f1:=change_poly(f1);
                               f2:=change_poly(f2)>>;
return {f1,f2}
end;

procedure transformation(f,a);
begin scalar g1,g2,h1,h2,b;
 g1:=sub({x=a(1,1)*xx+a(1,2)*yy,y=a(2,1)*xx+a(2,2)*yy},first f);
 g2:=sub({x=a(1,1)*xx+a(1,2)*yy,y=a(2,1)*xx+a(2,2)*yy},second f);
 b:=a^(-1);
 h1:=sub({xx=x,yy=y},b(1,1)*g1+b(1,2)*g2);
 h2:=sub({xx=x,yy=y},b(2,1)*g1+b(2,2)*g2);
 if (symbolic !*rounded) then <<h1:=change_poly(h1);h2=change_poly(h2)>>;
 return {h1,h2};
end;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% MAKING A 1000 NORMALIZATION %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure normalization(f);
begin scalar m,g,h;
on ratarg;
m:=0;
g:=first f;
while g neq 0 do
 << h:=nlterm(g,x,y);
    g:=g-h;
    h:=coeffn(coeffn(h,x,deg(h,x)),y,deg(h,y));
    if evalf(m)<abs(evalf(h)) then m:=abs(h);
 >>;
g:=second f;
while g neq 0 do
 << h:=nlterm(g,x,y);
    g:=g-h;
    h:=coeffn(coeffn(h,x,deg(h,x)),y,deg(h,y));
    if evalf(m)<abs(evalf(h)) then m:=abs(h);
 >>;
off ratarg;
return {1000*(first f)/m,1000*(second f)/m};
end;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% CALCULATE SEP TANGENT TO THE X_AXES %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

procedure manifold(g,n,stable);
begin scalar f,l,p,h,dfx,l1,l2,ok,n1,r;
r:=symbolic !*rounded;
l1:=coeffn(coeffn(first g,x,1),y,0);
l2:=coeffn(coeffn(second g,y,1),x,0);
f:=<< for l:=2:n sum mkid(sep_coeff!*,l)*x^l >>;
dfx:=<< for l:=2:n sum mkid(sep_coeff!*,l)*l*x^(l-1) >>;
weight x=1,y=1; wtlevel n;   %x^l*y^j=0 <=> l+j>n
h:=sub(y=f,second g-l2*y)-dfx*sub(y=f,first g-l1*x);
for l:=2:n do <<set(mkid(sep_coeff!*,l),coeffn(h,x,l)/(l*l1-l2));
%if eeq(mkid(sep_coeff!*,l),0) then 
% <<clear mkid('sep_coeff!*,l); set(mkid(sep_coeff!*,l),0); >>;
>>;
clear x,y; % clear weigthlevel


% TEST THE SEPERATRICE FOR x=-2*epsilon...2*epsilon

% TEST FOR x=0..2*epsilon
n1:=n;
if test_sep=0 then
<<
ok:=separatrice_test(g,n,stable,1);
while (ok=0) and (n1<max_level) do  
 << n1:=n1+1;
    if n1=numeric_level then on rounded;
    weight x=1,y=1;wtlevel n1;
    f:=f+mkid(sep_coeff!*,n1)*x^n1;
    dfx:=dfx+n1*mkid(sep_coeff!*,n1)*x^(n1-1);
    h:=sub(y=f,second g-l2*y)-dfx*sub(y=f,first g-l1*x);
    set(mkid(sep_coeff!*,n1),coeffn(h,x,n1)/(n1*l1-l2));
%    if eeq(mkid(sep_coeff!*,n1),0) then 
%     <<clear mkid('sep_coeff!*,n1);set(mkid(sep_coeff!*,n1),0);>>;
    clear x,y;
    ok:=separatrice_test(g,n1,stable,1);
  >>;

write "TEST FOR x=0..(-2*epsilon)";

ok:=separatrice_test(g,n1,stable,-1);
while (ok=0) and (n1<max_level) do 
 << n1:=n1+1;
    if n1=numeric_level then on rounded;
    weight x=1,y=1;wtlevel n1;
    f:=f+mkid(sep_coeff!*,n1)*x^n1;
    dfx:=dfx+n1*mkid(sep_coeff!*,n1)*x^(n1-1);
    h:=sub(y=f,second g-l2*y)-dfx*sub(y=f,first g-l1*x);
    set(mkid(sep_coeff!*,n1),coeffn(h,x,n1)/(n1*l1-l2));
%    if eeq(mkid(sep_coeff!*,n1),0) then 
%     <<clear mkid('sep_coeff!*,n1); set(mkid('sep_coeff!*,n1),0);>>;
    clear x,y;
    ok:=separatrice_test(g,n1,stable,-1);
  >>;
>>;

if not r then off rounded;
p:=<< for l:=2:n1 sum mkid(sep_coeff!*,l)*t^l >>;
for l:=2:n1 do clear mkid('sep_coeff!*,l);
return p;
end;

procedure center_manifold(g,n,stable);
begin scalar r,l,f,dfx,h,k,p,dx,n1,b,m,a_m,type,stable2;

 r:=symbolic !*rounded;
 l:=coeffn(coeffn(second g,y,1),x,0);
 f:=<< for k:=2:n sum mkid(sep_coeff!*,k)*x^k >>;
 dfx:=<< for k:=2:n sum mkid(sep_coeff!*,k)*k*x^(k-1)>>;
 weight x=1,y=1; wtlevel n;   %x^k*y^j=0 <=> k+j>n
 h:=-(sub(y=f,second g-l*y)-dfx*sub(y=f,first g))/l;
 for k:=2:n do <<set(mkid(sep_coeff!*,k),coeffn(h,x,k));
%  if eeq(mkid(sep_coeff!*,l),0) then 
%   <<clear mkid('sep_coeff!*,l); set(mkid(sep_coeff!*,l),0); >>;
 >>;
 p:=<< for k:=2:n sum mkid(sep_coeff!*,k)*x^k >>;
 dx:=sub(y=p,first g);
 clear x,y;
 n1:=n;
 while (dx=0) do
 << n1:=n1+1;
    weight x=1,y=1; wtlevel n1;
    f:=f+mkid(sep_coeff!*,n1)*x^n1;
    dfx:=dfx+n1*mkid(sep_coeff!*,n1)*x^(n1-1);
    h:=-(sub(y=f,second g-l*y)-dfx*sub(y=f,first g))/l;
    set(mkid(sep_coeff!*,n1),coeffn(h,x,n1));
 %   if eeq(mkid(sep_coeff!*,n1),0) then 
 %    <<clear mkid('sep_coeff!*,n1); set(mkid(sep_coeff!*,n1),0); >>;
    p:=p+mkid(sep_coeff!*,n1)*x^n1;
    dx:=sub(y=p,first g);
    clear x,y  
 >>;

 %
 % Determing the type of semi_elementary
 %
 b:=first_term(dx,x);
 on ratarg; m:=deg(b,x); off ratarg;
 a_m:=sign(evalf(sub(x=1,b)));
 if evenp(m) then
  if a_m=1 then
   if  stable=0 then type:=1 % saddle-node
                else type:=2 % saddle-node
           else
   if stable=0 then type:=3  % saddle-node
               else type:=4  % saddle-node
            else
  if a_m=1 then
   if stable=0 then type:=5  % unstable node
               else type:=6  % saddle
           else
   if stable=0 then type:=7  % saddle
               else type:=8; % stable node

 
 %TESTING THE SEPERATRICE
if test_sep=0 then
<< % TEST FOR x=0..2*epsilon
 if member(type,{2,6}) then stable2:=0 else stable2:=1;
 if member(type,{2,6,3,7}) then
  << ok:=separatrice_test(g,n1,stable2,1);
     while (ok=0) and (n1<max_level) do  % NOG NAKIJKEN !!!!!!
     << n1:=n1+1;
        if n1=numeric_level then on rounded;
        weight x=1,y=1;wtlevel n1;
        f:=f+mkid(sep_coeff!*,n1)*x^n1;
        dfx:=dfx+n1*mkid(sep_coeff!*,n1)*x^(n1-1);
        h:=-(sub(y=f,second g-l*y)-dfx*sub(y=f,first g))/l;
        set(mkid(sep_coeff!*,n1),coeffn(h,x,n1));
 %       if eeq(mkid(sep_coeff!*,n1),0) then 
 %        <<clear mkid('sep_coeff!*,n1); set(mkid(sep_coeff!*,n1),0); >>;
        p:=p+mkid(sep_coeff!*,n1)*x^n1;
        clear x,y;
        ok:=separatrice_test(g,n1,stable2,1);
     >>;


   >>;

 %TEST FOR x=0..(-2*epsilon)
 if member(type,{4,6}) then stable2:=0 else stable2:=1;
 if member(type,{4,6,1,7}) then
  <<ok:=separatrice_test(g,n1,stable2,-1);
    while (ok=0) and (n1<max_level) do 
     << n1:=n1+1;
        if n1=numeric_level then on rounded;
        weight x=1,y=1;wtlevel n1;
        f:=f+mkid(sep_coeff!*,n1)*x^n1;
        dfx:=dfx+n1*mkid(sep_coeff!*,n1)*x^(n1-1);
        h:=-(sub(y=f,second g-l*y)-dfx*sub(y=f,first g))/l;
        set(mkid(sep_coeff!*,n1),coeffn(h,x,n1));
  %      if eeq(mkid(sep_coeff!*,n1),0) then 
  %       <<clear mkid('sep_coeff!*,n1); set(mkid(sep_coeff!*,n1),0); >>;
        p:=p+mkid(sep_coeff!*,n1)*x^n1;
        clear x,y;
        ok:=separatrice_test(g,n1,stable2,-1);
    >>;
   >>;
>>;
 for k:=2:n1 do clear mkid('sep_coeff!*,k);
 return {p,dx,type};
end;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% (x0,y0) is a saddle point%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure saddle(f,x0,y0,a,b,c,d,rho,sigma,chart);
begin scalar ff,g,g1,lambda1,lambda2,v1,v2,sep1,sep2,stable,a11,a12,a21,a22;

write "(",x0,",",y0,") is a saddle point";
openfile(result_file);
write "(",x0,",",y0,") is a saddle point";
out 't;

lambda1:=(rho-sqrt(sigma))/2;
lambda2:=(rho+sqrt(sigma))/2;
v1:=eigenvector(a-lambda1,b,c,d-lambda1);
v2:=eigenvector(a-lambda2,b,c,d-lambda2);
if save_all then << openfile(result_file);
 write " The eigenvalues are :",lambda1," ",lambda2;
 write " The eigenvectors are :",v1," ",v2;
 out 't;
>>;
 %
 %MOVING (x0,y0) TO THE ORIGIN
 %
 
ff:=translation(f,x0,y0);
trans:=mat((first v1,first v2),(second v1,second v2));
a11:=first v1; a12:=first v2; a21:=second v1; a22:=second v2;

 
 %
 % MAKE A TRANSFORMATION SO THAT THE EIGENVECTORS ARE (1,0) RESP (0,1)
 %
 
g1:=transformation(ff,trans);

 
 %
 % MAKE A 1000-NORMALIZATION 
 %

g:=normalization(g1); 
 %
 % CALCULATE THE SEP THAT IS TANGENT TO THE X-AXES
 %

write "Moving this saddle point to the origin";
write "The new vector field is ",ff;
write "Transform this vector field so that the linear part is diagonal";
write "The new vector field is ",g1;  
write "Make a 1000-normalization";
write "So the new vector field is ",g;
write "We calculate the invariant separatrices";

if save_all then 
<< openfile(result_file);
   write "Moving this saddle point to the origin";
   write "The new vector field is ",ff;
   write "Transform this vector field so that the linear part is diagonal";
   write "The new vector field is ",g1;  
   write "Make a 1000-normalization";
   write "So the new vector field is ",g;
   write "We calculate the invariant separatrices";
   out 't;
>>;   

if evalf(lambda1)<0 then stable:=1 else stable:=0;
sep1:=manifold(g,taylor_level,stable); 
 
 %
 % CALCULATE THE SEP THAT IS TANGENT TO THE Y-AXES
 %
 
if evalf(lambda2)<0 then stable:=1 else stable:=0;
sep2:=manifold(transformation(g,mat((0,1),(1,0))),taylor_level,stable);


if gt(lambda1,0) then
  <<write "The stable separatrice is ",{sep2,t};
    write "The unstable separatrice is ",{t,sep1}; 
    openfile(result_file);
    if save_all then
    << write "The stable separatrice is ",{sep2,t};
       write "The unstable separatrice is ",{t,sep1};
    >>;
    out 't;
  >>  
		  else
  <<write "The stable separatrice is ",{t,sep1};
    write "the unstable separatrice is ",{sep2,t}; 
    openfile(result_file);
    if save_all then
    << write "The stable separatrice is ",{t,sep1};
       write "the unstable separatrice is ",{sep2,t};
    >>;  
   out 't;
  >>;
write "######################################################################################";
openfile(result_file);
write "######################################################################################";
out 't;
write_saddle(x0,y0,a11,a12,a21,a22,
              g,{sep1,sep2},{sign(evalf(lambda1)),sign(evalf(lambda2))},chart);
          
end;
 

procedure strong_focus(sigma,rho,x0,y0,chart);
begin scalar lambda1,lambda2,l;

lambda1:=(rho-sqrt(sigma))/2;
lambda2:=(rho+sqrt(sigma))/2;
if gt(rho,0) then <<write "(",x0,",",y0,") is a strong unstable focus";
                    openfile(result_file);
                    write "(",x0,",",y0,") is a strong unstable focus";
                    out 't;
                    l:=1; >>
	      else <<write "(",x0,",",y0,") is a strong stable focus";
                    openfile(result_file);
                    write "(",x0,",",y0,") is a strong stable focus";
                    out 't;
                    l:=-1; >>;
write "######################################################################################";
openfile(result_file);
write "######################################################################################";
out 't;
write_strong_focus(x0,y0,l,chart);
end;

procedure node(f,sigma,rho,x0,y0,chart);
begin scalar l;
if llt(rho,0) then 
 <<write "(",x0,",",y0,") is a stable node";
   openfile(result_file);
   write "(",x0,",",y0,") is a stable node";
   out 't;
   l:=-1>>
             else
 <<write "(",x0,",",y0,") is a unstable node";
   openfile(result_file);
   write "(",x0,",",y0,") is a unstable node";
   out 't;
   l:=1>>;
write "######################################################################################";
openfile(result_file);
write "######################################################################################";
out 't;
write_node(x0,y0,l,chart);
end;   

procedure degenerate(f,x0,y0,chart);
begin scalar trajectors;
write "(",x0,",",y0,") is a non-elementary point";
write "blowing up the vector field";
openfile(result_file);
write "(",x0,",",y0,") is a non-elementary point";
out 't;
trajectors:=blowing_up(f,x0,y0,1); % 1 finite point 
write "######################################################################################";
openfile(result_file);
write "######################################################################################";
out 't;
write_degenerate(x0,y0,trajectors,chart);
end;

procedure semi_elementary(f,x0,y0,a,b,c,d,lambda2,chart);
begin scalar v1,v2,ff,g,g1,center_sep,sep1,sep2,type,stable,a11,a12,a21,a22;
 
v1:=eigenvector(a,b,c,d);
v2:=eigenvector(a-lambda2,b,c,d-lambda2);
trans:=mat((first v1,first v2),(second v1,second v2));
a11:=first v1; a12:=first v2; a21:=second v1; a22:=second v2;
ff:=translation(f,x0,y0);
g1:=transformation(ff,trans);
g:=normalization(g1);

write "(",x0,",",y0,") is semi-hyperbolic";
write "Moving this point to the origin";
write "The new vector field is",ff;
write "Transform this vector vector field so that the linear part is diagonal";
write "The new vector field is",g1;
write "Make a 1000-normalization";
write "So the new vector field is ",g;
write "We calculate the invariant separatrices";
openfile(result_file);
write "(",x0,",",y0,") is semi-hyperbolic";
if save_all then 
<< write "Moving this point to the origin";
   write "The new vector field is",ff;
   write "Transform this vector vector field so that the linear part is diagonal";
   write "The new vector field is",g1;
   write "Make a 1000-normalization";
   write "So the new vector field is ",g;
   write "We calculate the invariant separatrices";
>>;   
out 't;

 %
 % CALCULATE THE CENTER SEP
 % 
if llt(lambda2,0) then stable:=1 else stable:=0;
center_sep:=center_manifold(g,taylor_level,stable);
sep1:=sub(x='t,first center_sep);
sep2:=sub(x='t,manifold(transformation(g,mat((0,1),(1,0))),
          taylor_level,stable));
type:=last center_sep;
if member(type,{1,3}) then
  << write "we have a saddle-node";
     write "center separatrice ",{t,sep1};
     write "unstable separatrice ",{sep2,t};
     openfile(result_file);
     write "we have a saddle-node";
     if save_all then
     << write "center separatrice ",{t,sep1};
        write "unstable separatrice ",{sep2,t};
     >>; 
     out 't;
  >>
else if member(type,{2,4}) then
   << write "we have a saddle-node";
      write "center separatrice ",{t,sep1};
      write "stable separatrice ",{sep2,t};
      openfile(result_file);
      write "we have a saddle-node";
      if save_all then
      << write "center separatrice ",{t,sep1};
         write "stable separatrice ",{sep2,t};
      >>;
      out 't;
   >>
else if type=5 then
   <<write "we have a unstable node";
     write "center separatrice ",{t,sep1};
     write "unstable separatrice ",{sep2,t};
     openfile(result_file);
     write "we have a unstable node";
     if save_all then 
     <<write "center separatrice ",{t,sep1};
       write "unstable separatrice ",{sep2,t};
     >>;
     out 't;
   >>
else if type=6 then
   <<write "we have a saddle";
     write "unstable center separatrice ",{t,sep1};
     write "stable separatrice",{sep2,t};
     openfile(result_file);
     write "we have a saddle";
     if save_all then
     <<write "unstable center separatrice ",{t,sep1};
       write "stable separatrice",{sep2,t};
     >>;
     out 't;
    >>
else if type=7 then
   << write "we have a saddle";
      write "stable center separatrice ",{t,sep1};
      write "unstable separatrice ",{sep2,t};
      openfile(result_file);
      write "we have a saddle";
      if save_all then
      <<write "stable center separatrice ",{t,sep1};
        write "unstable separatrice ",{sep2,t};
      >>;
      out 't;
   >>
else << write "we have a stable node";
	write "center separatrice ",{t,sep1};
	write "stable separatrice ",{sep2,t};
        openfile(result_file);
        write "we have a stable node";
        if save_all then
	<<write "center separatrice ",{t,sep1};
	  write "stable separatrice ",{sep2,t};
	>>;
	out 't;
     >>;
write "######################################################################################";
openfile(result_file);
write "######################################################################################";
out 't;
write_semi_elementary(x0,y0,first v1,first v2,second v1,second v2,g,
 append({type},{1}),{sep1,sep2},chart);
end;

procedure type(f,x0,y0,chart);
begin scalar jac,delta,rho,sigma;
 
 jac:=jacobian(f,x0,y0); 
 delta:=first jac*part(jac,4)-second jac*part(jac,3);
 rho:=first jac+part(jac,4); 
 sigma:=rho**2-4*delta;
  
if save_all then 
<<openfile(result_file);
  write "(",x0,",",y0,") is a singular point";
  write "Jacobian=",mat((first jac,second jac),(part(jac,3),part(jac,4)));
  write "Delta=",delta;
  write "Rho=",rho;
  write "Sigma=",sigma;
  out 't;
>>;

 if llt(delta,0) then saddle(f,x0,y0,first jac,second jac,part(jac,3),
                             part(jac,4),rho,sigma,chart)
 else if gt(delta,0) then
  <<if gteq(sigma,0) then node(f,sigma,rho,x0,y0,chart)
    else if eeq(rho,0) then weak_focus(f,x0,y0,chart) 
	 else strong_focus(sigma,rho,x0,y0,chart)
  >>
  else 
   if eeq(rho,0) then degenerate(f,x0,y0,chart)
                else << semi_elementary(f,x0,y0,first jac,second jac,    
                                        part(jac,3),part(jac,4),rho,chart)>>;
end;


