procedure coeff_degree1(f,x);
% input f=a1*x^n1+a2*x^n2+...
% output {{n1,a1},{n2,a2},...}
begin scalar h,l,j,s;
h:={};
l:=coeff(f,x);
for j:=1:length(l) do
 << s:=part(l,j);
    if s neq 0 then h:=append(h,{{j-1,s}})
 >>;
if h={} then h:={{0,0}};
return h;
end;

procedure coeff_degree2(f,x,y);
% input f=a1*x^n1*y^m1+a2*x^n2*y^m2+....
% output {{n1,m1,a1},{n2,m2,a2},...}
begin scalar j,k,l1,l2,s1,s2,h;
h:={};
l1:=coeff(f,x);
for j:=1:length(l1) do
 << s1:=part(l1,j);
    if s1 neq 0 then
      <<l2:=coeff(s1,y);
	for k:=1:length(l2) do
	<< s2:=part(l2,k);
	   if s2 neq 0 then
            h:=append(h,{{j-1,k-1,s2}})
        >>
      >>
 >>;
if h={} then h:={{0,0,0}};
return h;
end;

procedure coeff_degree3(f,x,y,z);
%input f=a1*x^n1*y^m1*z^k1+....
%output {{n1,m1,k1,a1},....}
begin scalar j,k,l,l1,l2,l3,s1,s2,s3,h;
h:={};
l1:=coeff(f,x); 
for j:=1:length(l1) do
<<s1:=part(l1,j);
  if s1 neq 0 then
  <<l2:=coeff(s1,y); 
    for k:=1:length(l2) do
    <<s2:=part(l2,k);
      if s2 neq 0 then
      <<l3:=coeff(s2,z);
        for l:=1:length(l3) do
        <<s3:=part(l3,l);
          if s3 neq 0 then
          h:=append(h,{{j-1,k-1,l-1,s3}})
        >>
      >>
    >>
  >>
>>;
if h={} then h:={{0,0,0,0}};
return h;
end;


procedure write_vec_field(vec_field);
begin scalar l,j,h,r;
r:=symbolic !*rounded;
on rounded;
% x'=....
l:=coeff_degree2(first vec_field,x,y);
write length(l);
for j:=1:length(l) do 
<< h:=part(l,j);
   if last h>=0 then write first h," ",second h," ",last h
    else write first h," ",second h," -",abs(last h)
>>;
% y'=....
l:=coeff_degree2(second vec_field,x,y);
write length(l);
for j:=1:length(l) do
<< h:=part(l,j);
   if last h>=0 then write first h," ",second h," ",last h
   else write first h," ",second h," -",abs(last h)
>>;
if not r then off rounded;
end;

procedure write_vec_field_cylinder(vec_field);
begin scalar l,j,h,ro;
ro:=symbolic !*rounded;
on rounded;
% x'=....
l:=coeff_degree3(first vec_field,r,Co,Si);
write length(l);
for j:=1:length(l) do 
<< h:=part(l,j);
   if last h>=0 then write first h," ",second h," "," ",third h," ",last h
    else write first h," ",second h," ",third h," -",abs(last h)
>>;
% y'=....
l:=coeff_degree3(second vec_field,r,Co,Si);
write length(l);
for j:=1:length(l) do
<< h:=part(l,j);
   if last h>=0 then write first h," ",second h," ",third h," ",last h
   else write first h," ",second h," ",third h," -",abs(last h)
>>;
if not ro then off rounded;
end;


procedure write_saddle(x0,y0,a,b,c,d,vec_field,sep,type,chart);
begin scalar r,j,l,k,g,u;
openfile(log_file);
r:=symbolic !*rounded;
on rounded;
write 1; % (x0,y0) is a saddle
write_number(x0);
write_number(y0);

%WRITE THE TRANS MATRIX
write_number(a);
write_number(b);
write_number(c);
write_number(d);

write_vec_field(vec_field);
%WRITE separatrices
write chart;
 % if length(sep)=1 then we have a saddle at infinity with one sep
for j:=1:length(sep) do
 <<
    write_number(part(type,j));
    l:=coeff_degree1(part(sep,j),'t);
    write length(l);
    for k:=1:length(l) do
     << g:=part(l,k);
        if second g>=0 then write first g," ",second g
        else write first g," -",abs(second g);
     >>;
  >>;
write epsilon;
if not r then off rounded;
out 't;
end;

procedure write_node(x0,y0,l,chart);
begin  scalar r;
openfile(log_file);
r:=symbolic !*rounded;
on rounded;
write 2;% {x0,y0} is a node
write_number(x0);
write_number(y0);
write_number(l);% (1) unstable (-1) stable
write chart;
if not r then off rounded;
out 't;
end;

procedure write_weak_focus(x0,y0,chart,type);
begin
openfile(log_file);
r:=symbolic !*rounded;
on rounded;
write 3; % {x0,y0} is a weak focus
write_number(x0);
write_number(y0);
write type;
write chart;
if not r then off rounded;
out 't;
end;

procedure write_strong_focus(x0,y0,l,chart);
begin  scalar r;
openfile(log_file);
r:=symbolic !*rounded;
on rounded;
write 4; % {x0,y0} is a strong focus
write l; % (1) unstable (-1) stable
write_number(x0);
write_number(y0);
write chart;
if not r then off rounded;
out 't;
end;

procedure write_semi_elementary(x0,y0,a,b,c,d,vec_field,type,sep,chart);
begin scalar r,l,k,g;
openfile(log_file);
r:=symbolic !*rounded;
on rounded;
write 5;
write_number(x0);
write_number(y0);

%WRITE THE TRANS MATRIX
write_number(a);
write_number(b);
write_number(c);
write_number(d);

write_vec_field(vec_field);
write first type;
write second type; % =0 means that line at infinity is center separatrices
write chart;
if ((chart=0) and member(first type,{1,2,3,4,6,7})) or
   ((chart neq 0) and (second type=0) and member(first type,{1,2,3,4,6,7})) or
   ((chart neq 0) and (second type=1) and member(first type,{2,3,6,7}))
then
<<
  %WRITE separatrices
  for j:=1:length(sep) do
  <<l:=coeff_degree1(part(sep,j),'t);
    write length(l);
    for k:=1:length(l) do
    << g:=part(l,k);
       if second g>=0 then write first g," ",second g    
       else write first g," -",abs(second g);
    >>
  >>
>>;
write epsilon;
if not r then off rounded;
out 't;
end;

procedure write_separatrice(sep);
begin scalar l,j,trans,a,k;
trans:=first sep;
write length(trans);
for l:=length(trans) step -1 until 1 do
 << a:=part(trans,l);
    for j:=1:9 do write_number(part(a,j));
 >>;
write_number(first second sep);
write_number(second second sep);
trans:=part(sep,3);
write_number(first trans);
write_number(second trans);
write_number(third trans);
write_number(last trans);
write_vec_field(part(sep,4));
l:=coeff_degree1(part(sep,5),'t);
write length(l);
for k:=1:length(l) do
 << g:=part(l,k);
    if second g>=0 then write first g," ",second g    
    else write first g," -",abs(second g);
 >>;
write part(sep,6);
end;


    
    
procedure write_degenerate(x0,y0,trajectors,chart);
begin scalar r,l;
openfile(log_file);
r:=symbolic !*rounded;
on rounded;
write 6;
write_number(x0);
write_number(y0);
write epsilon;
write length(trajectors);
for l:=1:length(trajectors) do
write_separatrice(part(trajectors,l));
write chart;
if not r then off rounded;
out 't;
end;

% write gcf to file
procedure write_gcf(gcff);
begin scalar r,l,g,d;

r:=symbolic !*rounded;
on rounded;
openfile(vec_table);
d:=ddeg(gcff,x,y);
if d=0 then write 0  % no comman factors 
 else
 <<write d;
   l:=coeff_degree2(gcff,x,y);
   write length(l);
   for j:=1:length(l) do
    << h:=part(l,j);
       if last h>=0 then write first h," ",second h," ",last h
       else write first h," ",second h," -",abs(last h)
    >>;

  g:=num(sub(x=1/z2^p,y=z1/z2^q,gcff)); 
  l:=coeff_degree2(g,z1,z2);
  write length(l);
  for j:=1:length(l) do
   << h:=part(l,j);
      if last h>=0 then write first h," ",second h," ",last h
      else write first h," ",second h," -",abs(last h)
   >>;    
  g:=num(sub(x=z1/z2^p,y=1/z2^q,gcff)); 
  l:=coeff_degree2(g,z1,z2);
  write length(l);
  for j:=1:length(l) do
   << h:=part(l,j);
      if last h>=0 then write first h," ",second h," ",last h
      else write first h," ",second h," -",abs(last h)
   >>;   
  g:=num(sub(x=-1/z2^p,y=z1/z2^q,gcff)); 
  l:=coeff_degree2(g,z1,z2);
  write length(l);
  for j:=1:length(l) do
   << h:=part(l,j);
      if last h>=0 then write first h," ",second h," ",last h
      else write first h," ",second h," -",abs(last h)
   >>;  
  g:=num(sub(x=z1/z2^p,y=-1/z2^q,gcff)); 
  l:=coeff_degree2(g,z1,z2);
  write length(l);
  for j:=1:length(l) do
   << h:=part(l,j);
      if last h>=0 then write first h," ",second h," ",last h
      else write first h," ",second h," -",abs(last h)
   >>;
 if ((p neq 1) or (q neq 1)) then
 % we work on the poincare-lyapunov disc
 <<
  clear Co,Si,s;
  g:=num(sub(x=Co/s^p,y=Si/s^q,gcff)); 
  l:=coeff_degree3(g,s,Co,Si);
  write length(l);
  for j:=1:length(l) do 
  << h:=part(l,j);
     if last h>=0 then write first h," ",second h," "," ",third h," ",last h
     else write first h," ",second h," ",third h," -",abs(last h)
  >>;
 >>;        
>>;
 
out 't;   
if not r then off rounded;
end;

