#include <stdlib.h>
#include <stdio.h>
#include "lyapunov.h"

// --------------------------------------------------------------------------
//                      INS_HOM_POLY
// --------------------------------------------------------------------------

void ins_hom_poly( struct hom_poly *f, int degz, int degzb, double re, double im )
{
	int total_degree = degz + degzb, ok = 1;
	struct hom_poly *w = f->next_hom_poly;

	while ( w != NULL && ok )
	{
		if ( total_degree == w->total_degree )
		{
			ok = 0;
			ins_poly( w->p, degz, degzb, re, im );
			break;
		}
		if ( total_degree > w->total_degree )
			break;
		f = w;
		w = w->next_hom_poly;
	}
	if ( ok )
	{
		f->next_hom_poly = ( struct hom_poly * ) malloc( sizeof( struct hom_poly ) );
		f = f->next_hom_poly;
		f->total_degree = total_degree;
		f->p = ( struct poly * ) malloc( sizeof( struct poly ) );
		f->p->next_poly = NULL;
		f->next_hom_poly = w;
		ins_poly( f->p, degz, degzb, re, im );
	}
}

// --------------------------------------------------------------------------
//                      INS_POLY
// --------------------------------------------------------------------------

void ins_poly( struct poly *f, int degz, int degzb, double re, double im )
{
	struct poly *w = f->next_poly;

	while ( w != NULL )
	{
		if ( degz > w->degz )
			break;
		if ( degz == w->degz && degzb > w->degzb )
			break;
		f = w;
		w = w->next_poly;
	}
	f->next_poly = ( struct poly * ) malloc( sizeof( struct poly ) );
	f = f->next_poly;
	f->degz = degz;
	f->degzb = degzb;
	f->re = re;
	f->im = im;
	f->next_poly = w;
}

// --------------------------------------------------------------------------
//                      COPY_POLY
// --------------------------------------------------------------------------

struct poly *copy_poly( struct poly *f )
{
	struct poly *g, *h;

	g = ( struct poly * ) malloc( sizeof( struct poly ) );
	g->next_poly = NULL;
	h = g;
	while ( ( f = f->next_poly ) != NULL )
	{
		g->next_poly = ( struct poly * ) malloc( sizeof( struct poly ) );
		g = g->next_poly;
		g->degz = f->degz;
		g->degzb = f->degzb;
		g->re = f->re;
		g->im = f->im;
		g->next_poly = NULL;
	}
	return ( h );
}

// --------------------------------------------------------------------------
//                      SEARCH_POLY
// --------------------------------------------------------------------------

struct poly *search_poly( struct poly *f, int degz, int degzb, int *present )
{
	struct poly *w;

	w = f;
	*present = 0;
	while ( ( f = f->next_poly ) != NULL )
	{
		if ( degz > f->degz )
			break;
		if ( degz == f->degz && degzb == f->degzb )
		{
			*present = 1;
			break;
		}
		if ( degz == f->degz && degzb > f->degzb )
			break;
		w = f;
	}
	return ( w );
}

// --------------------------------------------------------------------------
//                      INS_COEFF
// --------------------------------------------------------------------------

void ins_coeff( struct poly *f, int degz, int degzb, double re, double im )
{

	struct poly *w;

	w = ( struct poly * ) malloc( sizeof( struct poly ) );
	w->degz = degz;
	w->degzb = degzb;
	w->re = re;
	w->im = im;
	w->next_poly = f->next_poly;
	f->next_poly = w;
}

// --------------------------------------------------------------------------
//                      DELETE_COEFF
// --------------------------------------------------------------------------

void delete_coeff( struct poly *f )
{
	struct poly *w;

	w = f->next_poly->next_poly;
	free( f->next_poly );
	f->next_poly = w;
}

// --------------------------------------------------------------------------
//                      ADD_POLY
// --------------------------------------------------------------------------

void add_poly( struct poly *f, struct poly *g )
{
	int present;
	struct poly *w1, *w2;

	while ( ( g = g->next_poly ) != NULL )
	{
		w1 = search_poly( f, g->degz, g->degzb, &present );
		if ( !present )
			ins_coeff( w1, g->degz, g->degzb, g->re, g->im );
		else
		{
			w2 = w1->next_poly;
			w2->re += g->re;
			w2->im += g->im;
			if ( !( w2->re || w2->im ) )
			{
				w1->next_poly = w2->next_poly;
				free( w2 );
			}
		}
	}
}

// --------------------------------------------------------------------------
//                      SUB_POLY
// --------------------------------------------------------------------------

void sub_poly( struct poly *f, struct poly *g )
{
	int present;
	struct poly *w1, *w2;

	while ( ( g = g->next_poly ) != NULL )
	{
		w1 = search_poly( f, g->degz, g->degzb, &present );
		if ( !present )
			ins_coeff( w1, g->degz, g->degzb, -g->re, -g->im );
		else
		{
			w2 = w1->next_poly;
			w2->re -= g->re;
			w2->im -= g->im;
			if ( !( w2->re || w2->im ) )
			{
				w1->next_poly = w2->next_poly;
				free( w2 );
			}
		}
	}
}

// --------------------------------------------------------------------------
//                      TERM_PROD
// --------------------------------------------------------------------------

void term_prod( struct poly *f, int degz, int degzb, double re, double im )
{
	double r, i;

	while ( ( f = f->next_poly ) != NULL )

	{
		f->degz += degz;
		f->degzb += degzb;
		r = f->re * re - f->im * im;
		i = f->re * im + f->im * re;
		f->re = r;
		f->im = i;
	}
}

// --------------------------------------------------------------------------
//                      PROD_POLY
// --------------------------------------------------------------------------

void prod_poly( struct poly *f, struct poly *g )
{
	struct poly *w1, *w2;

	if ( ( f->next_poly != NULL ) && ( g->next_poly != NULL ) )
	{
		w2 = copy_poly( f );
		g = g->next_poly;
		term_prod( f, g->degz, g->degzb, g->re, g->im );
		while ( ( g = g->next_poly ) != NULL )
		{
			w1 = copy_poly( w2 );
			term_prod( w1, g->degz, g->degzb, g->re, g->im );
			add_poly( f, w1 );
			delete_poly( w1 );
		}
		delete_poly( w2 );
	}
	else
	{
		delete_poly( f );
		/* we must return the zero poly and not a empty list !! */
		f = ( struct poly * ) malloc( sizeof( struct poly ) );
		f->next_poly = NULL;
	}
}

// --------------------------------------------------------------------------
//                      MULTC_POLY
// --------------------------------------------------------------------------

void multc_poly( struct poly *f, double re, double im )
{
	double r, i;

	while ( ( f = f->next_poly ) != NULL )
	{
		r = f->re * re - f->im * im;
		i = f->re * im + f->im * re;
		f->re = r;
		f->im = i;
	}
}

// --------------------------------------------------------------------------
//                      DELETE_POLY
// --------------------------------------------------------------------------

void delete_poly( struct poly *f )
{
	struct poly *w;

	if ( f != NULL )
		do
		{
			w = f;
			f = f->next_poly;
			free( w );
		}
		while ( f != NULL );
}

// --------------------------------------------------------------------------
//                      CONJ_POLY
// --------------------------------------------------------------------------

struct poly *conj_poly( struct poly * f )
{
	struct poly *g;

	g = ( struct poly * ) malloc( sizeof( struct poly ) );
	g->next_poly = NULL;
	while ( ( f = f->next_poly ) != NULL )
		ins_poly( g, f->degzb, f->degz, f->re, -f->im );
	return ( g );
}

// --------------------------------------------------------------------------
//                      PRINT_POLY
// --------------------------------------------------------------------------

void print_poly( struct poly * f )
{
	if( f )
	while( ( f = f->next_poly ) )
		printf( "%f  %f  %d  %d\n", f->re, f->im, f->degz, f->degzb );
	printf( "\n\n" );
}

// --------------------------------------------------------------------------
//                      READ_POLY
// --------------------------------------------------------------------------

void read_poly( struct poly *f )
{
	int degz, degzb;
	double re, im;

	for ( ;; )
	{
		scanf( "%lf", &re );
		scanf( "%lf", &im );
		scanf( "%d", &degz );
		scanf( "%d", &degzb );
		if ( degz < 0 )
			break;
		ins_poly( f, degz, degzb, re, im );
	}
}

/*
void main()
{
 
  struct poly *f,*g,*h;
  
  f=(struct poly *) malloc(sizeof(struct poly));
  f->next_poly=NULL;
  printf("read f\n");
  read_poly(f);
  print_poly(f); 
  g=conj_poly(f);
  print_poly(g); 
  delete_poly(g);
  g=(struct poly *) malloc(sizeof(struct poly));
  printf("read g\n");
  read_poly(g);
  print_poly(g);
  h=copy_poly(g);
  print_poly(h);
  add_poly(h,f);
  print_poly(h);
  h=copy_poly(g);
  sub_poly(h,f);
  print_poly(g);
  h=copy_poly(g);
  prod_poly(h,f);
  print_poly(h);
}
 
*/
