#include <qprocess.h>

#ifndef MAXNUMPARAMS
#include "custom.h"
#endif

class QTextEdit;
class QGcfDlg;
class QPushButton;
class QTextStream;

#define TYPEOFSTUDY_ALL     0
#define TYPEOFSTUDY_FIN     1
#define TYPEOFSTUDY_INF     2
#define TYPEOFSTUDY_ONE     3

#define PACKAGE_MAPLE       1
#define PACKAGE_REDUCE      0

typedef struct _VFRegion
{
    int vfindex;        // which vector field
    int * signs;
} VFREGION;

typedef struct _CurveRegion
{
    int  curveindex;        // which curve are we talking about
    int * signs;
} CURVEREGION;

class QInputVF : public QObject
{
Q_OBJECT
public:
    QString filename;

    int symbolicpackage;        // 0 for reduce, 1 for maple
    int typeofstudy;            // 0, 1, 2, 3 = all, inf, fin, one
    
    QString x0;
    QString y0;
    
//    int defaulttypeofview;
//    double defaultxmin;
//    double defaultxmax;
//    double defaultymin;
//    double defaultymax;
    
    int p;
    int q;
    int numparams;
    QString parlabel[MAXNUMPARAMS];
    int numvf;
    int numcurves;
    QString * * curve;
    int * numpointscurve;

    int numselected;
    int * selected;

    int numvfregions;
    int numcurveregions;
    VFREGION * vfregions;
    CURVEREGION * curveregions;
    
    QString * * xdot;
    QString * * ydot;
    QString * * gcf;

    bool * numeric;
    int * precision;
    QString * * epsilon;
    bool * testsep;
    int * taylorlevel;
    int * numericlevel;
    int * maxlevel;
    int * weakness;
    
    QString *** parvalue;

    bool changed;               // set when data needs to be saved
    bool evaluated;             // set when data has been evaluated
    bool evaluating;            // set while evaluating
    bool evaluating_piecewiseconfig;
    bool evaluating_gcf;         // true when evaluation is of GCF kind
    bool cleared;               // initial state, when records are clear
    bool processfailed;         // true when process failed;
    QString processError;       // only relevant when processfailed=true
    QGcfDlg * gcfDlg;
    QString EvalFile;
    QString EvalFile2;
    QProcess * EvalProcess;
    QTextEdit * ProcessText;
    QPushButton * ProcessButton;
    QPushButton * ProcessClearButton;

    void AddVectorField( void );
    void DeleteVectorField( int );

    void AddSeparatingCurve( void );
    void DeleteSeparatingCurve( int );

    void ResampleCurve( int );
    int GetVFIndex_R2( const double * );
    int GetVFIndex_U1( const double * );
    int GetVFIndex_U2( const double * );
    int GetVFIndex_V1( const double * );
    int GetVFIndex_V2( const double * );
    int GetVFIndex_UU1( const double * );
    int GetVFIndex_UU2( const double * );
    int GetVFIndex_VV1( const double * );
    int GetVFIndex_VV2( const double * );
    int GetVFIndex_sphere( const double * );
    int GetVFIndex_psphere( const double * );
    int GetVFIndex_plsphere( const double * );
    int GetVFIndex_cyl( const double * );

    void MarkVFRegion( int, const double * );
    void UnmarkVFRegion( int, const double * );
    void MarkCurveRegion( int, const double * );
    void UnmarkCurveRegion( int, const double * );
    void ClearVfMarks( void );
    void ClearCurveMarks( void );
    bool IsCurvePointDrawn( int, const double * );
    void ResampleGcf( int );

public:
    bool HasCommonString( QString ** );
    bool HasCommonInt( int * );
    bool HasCommonBool( bool * );
    bool HasCommonParvalue( int );
    QString CommonString( QString ** );
    int CommonInt( int * );
    bool CommonBoolean( bool * );
    QString CommonParvalue( int );
    void SetCommonString( QString **, QString );
    void SetCommonInt( int *, int );
    void SetCommonBool( bool *, bool );
    void SetCommonParvalue( int, QString );

public:
    QInputVF();
    ~QInputVF();
    
    QString getfilename( void ) const;              // filename.inp
    QString getbarefilename( void ) const;          // filename
    QString getfilename_finresults( void ) const;   // filename_fin.res
    QString getfilename_infresults( void ) const;   // filename_inf.res
    QString getfilename_curveresults( void ) const; // filename_curves.res
    QString getfilename_fintable( void ) const;     // filename_fin.tab
    QString getfilename_inftable( void ) const;     // filename_inf.tab
    QString getfilename_vectable( void ) const;     // filename_vec.tab
    QString getfilename_gcf( void ) const;          // filename_gcf.tab (temporary file)
    QString getfilename_gcfresults( void ) const;   // filename_gcf.res (temporary file, only used in case of reduce)
    QString getreducefilename( void ) const;        // filename.red
    QString getmaplefilename( void ) const;         // filename.mpl
    QString getrunfilename( void ) const;           // filename.run

    bool load( void );
    bool save( void );
    void reset( int );
    bool checkevaluated( void );

    static bool FileExists( QString );

    void PrepareReduceParameters( QTextStream * );
    void PrepareReduceVectorField( QTextStream * );
    void PrepareMapleParameters( QTextStream * );
    void PrepareMapleVectorField( QTextStream * );
    void PrepareMaplePiecewiseConfig( QTextStream * );
    void PrepareReducePiecewiseConfig( QTextStream * );
    void PrepareReduceCurves( QTextStream * );
    void PrepareMapleCurves( QTextStream * );

    QString booleanString( int value ) const;
    QString ConvertMapleUserParameterLabels( QString );
    QString ConvertReduceUserParameterLabels( QString );

    void PrepareFile( QTextStream *, bool ); // used by Prepare()

    void Prepare( void );
    void Evaluate( void );

    bool PrepareGcf( struct term2 * f, double, double, int, int, int, int );
    bool PrepareGcf_LyapunovCyl( double, double, int, int, int, int );
    bool PrepareGcf_LyapunovR2( int, int, int, int );
    bool EvaluateGcf( int, int, int, int );
    void CreateProcessWindow( void );
    void EvaluateCurves( void );
    void PrepareCurves( void );

public slots:
    void finishEvaluation( int );
    void catchProcessError( QProcess::ProcessError );
    void ReadProcessStdout( void );
    void OnTerminateButton( void );
    void OnClearButton( void );
    void finishCurvesEvaluation( void );
    void finishGcfEvaluation( void );
};

extern QInputVF * ThisVF;

extern bool IsInsideRegion_R2( const int *, const double * );
extern bool IsInsideRegion_U1( const int *, const double * );
extern bool IsInsideRegion_U2( const int *, const double * );
extern bool IsInsideRegion_V1( const int *, const double * );
extern bool IsInsideRegion_V2( const int *, const double * );
extern bool IsInsideRegion_cyl( const int *, const double * );
extern bool IsInsideRegion_sphere( const int *, const double * );

