#include <qfile.h>
#include <math.h>
#include "table.h"
#include "custom.h"
#include "file_vf.h"
#include "win_p5.h"
#include "math_p4.h"
#include "win_inputsphere.h"

extern void polarcoord_to_plsphere( double, double, double * );

void QInputVF::MarkVFRegion( int index, const double * p )
{
    int i,k;

    if( VFResults.curves == NULL )
        return;

    for( i = numvfregions-1; i >= 0; i-- )
    {
        if( IsInsideRegion_sphere( vfregions[i].signs, p ) )
            break;
    }
    if( i < 0 )
    {
        vfregions = (VFREGION *)realloc( vfregions, sizeof(VFREGION) * (numvfregions+1) );
        if( numcurves > 0 )
            vfregions[numvfregions].signs = (int *)malloc( sizeof(int) * numcurves );
        else
            vfregions[numvfregions].signs = NULL;
        for( k = numcurves-1; k >= 0; k-- )
        {
            if( eval_curve( VFResults.curves+k, p ) < 0 )
                vfregions[numvfregions].signs[k] = -1;
            else
                vfregions[numvfregions].signs[k] = +1;
        }
        vfregions[numvfregions].vfindex = -1;
        numvfregions ++;
        i = numvfregions - 1;
    }
    if( vfregions[i].vfindex == index )
        return;

    vfregions[i].vfindex = index;
}

void QInputVF::UnmarkVFRegion( int index, const double * p )
{
    int i;

    if( VFResults.curves == NULL )
        return;

    for( i = numvfregions-1; i >= 0; i-- )
    {
        if( IsInsideRegion_sphere( vfregions[i].signs, p ) )
            break;
    }
    if( i < 0 )
        return; // region cannot be found???

    if( vfregions[i].vfindex != index )
        return; // region does not have the corresponding index

    free( vfregions[i].signs );
    if( i < numvfregions-1 )
        memmove( vfregions+i, vfregions+i+1, sizeof(VFREGION) * (numvfregions-i-1) );
    if( numvfregions == 1)
    {
        free( vfregions );
        vfregions = NULL;
    }
    numvfregions--;
}

void QInputVF::MarkCurveRegion( int index, const double * p )
{
    int i;
    int * signs;
    int k;

    if( numcurves == 0 )
        return;
    if( VFResults.curves == NULL )
        return;

    signs = (int *)malloc( sizeof(int) * numcurves );
    for( i = numcurves-1; i >= 0; i-- )
    {
        if( i == index )
            signs[i] = 0;
        else
            if( eval_curve( VFResults.curves+i, p ) < 0 )
                signs[i] = -1;
            else
                signs[i] = 1;
    }
    for( k = 0; k < numcurveregions; k++ )
    {
        if( !memcmp( signs, curveregions[k].signs, sizeof(int)*numcurves ) && curveregions[k].curveindex == index )
        {
            // curve mark already exists
            free(signs);
            return;
        }
    }

    curveregions = (CURVEREGION *)realloc( curveregions, sizeof(CURVEREGION) * (numcurveregions+1) );
    curveregions[numcurveregions].curveindex = index;
    curveregions[numcurveregions].signs = signs;
    numcurveregions++;
    ResampleCurve(index);
}

void QInputVF::UnmarkCurveRegion( int index, const double * p )
{
    int i;
    int * signs;
    int k;

    if( numcurves == 0 )
        return;
    if( VFResults.curves == NULL )
        return;

    signs = (int *)malloc( sizeof(int) * numcurves );
    for( i = numcurves-1; i >= 0; i-- )
    {
        if( i == index )
            signs[i] = 0;
        else
            if( eval_curve( VFResults.curves+i, p ) < 0 )
                signs[i] = -1;
            else
                signs[i] = 1;
    }
    for( k = numcurveregions-1; k >= 0; k-- )
    {
        if( !memcmp( signs, curveregions[k].signs, sizeof(int)*numcurves ) && curveregions[k].curveindex == index )
        {
            // curve mark exists
            break;
        }
    }
    free( signs );
    if( k >= 0 )
    {
        free( curveregions[k].signs );
        if( k < numcurveregions-1 )
            memmove( curveregions+k, curveregions+k+1, sizeof(CURVEREGION) * (numcurveregions-k-1) );
        numcurveregions--;
        ResampleCurve(index);
    }
}

void QInputVF::ClearVfMarks( void )
{
    int i;
    if( vfregions != NULL )
    {
        if( numcurves > 0 )
            for( i = 0; i < numvfregions; i++ )
                free( vfregions[i].signs );

        free( vfregions );
        vfregions = NULL;
        numvfregions = 0;
    }
    return;
}

void QInputVF::ClearCurveMarks( void )
{
    int i;
    if( curveregions != NULL )
    {
        if( numcurves > 0 )
            for( i = 0; i < numcurveregions; i++ )
                free( curveregions[i].signs );

        free( curveregions );
        curveregions = NULL;
        numcurveregions = 0;
    }
    return;
}

bool QInputVF::IsCurvePointDrawn( int index, const double * pcoord )
{
    int k, j;

    if( numcurves == 0 )
        return true;
    if( VFResults.curves == NULL )
        return false;

    for( j = numcurveregions-1; j >= 0; j-- )
    {
        if( curveregions[j].curveindex != index )
            continue;
        for( k = numcurves-1; k >= 0; k-- )
        {
            if( k == index )
                continue;

            if( eval_curve( VFResults.curves+k, pcoord ) < 0 )
            {
                if( curveregions[j].signs[k] > 0 )
                    break;
            }
            else
            {
                if( curveregions[j].signs[k] < 0 )
                    break;
            }
        }
        if( k < 0 )
        {
            // matching region found --> disable curve draw.

            return false;
        }
    }
    return true;
}

void QInputVF::ResampleCurve( int i )
{
    struct orbits_points * sep;

    if( VFResults.curves == NULL )
        return;

    sep = VFResults.curves[i].sep_points;
    while( sep != NULL )
    {
        if( IsCurvePointDrawn( i, sep->pcoord ) )
            sep->color = CSEPCURVE;
        else
            sep->color = CSHADEDCURVE;

        sep = sep->nextpt;
    }
}

void QInputVF::ResampleGcf( int i )
{
    struct orbits_points * * sep;
    struct orbits_points * sepx;

    if( VFResults.curves == NULL )
        return;
    if( VFResults.vf == NULL )
        return;

    sep = &(VFResults.vf[i]->gcf_points);

    while( *sep != NULL )
    {
        if( GetVFIndex_sphere( (*sep)->pcoord ) != i )
        {
            sepx = *sep;
            *sep = sepx->nextpt;
            free( sepx );
            if( *sep != NULL )
                (*sep)->dashes = 0;
        }
        else
            sep = &((*sep)->nextpt);
    }
}

static void insert_curve_point( double x0, double y0, double z0, int dashes, struct orbits_points * * lastpt )
{
    (*lastpt) = (struct orbits_points *) malloc(sizeof(struct orbits_points));
    (*lastpt)->nextpt = NULL;

    (*lastpt)->pcoord[0] = x0;
    (*lastpt)->pcoord[1] = y0;
    (*lastpt)->pcoord[2] = z0;

    (*lastpt)->dashes = dashes;
    (*lastpt)->color = CSEPCURVE;
    (*lastpt)->nextpt = NULL;
}

bool ReadCurvePoints( FILE * fp, struct orbits_points * * psep, int index )
{
    int k;
    int numpoints;
    double x, y;
    double pcoord[3];
    int d,c;
    int chartindex;
    void (*chart)(double,double,double *);
    static void (*charts[12])(double,double,double *) =
    {
        R2_to_psphere,
        U1_to_psphere,
        U2_to_psphere,
        VV1_to_psphere,
        VV2_to_psphere,
        NULL,
        polarcoord_to_plsphere,
        cylinder_to_plsphere,
        cylinder_to_plsphere,
        cylinder_to_plsphere,
        cylinder_to_plsphere,
        NULL
    };

    while( *psep != NULL )
        psep = &((*psep)->nextpt);

    chartindex= (VFResults.plweights)  ? 6 : 0;
    while( 1 )
    {
        chart = charts[chartindex];
        if( chart == NULL )
            break;
        chartindex++;

        fscanf( fp, "%d\n", &numpoints );

        if( numpoints < 0 )
        {
            return false;
        }

        if( numpoints == 0 )
            continue;

        k=0;
        while( k < numpoints )
        {
            d=0;
            while( k < numpoints )
            {
                if( fscanf( fp, "%lf %lf", &x, &y ) != 2 )
                    break;
                k++;
                (*chart)(x,y,pcoord);
                insert_curve_point( pcoord[0], pcoord[1], pcoord[2], d, psep);
                psep = &((*psep)->nextpt);
                d=1;
                /*P5 d=GcfDashes; */
            }
            if( k < numpoints )
            {
                for( c = getc(fp); isspace(c); )
                    c =getc(fp);
                if( c != ',' )
                    return false;
            }
        }
    }
    return true;
}
