#include <qobject.h>
#include <math.h>
#include "custom.h"
#include "table.h"
#include "math_p4.h"
#include "file_vf.h"

extern int FindSepColor(struct term2 * f, int type,double * y );
extern int change_type(int);
extern bool PrepareVfForIntegration( double * );
extern bool IsARealSingularity( double, double, int, int );

// ---------------------------------------------------------------------------
//                  POWER
// ---------------------------------------------------------------------------

static struct term2 * vec_field[2] = { NULL, NULL };

void eval_blow_vec_field( const double * y, double * f)
{
    f[0] = eval_term2(vec_field[0],y);
    f[1] = eval_term2(vec_field[1],y);
}

extern struct orbits_points * integrate_sep( QWinSphere *, double *, double, int, int, int, struct orbits_points ** );

// ---------------------------------------------------------------------------
//                  POWER
// ---------------------------------------------------------------------------
//
// Calculates a^b with the convention that a^0=1  (hence also 0^0=1)

static double power(double a,int b)
{
    if( b == 0 )
        return 1.0;
    else
        return pow( a, (double)b );
}

// ---------------------------------------------------------------------------
//                  MAKETRANSFORMATIONS
// ---------------------------------------------------------------------------

void make_transformations(struct transformations * trans, double x0, double y0, double * point)
{
    double x,y;
    
    x = x0;
    y = y0;
    while( trans != NULL )
    {
        point[0] = trans->x0 + ((double)trans->c1) * power(x, trans->d1) * power(y, trans->d2);
        point[1] = trans->y0 + ((double)trans->c2) * power(x, trans->d3) * power(y, trans->d4);
        x = point[0];
        y = point[1];
        trans = trans->next_trans;
    }
}

// ---------------------------------------------------------------------------
//                  INTEGRATE_BLOW_UP
// ---------------------------------------------------------------------------

struct orbits_points * integrate_blow_up( QWinSphere * spherewnd, double * pcoord2,
                    struct blow_up_points * de_sep, double step,int dir, int type,
                    struct orbits_points ** orbit, int chart )
{
    int i;
    double hhi,hhi0;
    double y0[2];
    double h_min;
    double h_max;
    double point[2];
    double pcoord[3];
    double y[2];
    int color;
    int dashes;
    int ok;
    struct orbits_points *first_orbit;
    struct orbits_points *last_orbit;

    h_min=VFResults.config_hmi;
    h_max=VFResults.config_hma;

    first_orbit = NULL;
    last_orbit = NULL;
    ok = true;
    vec_field[0] = de_sep->vector_field[0];
    vec_field[1] = de_sep->vector_field[1];
    
    y[0] = de_sep->point[0];
    y[1] = de_sep->point[1];
    make_transformations( de_sep->trans,
                            de_sep->x0 + de_sep->a11*y[0] + de_sep->a12*y[1],
                            de_sep->y0 + de_sep->a21*y[0] + de_sep->a22*y[1], point );
    switch( chart )
    {
    case CHART_R2:      MATHFUNC(R2_to_sphere)(point[0],point[1],pcoord);   break;
    case CHART_U1:      MATHFUNC(U1_to_sphere)(point[0],point[1],pcoord);   break;
    case CHART_V1:      MATHFUNC(V1_to_sphere)(point[0],point[1],pcoord);   break;
    case CHART_U2:      MATHFUNC(U2_to_sphere)(point[0],point[1],pcoord);   break;
    case CHART_V2:      MATHFUNC(V2_to_sphere)(point[0],point[1],pcoord);   break;
    }
    
    if( !PrepareVfForIntegration( pcoord ) )
        return NULL;

    if( VFResults.plweights == false && (chart==CHART_V1 || chart==CHART_V2) )
        dir = VFResultsK->dir_vec_field * dir;

    hhi = (double)dir*step;
    y[0] = de_sep->point[0];
    y[1] = de_sep->point[1];

    for( i = 1; i <= VFResults.config_intpoints; ++i )
    {
        hhi0 = hhi;
        y0[0]=y[0];
        y0[1]=y[1];
        while( 1 )
        {
            y[0]=y0[0];
            y[1]=y0[1];
            rk78( eval_blow_vec_field, y, &hhi0, VFResults.config_hmi, VFResults.config_hma, VFResults.config_tolerance );
            make_transformations( de_sep->trans,
                                de_sep->x0 + de_sep->a11*y[0] + de_sep->a12*y[1],
                                de_sep->y0 + de_sep->a21*y[0] + de_sep->a22*y[1], point );
            switch( chart )
            {
            case CHART_R2: MATHFUNC(R2_to_sphere)(point[0],point[1],pcoord); break;
            case CHART_U1: if(point[1]>=0 || !VFResultsK->singinf) MATHFUNC(U1_to_sphere)(point[0],point[1],pcoord);
                                else VV1_to_psphere(point[0],point[1],pcoord); break;
            case CHART_V1: MATHFUNC(V1_to_sphere)(point[0],point[1],pcoord); break;
            case CHART_U2: if(point[1]>=0 || !VFResultsK->singinf) MATHFUNC(U2_to_sphere)(point[0],point[1],pcoord);
                                else VV2_to_psphere(point[0],point[1],pcoord); break;
            case CHART_V2: MATHFUNC(V2_to_sphere)(point[0],point[1],pcoord); break;
            }

            if( ThisVF->GetVFIndex_sphere( pcoord ) == VFResults.K )
                break;
            h_min = VFResults.config_branchhmi;
            h_max /= 2;
            hhi0 = fabs(hhi0*(hhi))/2/(hhi);
            if( fabs(hhi0) < h_min || h_max < VFResults.config_branchhmi )
            {
                hhi0 = h_min;
                if( hhi<0 ) hhi0=-hhi0;
                de_sep->integrating_in_local_chart=false;
                break;
            }
        }
        hhi = hhi0;
        
        dashes = true;
        switch( chart )
        {
        case CHART_R2:
            MATHFUNC(R2_to_sphere)(point[0],point[1],pcoord);
            color = FindSepColor(VFResultsK->gcf,type,point);
            break;

        case CHART_U1:
            if(point[1]>=0 || !VFResultsK->singinf)
            {
                MATHFUNC(U1_to_sphere)(point[0],point[1],pcoord);
                if(!ok)
                {
                    dashes=false; ok=true;
                    if(VFResultsK->dir_vec_field==1)dir*=-1;
                }
                type=de_sep->type;
                color=FindSepColor(VFResultsK->gcf_U1,type,point);
            }
            else
            {
                VV1_to_psphere(point[0],point[1],pcoord);
                if(ok)
                {
                    dashes=false; ok=false;
                    if(VFResultsK->dir_vec_field==1)dir*=-1;
                }
                if(VFResultsK->dir_vec_field==1)
                    type=change_type(de_sep->type);
                else
                    type=de_sep->type;

                psphere_to_V1(pcoord[0],pcoord[1],pcoord[2],point);
                color=FindSepColor(VFResultsK->gcf_V1,type,point);
            }
            break;

        case CHART_V1:
            MATHFUNC(V1_to_sphere)(point[0],point[1],pcoord);
            if( !VFResults.plweights )
                psphere_to_V1(pcoord[0],pcoord[1],pcoord[2],point);
            color=FindSepColor(VFResultsK->gcf_V1,type,point);
            break;

        case CHART_U2:
            if(point[1]>=0 || !VFResultsK->singinf)
            {
                MATHFUNC(U2_to_sphere)(point[0],point[1],pcoord);
                if(!ok)
                {
                    dashes=false; ok=true;
                    if(VFResultsK->dir_vec_field==1)dir*=-1;
                }
                type=de_sep->type;
                color=FindSepColor(VFResultsK->gcf_U2,type,point);
            }
            else
            {
                VV2_to_psphere(point[0],point[1],pcoord);
                if(ok){dashes=false; ok=false;
                if(VFResultsK->dir_vec_field==1)dir*=-1;}
                if(VFResultsK->dir_vec_field==1)
                    type=change_type(de_sep->type);
                else
                    type=de_sep->type;
                psphere_to_V2(pcoord[0],pcoord[1],pcoord[2],point);
                color=FindSepColor(VFResultsK->gcf_V2,type,point);
            }
            break;

        case CHART_V2:
            MATHFUNC(V2_to_sphere)(point[0],point[1],pcoord);
            if( !VFResults.plweights )
                psphere_to_V2(pcoord[0],pcoord[1],pcoord[2],point);
            color=FindSepColor(VFResultsK->gcf_V2,type,point);
            break;

        default:
             color = 0;
             break;
        }
        if(last_orbit==NULL)
        {
            first_orbit=(struct orbits_points *) malloc(sizeof(struct orbits_points));
            last_orbit=first_orbit;
        }
        else
        {
            last_orbit->nextpt=(struct orbits_points *) malloc(sizeof(struct orbits_points));
            last_orbit=last_orbit->nextpt;
        }
        last_orbit->pcoord[0]=pcoord[0];
        last_orbit->pcoord[1]=pcoord[1];
        last_orbit->pcoord[2]=pcoord[2];
        last_orbit->color=color;
        last_orbit->dashes=dashes * VFResults.config_dashes;
        last_orbit->dir=((VFResults.plweights == false) && (chart==CHART_V1 || chart==CHART_V2)) ? VFResultsK->dir_vec_field*dir : dir;
        last_orbit->type=type;
        last_orbit->nextpt=NULL;
        
        if(last_orbit->dashes)
            (*plot_l)( spherewnd, pcoord,pcoord2,color);
        else
            (*plot_p)( spherewnd, pcoord,color);

        if(y[0]*y[0]+y[1]*y[1]>=1.0)
            de_sep->integrating_in_local_chart=false;

        if( !de_sep->integrating_in_local_chart )   // either outside ball, or a different vf region
            break;
        copy_x_into_y(pcoord,pcoord2);
    }
    de_sep->point[0]=y[0];
    de_sep->point[1]=y[1];
    
    set_current_step(fabs(hhi));

    if( first_orbit != NULL )
        *orbit=last_orbit;
    return(first_orbit);
}

// ---------------------------------------------------------------------------
//                  PLOT_SEP_BLOW_UP
// ---------------------------------------------------------------------------
//
// Starts separatrix integration, and plots the results meanwhile to the screen
//
// The first 100 points are drawn by looking at the formal power series development
// of the separatrix, as calculated by maple.  These points are drawn in an
// epsilon-ball around the singularity, where epsilon can be changed by the user.
//
// At the end, a normal integration cycle is added.

static struct orbits_points * plot_sep_blow_up(
                            QWinSphere * spherewnd,
                            double x0, double y0,
                            int chart, double epsilon,
                            struct blow_up_points *de_sep,
                            struct orbits_points ** orbit, int vfindex )
{
  double h,t=0,y,pcoord[3],pcoord2[3],point[2];
  int i,color,dir,dashes,type,ok=true;
  struct orbits_points *first_orbit=NULL,*last_orbit=NULL,*sep;

  switch(chart){
   case CHART_R2: MATHFUNC(R2_to_sphere)(x0,y0,pcoord); break;
   case CHART_U1: MATHFUNC(U1_to_sphere)(x0,y0,pcoord); break;
   case CHART_V1: MATHFUNC(V1_to_sphere)(x0,y0,pcoord); break;
   case CHART_U2: MATHFUNC(U2_to_sphere)(x0,y0,pcoord); break;
   case CHART_V2: MATHFUNC(V2_to_sphere)(x0,y0,pcoord); break;
  }
  
  if( !PrepareVfForIntegration(pcoord) )
    return NULL;


  /* if we have a line of singularities at infinity then we have to change the
     chart if the chart is V1 or V2 */
  if(VFResultsK->singinf){if(chart==CHART_V1)chart=CHART_U1;
               else if(chart==CHART_V2)chart=CHART_U2;
              }

  h = epsilon/100;
/*   h=find_step(de_sep->sep,epsilon,1.0)/100;  */

  type = de_sep->type;

  switch( de_sep->type )
  {
    case OT_STABLE:         dir=OT_STABLE;      break;
    case OT_UNSTABLE:       dir=OT_UNSTABLE;    break;
    case OT_CENT_STABLE:    dir=OT_STABLE;      break;
    case OT_CENT_UNSTABLE:  dir=OT_UNSTABLE;    break;
    default:                dir=0;              break;
  }
  point[0]=x0;point[1]=y0;
  first_orbit=(struct orbits_points *) malloc(sizeof(struct orbits_points));
  last_orbit=first_orbit;
  switch(chart){
   case CHART_R2: MATHFUNC(R2_to_sphere)(x0,y0,pcoord);
            color=FindSepColor(VFResultsK->gcf,de_sep->type,point);
            break;
   case CHART_U1: MATHFUNC(U1_to_sphere)(x0,y0,pcoord);
            color=FindSepColor(VFResultsK->gcf_U1,de_sep->type,point);
            break;
   case CHART_V1: MATHFUNC(V1_to_sphere)(x0,y0,pcoord);
            if(  !VFResults.plweights )psphere_to_V1(pcoord[0],pcoord[1],pcoord[2],point);
            color=FindSepColor(VFResultsK->gcf_V1,de_sep->type,point);
            break;
   case CHART_U2: MATHFUNC(U2_to_sphere)(x0,y0,pcoord);
            color=FindSepColor(VFResultsK->gcf_U2,de_sep->type,point);
            break;
   case CHART_V2: MATHFUNC(V2_to_sphere)(x0,y0,pcoord);
            if(  !VFResults.plweights  )psphere_to_V2(pcoord[0],pcoord[1],pcoord[2],point);
            color=FindSepColor(VFResultsK->gcf_V2,de_sep->type,point);
            break;
   default:
           color= 0;
           break;
  }
  
/*
    P5 Addition: Check if one can start with this separatrix.
*/


   y=eval_term1(de_sep->sep,h*100);
   make_transformations(de_sep->trans,
    de_sep->x0+de_sep->a11*h*100+de_sep->a12*y,
    de_sep->y0+de_sep->a21*h*100+de_sep->a22*y,point);
   switch(chart){
    case CHART_R2: MATHFUNC(R2_to_sphere)(point[0],point[1],pcoord2);
            break;
    case CHART_U1: if(point[1]>=0 || !VFResultsK->singinf) MATHFUNC(U1_to_sphere)(point[0],point[1],pcoord2);
                    else VV1_to_psphere(point[0],point[1],pcoord2); break;
    case CHART_V1: MATHFUNC(V1_to_sphere)(point[0],point[1],pcoord2); break;
    case CHART_U2:if(point[1]>=0 || !VFResultsK->singinf) MATHFUNC(U2_to_sphere)(point[0],point[1],pcoord2);
                    else VV2_to_psphere(point[0],point[1],pcoord2); break;
    case CHART_V2: MATHFUNC(V2_to_sphere)(point[0],point[1],pcoord2); break;
   }
   
    if( ThisVF->GetVFIndex_sphere( pcoord2 ) != vfindex )
    {
        free( first_orbit );
        return NULL;
    }

/*
    end ofP5 Addition
*/
  
  copy_x_into_y(pcoord,last_orbit->pcoord);
  last_orbit->color=color;
  last_orbit->dashes=0;  
  copy_x_into_y(pcoord,pcoord2);
  for(i=0;i<=99;i++){
   last_orbit->nextpt=(struct orbits_points *) malloc(sizeof
                 (struct orbits_points));
   last_orbit=last_orbit->nextpt;
   dashes=true;
   t=t+h;
   y=eval_term1(de_sep->sep,t);
   make_transformations(de_sep->trans,
    de_sep->x0+de_sep->a11*t+de_sep->a12*y,
    de_sep->y0+de_sep->a21*t+de_sep->a22*y,point);
   switch(chart){
    case CHART_R2: MATHFUNC(R2_to_sphere)(point[0],point[1],pcoord);
            color=FindSepColor(VFResultsK->gcf,de_sep->type,point);
            break;
    case CHART_U1: if(point[1]>=0 || !VFResultsK->singinf){MATHFUNC(U1_to_sphere)(point[0],point[1],pcoord);
                                          if(!ok){dashes=false; ok=true;}
                                          type=de_sep->type;
                                          color=FindSepColor(VFResultsK->gcf_U1,type,point);
                                         }
             else{VV1_to_psphere(point[0],point[1],pcoord);
                 if(ok){dashes=false; ok=false;}
                 psphere_to_V1(pcoord[0],pcoord[1],pcoord[2],point);
                 if(VFResultsK->dir_vec_field==1)type=change_type(de_sep->type);
                 else type=de_sep->type;
                 color=FindSepColor(VFResultsK->gcf_V1,type,point);
                 }
            break;
    case CHART_V1: MATHFUNC(V1_to_sphere)(point[0],point[1],pcoord);
            if( !VFResults.plweights )psphere_to_V1(pcoord[0],pcoord[1],pcoord[2],point);
            color=FindSepColor(VFResultsK->gcf_V1,de_sep->type,point);
            break;
    case CHART_U2:if(point[1]>=0 || !VFResultsK->singinf){MATHFUNC(U2_to_sphere)(point[0],point[1],pcoord);
                                         if(!ok){dashes=false; ok=true;}
                                          type=de_sep->type;
                                          color=FindSepColor(VFResultsK->gcf_U2,type,point);
                                        }
            else{VV2_to_psphere(point[0],point[1],pcoord);
                 if(ok){dashes=false; ok=false;}
                 psphere_to_V2(pcoord[0],pcoord[1],pcoord[2],point);
                 if(VFResultsK->dir_vec_field==1)type=change_type(de_sep->type);
                 else type=de_sep->type;
                 color=FindSepColor(VFResultsK->gcf_V2,type,point);
                }
            break;
    case CHART_V2: MATHFUNC(V2_to_sphere)(point[0],point[1],pcoord);
            if( !VFResults.plweights )psphere_to_V2(pcoord[0],pcoord[1],pcoord[2],point);
            color=FindSepColor(VFResultsK->gcf_V2,de_sep->type,point);
            break;
   }
   copy_x_into_y(pcoord,last_orbit->pcoord);
   last_orbit->color=color;
   last_orbit->dashes=dashes * VFResults.config_dashes;
   last_orbit->dir=dir;
   last_orbit->type=type;
   if(last_orbit->dashes)
    (*plot_l)(spherewnd,pcoord,pcoord2,color);
   else
    (*plot_p)(spherewnd,pcoord,color);
   copy_x_into_y(pcoord,pcoord2);
  }
  de_sep->point[0]=t; de_sep->point[1]=y;
  de_sep->integrating_in_local_chart=true;
  sep = last_orbit;
  last_orbit->nextpt=integrate_blow_up(spherewnd,pcoord2,de_sep,
    VFResults.config_step,dir,last_orbit->type,&sep,chart);
  if( first_orbit != NULL )
    *orbit=sep;
  return first_orbit;
}

void change_epsilon_de( QWinSphere * spherewnd, double epsilon)
{
    struct blow_up_points *separatrice;
    
    VFResults.selected_de_point->epsilon = epsilon;
    
    separatrice=VFResults.selected_de_point->blow_up;
    while( separatrice != NULL )
    {
        draw_selected_sep(spherewnd, separatrice->first_sep_point,CBACKGROUND);
        DeleteOrbitPoint(separatrice->first_sep_point);
        separatrice->last_sep_point=NULL;
        separatrice->first_sep_point=NULL;
        separatrice=separatrice->next_blow_up_point;
    }
}


void start_plot_de_sep( QWinSphere * spherewnd, int vfindex )
{
  struct orbits_points *points;
  double p[3];

    draw_sep(spherewnd,VFResults.selected_de_sep->first_sep_point);
    if( VFResults.selected_de_sep->last_sep_point )
    {
        if( VFResults.selected_de_sep->integrating_in_local_chart )
        {
            copy_x_into_y( VFResults.selected_de_sep->last_sep_point->pcoord, p );
            points=VFResults.selected_de_sep->last_sep_point;
            VFResults.selected_de_sep->last_sep_point->nextpt = integrate_blow_up(
                            spherewnd,
                            p,
                            VFResults.selected_de_sep,
                            VFResults.config_currentstep,
                            VFResults.selected_de_sep->last_sep_point->dir,
                            VFResults.selected_de_sep->last_sep_point->type,
                            &points,
                            VFResults.selected_de_point->chart );

            VFResults.selected_de_sep->last_sep_point = points;
        }
        else
        {
            copy_x_into_y( VFResults.selected_de_sep->last_sep_point->pcoord, p );
            points = VFResults.selected_de_sep->last_sep_point;
            VFResults.selected_de_sep->last_sep_point->nextpt = integrate_sep(
                            spherewnd,
                            p,
                            VFResults.config_currentstep,
                            VFResults.selected_de_sep->last_sep_point->dir,
                            VFResults.selected_de_sep->last_sep_point->type,
                            VFResults.config_intpoints,
                            &points );

            VFResults.selected_de_sep->last_sep_point = points;
        }
    }
    else
    {
        VFResults.selected_de_sep->first_sep_point = plot_sep_blow_up(
                            spherewnd,
                            VFResults.selected_de_point->x0,
                            VFResults.selected_de_point->y0,
                            VFResults.selected_de_point->chart,
                            VFResults.selected_de_point->epsilon,
                            VFResults.selected_de_sep,
                            &points, vfindex );

        VFResults.selected_de_sep->last_sep_point = points;
    }
}

void cont_plot_de_sep( QWinSphere * spherewnd )
{
    double p[3];
    struct orbits_points * points;

    copy_x_into_y( VFResults.selected_de_sep->last_sep_point->pcoord, p );
    points = VFResults.selected_de_sep->last_sep_point;
    if( VFResults.selected_de_sep->integrating_in_local_chart )
    {
        VFResults.selected_de_sep->last_sep_point->nextpt = integrate_blow_up(
                    spherewnd,
                    p,
                    VFResults.selected_de_sep,
                    VFResults.config_currentstep,
                    VFResults.selected_de_sep->last_sep_point->dir,
                    VFResults.selected_de_sep->last_sep_point->type,
                    &points,
                    VFResults.selected_de_point->chart );
    }
    else
    {
        VFResults.selected_de_sep->last_sep_point->nextpt = integrate_sep(
                    spherewnd,
                    p,
                    VFResults.config_currentstep,
                    VFResults.selected_de_sep->last_sep_point->dir,
                    VFResults.selected_de_sep->last_sep_point->type,
                    VFResults.config_intpoints,
                    &points);
    }

    VFResults.selected_de_sep->last_sep_point = points;
}

void plot_next_de_sep( QWinSphere * spherewnd, int vfindex )
{
    draw_sep( spherewnd, VFResults.selected_de_sep->first_sep_point );

    VFResults.selected_de_sep=VFResults.selected_de_sep->next_blow_up_point;
    if( VFResults.selected_de_sep == NULL )
        VFResults.selected_de_sep = VFResults.selected_de_point->blow_up;

    start_plot_de_sep( spherewnd, vfindex );
}


void select_next_de_sep( QWinSphere * spherewnd )
{
    draw_sep( spherewnd, VFResults.selected_de_sep->first_sep_point );
    
    VFResults.selected_de_sep = VFResults.selected_de_sep->next_blow_up_point;
    if( VFResults.selected_de_sep== NULL )
        VFResults.selected_de_sep = VFResults.selected_de_point->blow_up;

    draw_selected_sep(spherewnd, VFResults.selected_de_sep->first_sep_point, CW_SEP );
}

void plot_all_de_sep( QWinSphere * spherewnd, int vfindex, struct degenerate *point )
{
    struct blow_up_points * de_sep;
    struct orbits_points * sep;
    double p[3];

    while( point != NULL )
    {
        if( !IsARealSingularity( point->x0, point->y0, point->chart, vfindex ) )
        {
            point=point->next_de;
            continue;
        }
        if( point->notadummy )
        {
            de_sep = point->blow_up;
            while( de_sep != NULL )
            {
                sep=de_sep->last_sep_point;
                if(de_sep->last_sep_point)
                {
                    copy_x_into_y(de_sep->last_sep_point->pcoord,p);
                    if( de_sep->integrating_in_local_chart )
                        de_sep->last_sep_point->nextpt = integrate_blow_up(spherewnd,p,
                                                de_sep,VFResults.config_currentstep,de_sep->last_sep_point->dir,
                                                de_sep->last_sep_point->type,&sep,point->chart);
                    else
                        de_sep->last_sep_point->nextpt = integrate_sep(spherewnd,p,VFResults.config_currentstep,
                                                de_sep->last_sep_point->dir,de_sep->last_sep_point->type,
                                                VFResults.config_intpoints,&sep);
                    de_sep->last_sep_point=sep;
                }
                else
                {
                    de_sep->first_sep_point=plot_sep_blow_up(spherewnd,point->x0,point->y0,
                    point->chart,point->epsilon,de_sep,&sep,vfindex);
                    de_sep->last_sep_point=sep;
                }
                de_sep = de_sep->next_blow_up_point;
            }
        }
        point=point->next_de;
    }
}

