#include <qobject.h>
#include <math.h>
#include "custom.h"
#include "table.h"
#include "win_p5.h"
#include "math_p4.h"
#include "win_sphere.h"

#define __p                 VFResults.double_p              // p
#define __q                 VFResults.double_q              // q

static bool eval_orbit(double qp[3],double a,double b,double c,double pp[3],int dir );

extern bool PrepareVfForIntegration( double * );
extern void DrawOrbit( QWinSphere *, double *, struct orbits_points *, int );
void StoreLimitCycle(QWinSphere * spherewnd, double x,double y,double a,double b,double c);

extern bool stop_search_limit( void );                  // see WIN_LIMITCYCLES
extern void write_to_limit_window( double x, double y );

static bool less( double x0, double y0, double x1, double y1 )
{
    if( x0 < x1 )
        return true;

    if( x0==x1 && y0 < y1 )
        return true;

    return false;
}

// ---------------------------------------------------------------------------------------------
//                      SEARCHLIMITCYCLE
// ---------------------------------------------------------------------------------------------
//
// Loop to find limit cycles cutting some transverse section determined by two end points.

void SearchLimitCycle( QWinSphere * spherewnd, double x0, double y0, double x1, double y1, double grid )
{
    double p1[3],pf1[3],pb1[3],rf1[2],rb1[2];
    double p2[3],pf2[3],pb2[3],rf2[2],rb2[2],p3[3];
    double grid_dx;
    double grid_dy;
    double x,y,z;
    bool okf1, okb1;
    bool found;
    double a,b,c;           // ax+by+c=0 defines the equation of the line of the transverse section
    int counter;            // counter to keep steps: after so many steps, the progress window needs to be updated

    counter=0;

    // first make sure x0 < x1:

    if(x1<x0)
    {
        z=x0; x0=x1; x1=z;
        z=y0; y0=y1; y1=z;
    }
    else
        if((x0==x1) && (y1<y0))
        {
            z=y0;
            y0=y1;
            y1=z;
        }

    z=atan2(y1-y0,x1-x0);
    grid_dx=grid*cos(z);
    grid_dy=grid*sin(z);
    if((z<=PI_DIV4) && (z>=-PI_DIV4))
    {
        a = (y1-y0)/(x1-x0);        // abs(a) <= 1
        b = -1.0;
        c = y0 - a*x0;
    }
    else
    {
        a=-1.0;
        b = (x1-x0)/(y1-y0);        // abs(b) <= 1
        c=x0-b*y0;
    }

    // transverse section is located on the line a*X+b*Y+c*Z=0

    okf1 = false;
    okb1 = false;

    MATHFUNC(R2_to_sphere)(x1,y1,p2);

    found = true;
    while( found )
    {
        found = false;      // assume not found
        for(;;)
        {
            if( ++counter == VFResults.config_lc_value )
            {
                counter = 0;
                if( stop_search_limit() )
                {
                    okf1 = false;
                    okb1 = false;
                    break;
                }
            }

            MATHFUNC(R2_to_sphere)(x0,y0,p1);
            if( okf1 )
            {
                // the flag is still set from the previous time, where we found
                // a intersection through forward integration.  Now try backwards

                okf1 = false;
            }
            else
            {
                // check if forward integration intersects the segment

                okf1 = eval_orbit( p1, a, b, c, pf1, 1 );
                if( okf1 && MATHFUNC(less2)(p1,pf1) && MATHFUNC(less2)(pf1,p2) )
                    break;
            }
            if( okb1 )
            {
                // the flag is still set from the previous time, where we found
                // a intersection through backward integration.

                okb1 = false;
            }
            else
            {
                // check if backward integration intersects the segment

                okb1 = eval_orbit( p1, a, b, c, pb1, -1 );
                if( okb1 && MATHFUNC(less2)(p1,pb1) && MATHFUNC(less2)(pb1,p2) )
                    break;
            }

            // if both orbits intersect the line, but not on the segment,
            // then we have spiraling behaviour, so stop early

            if( okf1 && okb1 && MATHFUNC(less2)(pf1,p1) && MATHFUNC(less2)(p2,pb1) )
            {
                okf1 = false;
                okb1 = false;
                break;
            }
            if( okf1 && okb1 && MATHFUNC(less2)(p2,pf1) && MATHFUNC(less2)(pb1,p1) )
            {
                okf1 = false;
                okb1 = false;
                break;
            }

            // advance on the segment

            x0 += grid_dx;
            y0 += grid_dy;

            if( less(x1,y1,x0,y0) )
            {
                // stop at the end of the segment

                okf1 = false;
                okb1 = false;
                break;
            }
            write_to_limit_window( x0, y0 );
        }
        if( okf1 || okb1 ) // note: they cannot be both true here
        {
            for(;;)
            {
                x=x0; y=y0;
//              write_to_limit_window(x0,y0);
                if( ++counter == VFResults.config_lc_value )
                {
                    counter = 0;
                    if( stop_search_limit() )
                    {
                        okf1 = false;
                        okb1 = false;
                        break;
                    }
                }

                if( okf1 )
                {
                    MATHFUNC(sphere_to_R2)(pf1[0],pf1[1],pf1[2],rf1);
                    for(;;)
                    {
                        x=x+grid_dx; y=y+grid_dy;
                        if(less(rf1[0],rf1[1],x,y))
                            break;
                    }
                    if(less(x1,y1,x,y))
                        break;
                    MATHFUNC(R2_to_sphere)(x,y,p3);
                    if( eval_orbit(p3,a,b,c,pf2,1 ) )
                    {
                        if(MATHFUNC(less2)(pf2,p3) && MATHFUNC(less2)(p1,pf2))
                        {
                            MATHFUNC(sphere_to_R2)(pf2[0],pf2[1],pf2[2],rf2);
                            spherewnd->prepareDrawing();
                            StoreLimitCycle(spherewnd,(rf1[0]+rf2[0])/2,(rf1[1]+rf2[1])/2,a,b,c);
                            spherewnd->finishDrawing();
                            found=true;
                            p1[0]=p3[0]; p1[1]=p3[1]; p1[2]=p3[2];
                            x0=x; y0=y;
                            break;
                        }
                        if(MATHFUNC(less2)(pf2,p2))
                        {
                            p1[0]=p3[0]; p1[1]=p3[1]; p1[2]=p3[2];
                            pf1[0]=pf2[0]; pf1[1]=pf2[1]; pf1[2]=pf2[2];
                            x0=x; y0=y;
                        }
                    }
                    else
                    {
                        found = true;
                        x0 = x;
                        y0 = y;
                        break;
                    }
                }
                else // okb1=true
                {
                    MATHFUNC(sphere_to_R2)(pb1[0],pb1[1],pb1[2],rb1);
                    for(;;)
                    {
                        x=x+grid_dx; y=y+grid_dy;
                        if(less(rb1[0],rb1[1],x,y))
                            break;
                    }
                    if(less(x1,y1,x,y))
                        break;
                    MATHFUNC(R2_to_sphere)(x,y,p3);
                    if( eval_orbit(p3,a,b,c,pb2,-1) )
                    {
                        if(MATHFUNC(less2)(pb2,p3) && MATHFUNC(less2)(p1,pb2))
                        {
                            MATHFUNC(sphere_to_R2)(pb2[0],pb2[1],pb2[2],rb2);
                            spherewnd->prepareDrawing();
                            StoreLimitCycle(spherewnd,(rb1[0]+rb2[0])/2,(rb1[1]+rb2[1])/2,a,b,c);
                            spherewnd->finishDrawing();
                            found=true;
                            p1[0]=p3[0]; p1[1]=p3[1]; p1[2]=p3[2];
                            x0=x; y0=y;
                            break;
                        }
                        if(MATHFUNC(less2)(pb2,p2))
                        {
                            p1[0]=p3[0]; p1[1]=p3[1]; p1[2]=p3[2];
                            pb1[0]=pb2[0]; pb1[1]=pb2[1]; pb1[2]=pb2[2];
                            x0=x; y0=y;
                        }
                    }
                    else
                    {
                        found=true; x0=x; y0=y;
                        break;
                    }
                }
            }
        }
    }
}

// ---------------------------------------------------------------------------------------------
//                      EVAL_ORBIT
// ---------------------------------------------------------------------------------------------
//
// Given a point qp and a transverse section ax+by+cz = 0, the orbit through qp is integrated
// for a maximum number of integration points.  If meanwhile the section is crossed, this
// is detected.  If not, false is returned.
//
// We need to detect two intersections (a periodic orbit needs to intersect any line an even
// number of times).  At the second intersection it tries to locate as precisely as possible the intersection
// point of the orbit with the transverse section and stores it in pp.

static bool eval_orbit( double qp[3], double a, double b, double c, double pp[3], int dir )
{
    double p1[3], p2[3];
    double hhi, h_max, h_min, hhi2, i;
    int dashes, d;
    bool ok;

    ok = false;

    hhi = (double)dir * VFResults.config_step;
    h_max = VFResults.config_hma;
    h_min = VFResults.config_hmi;
    copy_x_into_y( qp, p1 );
    if( !PrepareVfForIntegration( p1 ) )
        return false;
    MATHFUNC(integrate_sphere_orbit)( p1[0], p1[1], p1[2], p2, &hhi, &dashes, &d, h_min, h_max);
    for( i = 0; i <= VFResults.config_lc_numpoints; i++ )
    {
        copy_x_into_y(p2,p1);
        if( !PrepareVfForIntegration( p1 ) )
            return false;
        MATHFUNC(integrate_sphere_orbit)( p1[0], p1[1], p1[2], p2, &hhi, &dashes, &d, h_min, h_max);
        if((MATHFUNC(eval_lc)(p1,a,b,c)*MATHFUNC(eval_lc)(p2,a,b,c)) <= 0 )
        {
            // we have crossed the line on which the transverse section is located
            // (a*X+b*Y+c*Z has changed sign)

            ok = true;
            break;
        }
    }

    if( !ok )
        return false;

    ok = false;
    copy_x_into_y(p2,p1);
    if( !PrepareVfForIntegration( p1 ) )
        return false;
    MATHFUNC(integrate_sphere_orbit)( p1[0], p1[1], p1[2], p2, &hhi, &dashes, &d, h_min, h_max );
    for( i = 0; i <= VFResults.config_lc_numpoints; i++ )
    {
        copy_x_into_y(p2,p1);
        hhi2=hhi;
        if( !PrepareVfForIntegration( p1 ) )
            return false;
        MATHFUNC(integrate_sphere_orbit)( p1[0], p1[1], p1[2], p2, &hhi, &dashes, &d, h_min, h_max);
        if((MATHFUNC(eval_lc)(p1,a,b,c)*MATHFUNC(eval_lc)(p2,a,b,c)) <= 0 )
        {
            ok=1;
            break;
        }
    }
    if( !ok )
        return false;

    if( fabs(hhi2) < fabs(hhi) )
        hhi = hhi2;

        /* find the intersection point p3 with the poincare section
            with a precision of 1e-8         */

    if((fabs(MATHFUNC(eval_lc)(p2,a,b,c))<=1e-8) || fabs(hhi2)<=h_min)
    {
        copy_x_into_y(p2,pp);
    }
    else
    {
        for(;;)
        {
            hhi2=hhi/2.0;
            if( !PrepareVfForIntegration( p1 ) )
                return false;
            MATHFUNC(integrate_sphere_orbit)(p1[0],p1[1],p1[2],pp,&hhi2,&dashes,&d,h_min,fabs(hhi2));
            if((fabs(MATHFUNC(eval_lc)(pp,a,b,c))<=1e-8) || (fabs(hhi2)<=h_min))
                break;
            if((MATHFUNC(eval_lc)(p1,a,b,c)*MATHFUNC(eval_lc)(pp,a,b,c))>0)
            {
                copy_x_into_y(pp,p1);
                hhi=(double)dir*(fabs(hhi)-fabs(hhi2));
            }
            else hhi=hhi2;
        }
    }
    return ok;
}

// ---------------------------------------------------------------------------------------------
//                      STORELIMITCYCLE
// ---------------------------------------------------------------------------------------------
//
// The found limit cycle is re-integrated, and stored in memory.
// It is meanwhile drawn on the screen.
// The limit cycle is found through forward integration.

void StoreLimitCycle(QWinSphere * spherewnd, double x,double y,double a,double b,double c)
{
    double p1[3],p2[3];
    double hhi,h_max,h_min;
    int dashes,d;
    struct orbits * LC;

    if( VFResults.current_lim_cycle == NULL )
    {
        VFResults.first_lim_cycle = (struct orbits *)malloc( sizeof(struct orbits) );
        VFResults.current_lim_cycle = VFResults.first_lim_cycle;
    }
    else
    {
        VFResults.current_lim_cycle->next = (struct orbits *)malloc( sizeof(struct orbits) );
        VFResults.current_lim_cycle = VFResults.current_lim_cycle->next;
    }
    LC = VFResults.current_lim_cycle;

    MATHFUNC(R2_to_sphere)(x,y,p1);
    copy_x_into_y( p1, LC->pcoord );
    LC->color = CLIMIT;
    LC->firstpt = NULL;
    LC->next = NULL;
    (*plot_p)(spherewnd,p1,CLIMIT);

    hhi = VFResults.config_step;
    h_max = VFResults.config_hma;
    h_min = VFResults.config_hmi;
    if( !PrepareVfForIntegration( p1 ) )
        return;
    MATHFUNC(integrate_sphere_orbit)( p1[0], p1[1], p1[2], p2, &hhi, &dashes, &d, h_min, h_max );

    LC->firstpt = (struct orbits_points *)malloc( sizeof(struct orbits_points) );
    LC->currentpt = LC->firstpt;
    copy_x_into_y( p2, LC->currentpt->pcoord );
    LC->currentpt->color = CLIMIT;
    LC->currentpt->dashes = VFResults.config_dashes;
    LC->currentpt->nextpt = NULL;
    if( VFResults.config_dashes )
        (*plot_l)( spherewnd, p1, p2, CLIMIT );
    else
        (*plot_p)( spherewnd, p2, CLIMIT );

    for(;;)
    {
        copy_x_into_y( p2,p1);
        if( !PrepareVfForIntegration( p1 ) )
            break;
        MATHFUNC(integrate_sphere_orbit)(p1[0],p1[1],p1[2],p2,&hhi,&dashes,&d,h_min,h_max);

        LC->currentpt->nextpt=(struct orbits_points *)malloc(sizeof(struct orbits_points));
        LC->currentpt=LC->currentpt->nextpt;
        copy_x_into_y(p2,LC->currentpt->pcoord);
        LC->currentpt->color=CLIMIT;
        LC->currentpt->dashes=VFResults.config_dashes;
        LC->currentpt->nextpt=NULL;

        if(VFResults.config_dashes)
            (*plot_l)(spherewnd,p1,p2,CLIMIT);
        else
            (*plot_p)(spherewnd,p2,CLIMIT);

        if((MATHFUNC(eval_lc)(p1,a,b,c)*MATHFUNC(eval_lc)(p2,a,b,c))<=0)
            break;
    }
    copy_x_into_y(p2,p1);
    if( !PrepareVfForIntegration( p1 ) )
        return;
    MATHFUNC(integrate_sphere_orbit)(p1[0],p1[1],p1[2],p2,&hhi,&dashes,&d,h_min,h_max);
    for(;;)
    {
        copy_x_into_y(p2,p1);
        if( !PrepareVfForIntegration( p1 ) )
            break;
        MATHFUNC(integrate_sphere_orbit)(p1[0],p1[1],p1[2],p2,&hhi,&dashes,&d,h_min,h_max);

        LC->currentpt->nextpt=(struct orbits_points *)malloc(sizeof(struct orbits_points));
        LC->currentpt=LC->currentpt->nextpt;
        copy_x_into_y(p2,LC->currentpt->pcoord);
        LC->currentpt->color=CLIMIT;
        LC->currentpt->dashes=VFResults.config_dashes;
        LC->currentpt->nextpt=NULL;
        if((MATHFUNC(eval_lc)(p1,a,b,c)*MATHFUNC(eval_lc)(p2,a,b,c))<=0)
            break;
        if(VFResults.config_dashes)
            (*plot_l)(spherewnd,p1,p2,CLIMIT);
        else
            (*plot_p)(spherewnd,p2,CLIMIT);
    }
    MATHFUNC(R2_to_sphere)(x,y,p2);
    copy_x_into_y(p2,LC->currentpt->pcoord);
    if(VFResults.config_dashes)
        (*plot_l)(spherewnd,p1,p2,CLIMIT);
    else
        (*plot_p)(spherewnd,p2,CLIMIT);
}

// -----------------------------------------------------------------------
//                      DRAWLIMITCYCLES
// -----------------------------------------------------------------------
//
// Draw limit cycles that were calculated earlier.  This is called during
// a repaint (but also during a print command).

void DrawLimitCycles( QWinSphere * spherewnd )
{
    struct orbits * orbit;

    orbit = VFResults.first_lim_cycle;
    while( orbit != NULL )
    {
        DrawOrbit( spherewnd, orbit->pcoord, orbit->firstpt, orbit->color );
        orbit=orbit->next;
    }
}

// -----------------------------------------------------------------------
//                      DELETELASTLIMITCYCLE
// -----------------------------------------------------------------------

void DeleteLastLimitCycle( QWinSphere * spherewnd )
{
    struct orbits * orbit1;
    struct orbits * orbit2;

    if( VFResults.current_lim_cycle == NULL )
        return;

    orbit2 = VFResults.current_lim_cycle;
    DrawOrbit( spherewnd, orbit2->pcoord, orbit2->firstpt, spherewnd->spherebgcolor );

    if( VFResults.first_lim_cycle == VFResults.current_lim_cycle )
    {
        VFResults.first_lim_cycle = NULL;
        VFResults.current_lim_cycle = NULL;
    }
    else
    {
        orbit1 = VFResults.first_lim_cycle;
        do
        {
            VFResults.current_lim_cycle = orbit1;
            orbit1 = orbit1->next;
        }
        while( orbit1 != orbit2 );

        VFResults.current_lim_cycle->next = NULL;
    }
    DeleteOrbitPoint( orbit2->firstpt );
    free( orbit2 );
}
