#include <qobject.h>
#include <math.h>
#include "custom.h"
#include "table.h"
#include "win_p5.h"
#include "math_p4.h"
#include "win_sphere.h"
#include "file_vf.h"

struct orbits_points * integrate_orbit( QWinSphere *, double *, double,int, int, int, struct orbits_points ** );

// -----------------------------------------------------------------------
//                      PREPAREVFFORINTEGRATION
// -----------------------------------------------------------------------

bool PrepareVfForIntegration( double * pcoord )
{
    int K;

    K = ThisVF->GetVFIndex_sphere( pcoord );
    if( K >= 0 )
    {
        VFResults.vfK = VFResults.vf[K];
        if( K != VFResults.K )
        {
            VFResults.K = K;
            //VFResults.config_currentstep = VFResults.config_step;
        }
        return true;
    }
    else
    {
        VFResults.vfK = NULL;
        return false;
    }
}

// -----------------------------------------------------------------------
//                      INTEGRATEORBIT
// -----------------------------------------------------------------------
//
// dir = -1: backwards, dir=0: continue, dir=+1: forwards
//
// Continues orbit integration

void IntegrateOrbit( QWinSphere * sphere, int dir )
{
    struct orbits_points *sep;
    double pcoord[3],ucoord[2];

    if( dir == 0 )
    {
        // continue orbit button has been pressed

        dir = VFResults.current_orbit->currentpt->dir;

        copy_x_into_y( VFResults.current_orbit->currentpt->pcoord, pcoord );
        if( !PrepareVfForIntegration( pcoord ) )
            return;
        sep = VFResults.current_orbit->currentpt;
        VFResults.current_orbit->currentpt->nextpt =
                    integrate_orbit( sphere, pcoord,VFResults.config_currentstep,dir,CORBIT, VFResults.config_intpoints,&sep);

        VFResults.current_orbit->currentpt = sep;
        return;
    }

    copy_x_into_y(VFResults.current_orbit->pcoord,pcoord);
    MATHFUNC(sphere_to_R2)(pcoord[0],pcoord[1],pcoord[2],ucoord);

    if( !PrepareVfForIntegration( pcoord ) )
        return;

    if( VFResults.config_kindvf == INTCONFIG_ORIGINAL )
        if(eval_term2(VFResultsK->gcf,ucoord) < 0)
            dir = -dir;

    if( VFResults.current_orbit->firstpt == NULL )
    {
        sep = NULL;
        VFResults.current_orbit->firstpt =
                    integrate_orbit(sphere,pcoord,VFResults.config_step,dir,CORBIT,VFResults.config_intpoints,&sep);
    }
    else
    {
        VFResults.current_orbit->currentpt->nextpt=(struct orbits_points *) malloc(sizeof(struct orbits_points));
        VFResults.current_orbit->currentpt=VFResults.current_orbit->currentpt->nextpt;
        copy_x_into_y(pcoord,VFResults.current_orbit->currentpt->pcoord);
        VFResults.current_orbit->currentpt->dashes=0;
        VFResults.current_orbit->currentpt->color=CORBIT;
        VFResults.current_orbit->currentpt->dir=dir;
        sep = VFResults.current_orbit->currentpt;
        VFResults.current_orbit->currentpt->nextpt =
                    integrate_orbit(sphere,pcoord,VFResults.config_step,dir,CORBIT,VFResults.config_intpoints,&sep);
    }
    VFResults.current_orbit->currentpt=sep;
}

// -----------------------------------------------------------------------
//                      STARTORBIT
// -----------------------------------------------------------------------
/* R=0 then point selected in the drawing canvas else in the orbit window */

bool StartOrbit( QWinSphere * sphere, double x, double y, bool R )
{
    double pcoord[3];
    double ucoord[2];

    if( R )
        MATHFUNC(R2_to_sphere)(x,y,pcoord);
    else
    {
        if( (MATHFUNC(is_valid_viewcoord)( x, y, pcoord )) == false )
            return false;
    }
    if( !PrepareVfForIntegration( pcoord ) )
        return false;

    if( VFResults.first_orbit == NULL )
    {
        VFResults.first_orbit = (struct orbits *)malloc( sizeof(struct orbits) );
        VFResults.current_orbit = VFResults.first_orbit;
    }
    else
    {
        VFResults.current_orbit->next = (struct orbits *)malloc( sizeof(struct orbits) );
        VFResults.current_orbit = VFResults.current_orbit->next;
    }

    copy_x_into_y( pcoord, VFResults.current_orbit->pcoord );
    VFResults.current_orbit->color = CORBIT;
    VFResults.current_orbit->firstpt = NULL;
    VFResults.current_orbit->next = NULL;

    MATHFUNC(sphere_to_viewcoord)( pcoord[0], pcoord[1], pcoord[2], ucoord );
    sphere->drawPoint( ucoord[0], ucoord[1], CORBIT );

    return true;
}

// -----------------------------------------------------------------------
//                      DRAWORBIT
// -----------------------------------------------------------------------

void DrawOrbit( QWinSphere * spherewnd, double * pcoord, struct orbits_points * points, int color )
{
    double pcoord1[3];

    copy_x_into_y( pcoord, pcoord1 );
    (*plot_p)( spherewnd, pcoord, color );

    while( points!=NULL )
    {
        if(points->dashes)
        {
            (*plot_l)(spherewnd, pcoord1,points->pcoord,color);
        }
        else
        {
            (*plot_p)(spherewnd, points->pcoord,color);
        }

        copy_x_into_y(points->pcoord,pcoord1);

        points = points->nextpt;
    }
}

// -----------------------------------------------------------------------
//                      DRAWORBITS
// -----------------------------------------------------------------------

void DrawOrbits( QWinSphere * spherewnd )
{
    struct orbits * orbit;

    for( orbit = VFResults.first_orbit; orbit != NULL; orbit = orbit->next )
    {
        DrawOrbit( spherewnd, orbit->pcoord, orbit->firstpt, orbit->color );
    }
}

// -----------------------------------------------------------------------
//                      DELETELASTORBIT
// -----------------------------------------------------------------------

void DeleteLastOrbit( QWinSphere * spherewnd )
{
    struct orbits *orbit1,*orbit2;

    if( VFResults.current_orbit == NULL )
        return;

    orbit2 = VFResults.current_orbit;
    DrawOrbit( spherewnd, orbit2->pcoord, orbit2->firstpt, spherewnd->spherebgcolor );

    if( VFResults.first_orbit == VFResults.current_orbit )
    {
        VFResults.first_orbit = NULL;
        VFResults.current_orbit = NULL;
    }
    else
    {
        orbit1 = VFResults.first_orbit;

        do
        {
            VFResults.current_orbit = orbit1;
            orbit1 = orbit1->next;
        }
        while( orbit1 != orbit2 );

        VFResults.current_orbit->next = NULL;
    }
    DeleteOrbitPoint( orbit2->firstpt );
    free( orbit2 );
}


// ---------------------------------------------------------------------------
//                  INTEGRATE_POINCARE_ORBIT
// ---------------------------------------------------------------------------
//
// integrate poincare sphere case p=q=1
//
// This calculates 1 step.
//
//
// ALTERED FOR P5

void integrate_poincare_orbit( double p0, double p1, double p2, double * pcoord,
                              double * hhi, int * dashes, int * dir, double h_min, double h_max)
{
    double y[2];
    double theta;
    double hhi0;

    if( pcoord[2]>ZCOORD )
    {
        hhi0 = *hhi;
        while( 1 )
        {
            *dashes = true;
            *dir = 1;
            psphere_to_R2(p0,p1,p2,y);
            rk78(eval_r_vec_field,y,&hhi0,h_min,h_max,VFResults.config_tolerance);
            if( ThisVF->GetVFIndex_R2( y ) == VFResults.K )
                break;
            h_min = VFResults.config_branchhmi;
            h_max /= 2;
            hhi0 = fabs(hhi0*(*hhi))/2/(*hhi);
            if( fabs(hhi0) < h_min || h_max < VFResults.config_branchhmi )
            {
                hhi0 = h_min;
                if( *hhi<0 ) hhi0=-hhi0;
                break;
            }
        }
        *hhi = hhi0;
        R2_to_psphere( y[0], y[1], pcoord );
    }
    else
    {
        theta=atan2(fabs(p1),fabs(p0));
        if((theta<PI_DIV4) && (theta>-PI_DIV4))
        {
            if(p0>0)
            {
                hhi0 = *hhi;
                while( 1 )
                {
                    *dashes = true;
                    *dir = 1;
                    psphere_to_U1(p0,p1,p2,y);
                    rk78(eval_U1_vec_field,y,&hhi0,h_min,h_max,VFResults.config_tolerance);
                    if(y[1]>=0 || !VFResultsK->singinf)
                    {
                        if( ThisVF->GetVFIndex_U1( y ) == VFResults.K )
                            break;
                    }
                    else
                    {
                        if( ThisVF->GetVFIndex_VV1( y ) == VFResults.K )
                            break;
                    }
                    h_min = VFResults.config_branchhmi;
                    h_max /= 2;
                    hhi0 = fabs(hhi0*(*hhi))/2/(*hhi);
                    if( fabs(hhi0) < h_min || h_max < VFResults.config_branchhmi )
                    {
                        hhi0 = h_min;
                        if( *hhi<0 ) hhi0=-hhi0;
                        break;
                    }
                }
                *hhi = hhi0;
                if(y[1]>=0 || !VFResultsK->singinf)
                    U1_to_psphere(y[0],y[1],pcoord);
                else
                {
                    VV1_to_psphere(y[0],y[1],pcoord);
                    if(VFResultsK->dir_vec_field==1)
                    {
                        *dir=-1;
                        *hhi=-(*hhi);
                    }
                    psphere_to_V1( pcoord[0], pcoord[1], pcoord[2], y );
                    *dashes=false;
                }
            }
            else
            {
                hhi0 = *hhi;
                while( 1 )
                {
                    *dashes = true;
                    *dir = 1;
                    psphere_to_V1(p0,p1,p2,y);
                    rk78(eval_V1_vec_field,y,&hhi0,h_min,h_max,VFResults.config_tolerance);
                    if(y[1]>=0 || !VFResultsK->singinf)
                    {
                        if( ThisVF->GetVFIndex_V1( y ) == VFResults.K )
                            break;
                    }
                    else
                    {
                        if( ThisVF->GetVFIndex_UU1( y ) == VFResults.K )
                            break;
                    }
                    h_min = VFResults.config_branchhmi;
                    h_max /= 2;
                    hhi0 = fabs(hhi0*(*hhi))/2/(*hhi);
                    if( fabs(hhi0) < h_min || h_max < VFResults.config_branchhmi )
                    {
                        hhi0 = h_min;
                        if( *hhi<0 ) hhi0=-hhi0;
                        break;
                    }
                }
                *hhi = hhi0;
                if(y[1]>=0 || !VFResultsK->singinf)
                    V1_to_psphere(y[0],y[1],pcoord);
                else
                {
                    UU1_to_psphere(y[0],y[1],pcoord);
                    if(VFResultsK->dir_vec_field==1)
                    {
                        *dir=-1;
                        *hhi=-(*hhi);
                    }
                    psphere_to_U1(pcoord[0],pcoord[1],pcoord[2],y);
                    *dashes=false;
                }
            }
        }
        else
        {
            if(p1>0)
            {
                hhi0 = *hhi;
                while( 1 )
                {
                    *dashes = true;
                    *dir = 1;
                    psphere_to_U2(p0,p1,p2,y);
                    rk78(eval_U2_vec_field,y,&hhi0,h_min,h_max,VFResults.config_tolerance);
                    if(y[1]>=0 || !VFResultsK->singinf)
                    {
                        if( ThisVF->GetVFIndex_U2( y ) == VFResults.K )
                            break;
                    }
                    else
                    {
                        if( ThisVF->GetVFIndex_VV2( y ) == VFResults.K )
                            break;
                    }
                    h_min = VFResults.config_branchhmi;
                    h_max /= 2;
                    hhi0 = fabs(hhi0*(*hhi))/2/(*hhi);
                    if( fabs(hhi0) < h_min || h_max < VFResults.config_branchhmi )
                    {
                        hhi0 = h_min;
                        if( *hhi<0 ) hhi0=-hhi0;
                        break;
                    }
                }
                *hhi = hhi0;
                if(y[1]>=0 || !VFResultsK->singinf)
                    U2_to_psphere(y[0],y[1],pcoord);
                else
                {
                    VV2_to_psphere(y[0],y[1],pcoord);
                    if(VFResultsK->dir_vec_field==1)
                    {
                        *dir=-1;
                        *hhi=-(*hhi);
                    }
                    psphere_to_V2(pcoord[0],pcoord[1],pcoord[2],y);
                    *dashes=false;
                }
            }
            else
            {
                hhi0 = *hhi;
                while( 1 )
                {
                    *dashes = true;
                    *dir = 1;
                    psphere_to_V2(p0,p1,p2,y);
                    rk78(eval_V2_vec_field,y,&hhi0,h_min,h_max,VFResults.config_tolerance);
                    if( y[1]>=0 || !VFResultsK->singinf)
                    {
                        if( ThisVF->GetVFIndex_V2( y ) == VFResults.K )
                            break;
                    }
                    else
                    {
                        if( ThisVF->GetVFIndex_UU2( y ) == VFResults.K )
                            break;
                    }
                    h_min = VFResults.config_branchhmi;
                    h_max /= 2;
                    hhi0 = fabs(hhi0*(*hhi))/2/(*hhi);
                    if( fabs(hhi0) < h_min || h_max < VFResults.config_branchhmi )
                    {
                        hhi0 = h_min;
                        if( *hhi<0 ) hhi0=-hhi0;
                        break;
                    }
                }
                *hhi = hhi0;
                if( y[1]>=0 || !VFResultsK->singinf)
                    V2_to_psphere(y[0],y[1],pcoord);
                else
                {
                    UU2_to_psphere(y[0],y[1],pcoord);
                    if(VFResultsK->dir_vec_field==1)
                    {
                        *dir=-1;
                        *hhi=-(*hhi);
                    }
                    psphere_to_U2(pcoord[0],pcoord[1],pcoord[2],y);
                    *dashes=false;
                }
            }
        }
    }
}

// ---------------------------------------------------------------------------
//                  INTEGRATE_LYAPUNOV_ORBIT
// ---------------------------------------------------------------------------
//
// integrate on the Poincare-Lyapunov sphere
//
// This calculates 1 step.

void integrate_lyapunov_orbit( double p0, double p1, double p2, double * pcoord,
                              double * hhi, int * dashes, int * dir, double h_min, double h_max)
{
    double y[2];
    double hhi0;

    hhi0 = *hhi;

    if( p0==0 )
    {
        while(1)
        {
            *dashes=true;
            *dir=1;
            y[0]=p1;
            y[1]=p2;
            rk78(eval_r_vec_field,y,hhi,h_min,h_max,VFResults.config_tolerance);
            if( ThisVF->GetVFIndex_R2( y ) == VFResults.K )
                break;
            h_min = VFResults.config_branchhmi;
            h_max /= 2;
            hhi0 = fabs(hhi0*(*hhi))/2/(*hhi);
            if( fabs(hhi0) < h_min || h_max < VFResults.config_branchhmi )
            {
                hhi0 = h_min;
                if( *hhi<0 ) hhi0=-hhi0;
                break;
            }
        }
        *hhi = hhi0;
        R2_to_plsphere(y[0],y[1],pcoord);
    }
    else
    {
        while(1)
        {
            *dashes=true;
            *dir=1;
            y[0]=p1;
            y[1]=p2;
            rk78(eval_vec_field_cyl,y,hhi,h_min,h_max,VFResults.config_tolerance);
            if(y[1]>=TWOPI)y[1]-=TWOPI;
            if( ThisVF->GetVFIndex_cyl( y ) == VFResults.K )
                break;
            h_min = VFResults.config_branchhmi;
            h_max /= 2;
            hhi0 = fabs(hhi0*(*hhi))/2/(*hhi);
            if( fabs(hhi0) < h_min || h_max < VFResults.config_branchhmi )
            {
                hhi0 = h_min;
                if( *hhi<0 ) hhi0=-hhi0;
                break;
            }
        }
        *hhi = hhi0;
        cylinder_to_plsphere(y[0],y[1],pcoord);
    }
}

// ---------------------------------------------------------------------------
//                  INTEGRATE_ORBIT
// ---------------------------------------------------------------------------
//
// Integrate a number of points (user-dependent)
//

struct orbits_points * integrate_orbit(QWinSphere * spherewnd, double pcoord[3],double step,int dir,int color,
    int points_to_int,struct orbits_points **orbit)
{
    int i,d,h;
    int dashes;
    double hhi;
    double pcoord2[3],h_min,h_max;
    struct orbits_points *first_orbit=NULL,*last_orbit=NULL;

    hhi = (double)dir * step;
    h_min = VFResults.config_hmi;
    h_max = VFResults.config_hma;

    copy_x_into_y( pcoord, pcoord2 );

    for( i=1; i <= points_to_int ; ++i )
    {
/* P5 ADDITION: */
        if( !PrepareVfForIntegration( pcoord ) )
            break;
/* P5 END OF ADDITION */

        MATHFUNC(integrate_sphere_orbit)(pcoord[0],pcoord[1],pcoord[2],pcoord,
                                        &hhi,&dashes,&d,h_min,h_max);

        if( (i % UPDATEFREQ_STEPSIZE) == 0 )
            set_current_step(fabs(hhi));

        if(last_orbit==NULL)
        {
            first_orbit=(struct orbits_points *) malloc(sizeof(struct orbits_points));
            last_orbit=first_orbit; h=dir;
        }
        else
        {
            last_orbit->nextpt=(struct orbits_points *) malloc(sizeof(struct orbits_points));
            h=last_orbit->dir;
            last_orbit=last_orbit->nextpt;
        }

        copy_x_into_y(pcoord,last_orbit->pcoord);
        last_orbit->color=color;
        last_orbit->dashes=dashes * VFResults.config_dashes;
        last_orbit->dir=d*h;
        last_orbit->nextpt=NULL;
        if(dashes * VFResults.config_dashes)
            (*plot_l)(spherewnd,pcoord,pcoord2,color);
        else
            (*plot_p)(spherewnd,pcoord,color);
        copy_x_into_y(pcoord,pcoord2);
    }
    set_current_step(fabs(hhi));
    if( first_orbit != NULL )
        *orbit=last_orbit;
    return(first_orbit);
}
