#include <qobject.h>
#include <math.h>
#include "custom.h"
#include "table.h"
#include "math_p4.h"
#include "file_vf.h"

extern bool PrepareVfForIntegration( double * );
extern struct orbits_points *integrate_sep( QWinSphere *, double *, double, int, int, int, struct orbits_points ** );
extern struct orbits_points * plot_separatrice( QWinSphere *, double, double, double, double, double, double,
                            double, struct sep *, struct orbits_points **, int, int );
extern bool IsARealSingularity( double, double, int, int );

// ---------------------------------------------------------------------------
//                  START_PLOT_SADDLE_SEP
// ---------------------------------------------------------------------------

void start_plot_saddle_sep( QWinSphere * spherewnd, int vfindex )
{
    double p[3];
    struct orbits_points *points;
    
    draw_sep( spherewnd,VFResults.selected_sep->first_sep_point );
    
    points=VFResults.selected_sep->last_sep_point;
    if( VFResults.selected_sep->last_sep_point != NULL )
    {
        copy_x_into_y(VFResults.selected_sep->last_sep_point->pcoord,p);
        VFResults.selected_sep->last_sep_point->nextpt = integrate_sep(
                        spherewnd, p,VFResults.config_currentstep,VFResults.selected_sep->last_sep_point->dir,
                        VFResults.selected_sep->last_sep_point->type,VFResults.config_intpoints,&points);
    }
    else
    {
        VFResults.selected_sep->first_sep_point=plot_separatrice(spherewnd, VFResults.selected_saddle_point->x0,
                                VFResults.selected_saddle_point->y0,
                                VFResults.selected_saddle_point->a11,VFResults.selected_saddle_point->a12,
                                VFResults.selected_saddle_point->a21,VFResults.selected_saddle_point->a22,
                                VFResults.selected_saddle_point->epsilon,
                                VFResults.selected_sep,&points,
                                VFResults.selected_saddle_point->chart,vfindex);
    }
    VFResults.selected_sep->last_sep_point=points;
}

void cont_plot_saddle_sep( QWinSphere * spherewnd )
{
    double p[3];
    struct orbits_points *points;
    
    if( VFResults.selected_sep->first_sep_point == NULL )
        return;
    
    copy_x_into_y(VFResults.selected_sep->last_sep_point->pcoord,p);
    points=VFResults.selected_sep->last_sep_point;

    VFResults.selected_sep->last_sep_point->nextpt = integrate_sep(spherewnd, p,VFResults.config_currentstep,
                                        VFResults.selected_sep->last_sep_point->dir,
                                        VFResults.selected_sep->last_sep_point->type,
                                        VFResults.config_intpoints,&points);

    VFResults.selected_sep->last_sep_point=points;
}

void plot_next_saddle_sep( QWinSphere * spherewnd, int vfindex )
{
    draw_sep(spherewnd,VFResults.selected_sep->first_sep_point);
    
    VFResults.selected_sep=VFResults.selected_sep->next_sep;
    if( VFResults.selected_sep == NULL )
        VFResults.selected_sep=VFResults.selected_saddle_point->separatrices;
    
    start_plot_saddle_sep( spherewnd, vfindex );
}

void select_next_saddle_sep( QWinSphere * spherewnd )
{
    draw_sep(spherewnd,VFResults.selected_sep->first_sep_point);
    
    VFResults.selected_sep=VFResults.selected_sep->next_sep;
    if( VFResults.selected_sep == NULL )
        VFResults.selected_sep=VFResults.selected_saddle_point->separatrices;

    draw_selected_sep(spherewnd,VFResults.selected_sep->first_sep_point,CW_SEP);
}

void plot_all_saddle_sep(QWinSphere * spherewnd, int vfindex, struct saddle *point)
{
  struct sep *sep1;
  struct orbits_points *points;
  double p[3];

    while( point != NULL )
    {
        if( !IsARealSingularity( point->x0, point->y0, point->chart, vfindex ) )
        {
            point=point->next_saddle;
            continue;
        }
        if( point->notadummy )
        {
            sep1=point->separatrices;
            while( sep1 != NULL )
            {
                points=sep1->last_sep_point;
                if(sep1->last_sep_point)
                {
                    copy_x_into_y(sep1->last_sep_point->pcoord,p);
                    sep1->last_sep_point->nextpt = integrate_sep(spherewnd, p,
                                    VFResults.config_currentstep,sep1->last_sep_point->dir,
                                    sep1->last_sep_point->type,VFResults.config_intpoints,&points);
                }
                else
                {
                    sep1->first_sep_point=plot_separatrice(spherewnd,point->x0,point->y0,
                                    point->a11,point->a12,point->a21,point->a22,point->epsilon,
                                    sep1,&points,point->chart,vfindex);
                }
                sep1->last_sep_point=points;
                sep1=sep1->next_sep;
            }
        }
    point=point->next_saddle;
    }
}

void change_epsilon_saddle( QWinSphere * spherewnd, double epsilon)
{
    struct sep *separatrice;
    
    VFResults.selected_saddle_point->epsilon=epsilon;
    separatrice=VFResults.selected_saddle_point->separatrices;

    while( separatrice != NULL )
    {
        draw_selected_sep(spherewnd,separatrice->first_sep_point,CBACKGROUND);
        DeleteOrbitPoint(separatrice->first_sep_point);
        separatrice->last_sep_point=NULL;
        separatrice->first_sep_point=NULL;
        separatrice=separatrice->next_sep;
    }
}

