#include <qobject.h>
#include <math.h>
#include "custom.h"
#include "table.h"
#include "math_p4.h"
#include "file_vf.h"

void (* change_epsilon)( QWinSphere *, double ) = NULL;
void (*start_plot_sep)( QWinSphere *, int ) = NULL;
void (*cont_plot_sep)( QWinSphere * ) = NULL;
void (*plot_next_sep)( QWinSphere *, int ) = NULL;
void (*select_next_sep)( QWinSphere * ) = NULL;

extern bool PrepareVfForIntegration( double * );
extern void plot_all_saddle_sep(QWinSphere * spherewnd, int vfindex, struct saddle *point);
extern void plot_all_se_sep(QWinSphere * spherewnd, int vfindex, struct semi_elementary *point);
extern void plot_all_de_sep( QWinSphere * spherewnd, int vfindex, struct degenerate *point );

// ---------------------------------------------------------------------------
//                  FINDSEPCOLOR
//                  FINDSEPCOLOR3
// ---------------------------------------------------------------------------
//
// The stability of separatrices is found for the reduced vector field.  In case
// a greatest common factor is present, we have to take into account the sign
// of this common factor to determine the plot color of the separatrice.
//
// The point y is a point in one of the charts.

int FindSepColor( struct term2 * f, int type, double * y )
{
    int color;

    if( eval_term2(f,y) >= 0 )
    {
        switch(type)
        {
            case OT_STABLE:         color = CSTABLE;        break;
            case OT_UNSTABLE:       color = CUNSTABLE;      break;
            case OT_CENT_STABLE:    color = CCENT_STABLE;   break;
            case OT_CENT_UNSTABLE:  color = CCENT_UNSTABLE; break;
            default:                color = 0;              break; // to avoid compiler warnings
        }
    }
    else
    {
        switch(type)
        {
            case OT_STABLE:         color = CUNSTABLE;      break;
            case OT_UNSTABLE:       color = CSTABLE;        break;
            case OT_CENT_STABLE:    color = CCENT_UNSTABLE; break;
            case OT_CENT_UNSTABLE:  color = CCENT_STABLE;   break;
            default:                color = 0;              break; // to avoid compiler warnings
        }
    }
    return(color);
}

int FindSepColor3( struct term3 * f, int type, double * y ) // y = (r,theta)
{
    int color;

    if( eval_term3(f,y) >= 0 )
    {
        switch(type)
        {
            case OT_STABLE:         color = CSTABLE;        break;
            case OT_UNSTABLE:       color = CUNSTABLE;      break;
            case OT_CENT_STABLE:    color = CCENT_STABLE;   break;
            case OT_CENT_UNSTABLE:  color = CCENT_UNSTABLE; break;
            default:                color = 0;              break;
        }
    }
    else
    {
        switch(type)
        {
            case OT_STABLE:         color = CUNSTABLE;      break;
            case OT_UNSTABLE:       color = CSTABLE;        break;
            case OT_CENT_STABLE:    color = CCENT_UNSTABLE; break;
            case OT_CENT_UNSTABLE:  color = CCENT_STABLE;   break;
            default:                color = 0;              break;
        }
    }
    return(color);
}

// ---------------------------------------------------------------------------
//                  CHANGE_TYPE
// ---------------------------------------------------------------------------
//
// Sometimes, when a separatrix is integrated close to the line of infinity,
// it is possible that the line of infinity is crossed (only when p=q=1).
// By crossing the line of singularity, the separatrix may change its stability
// (depending on the degree of the vector field).

int change_type(int type)
{
    int t;

    switch(type)
    {
    case OT_STABLE:         t = OT_UNSTABLE;        break;
    case OT_UNSTABLE:       t = OT_STABLE;          break;
    case OT_CENT_STABLE:    t = OT_CENT_UNSTABLE;   break;
    case OT_CENT_UNSTABLE:  t = OT_CENT_STABLE;     break;
    case OT_ORBIT:          t = OT_ORBIT;           break;
    default:                t = type;               break;
    }
    return t;
}

// ---------------------------------------------------------------------------
//                  INTEGRATE_POINCARE_SEP
// ---------------------------------------------------------------------------
//
// Integrate Separatrix on the Poincare Sphere, i.e. in case p=q=1
// This routine calcuates 1 integration step.
//
void integrate_poincare_sep( double p0, double p1, double p2, double * pcoord,
                            double * hhi, int * type, int * color, int * dashes, int * dir,
                            double h_min, double h_max)
{
    double y[2];
    double theta;
    double hhi0;

    if(pcoord[2]>ZCOORD)
    {
        hhi0 = *hhi;
        while( 1 )
        {
            *dashes=true;
            *dir=1;
            psphere_to_R2(p0,p1,p2,y);
            rk78(eval_r_vec_field,y,hhi,h_min,h_max,VFResults.config_tolerance);
            if( ThisVF->GetVFIndex_R2( y ) == VFResults.K )
                break;
            h_min = VFResults.config_branchhmi;
            h_max /= 2;
            hhi0 = fabs(hhi0*(*hhi))/2/(*hhi);
            if( fabs(hhi0) < h_min || h_max < VFResults.config_branchhmi )
            {
                hhi0 = h_min;
                if( *hhi<0 ) hhi0=-hhi0;
                break;
            }
        }
        *hhi = hhi0;
        R2_to_psphere(y[0],y[1],pcoord);
        *color=FindSepColor(VFResultsK->gcf,*type,y);
    }
    else
    {
        // infinite region (annulus)

        theta=atan2(fabs(p1),fabs(p0));
        if((theta<PI_DIV4) && (theta>-PI_DIV4))
        {
            if(p0>0)
            {
                hhi0 = *hhi;
                while( 1 )
                {
                    *dashes=true;
                    *dir=1;
                    psphere_to_U1(p0,p1,p2,y);
                    rk78(eval_U1_vec_field,y,hhi,h_min,h_max,VFResults.config_tolerance);
                    if(y[1]>=0 || !VFResultsK->singinf)
                    {
                        if( ThisVF->GetVFIndex_U1( y ) == VFResults.K )
                            break;
                    }
                    else
                    {
                        if( ThisVF->GetVFIndex_VV1( y ) == VFResults.K )
                            break;
                    }
                    h_min = VFResults.config_branchhmi;
                    h_max /= 2;
                    hhi0 = fabs(hhi0*(*hhi))/2/(*hhi);
                    if( fabs(hhi0) < h_min || h_max < VFResults.config_branchhmi )
                    {
                        hhi0 = h_min;
                        if( *hhi<0 ) hhi0=-hhi0;
                        break;
                    }
                }
                *hhi = hhi0;
                if(y[1]>=0 || !VFResultsK->singinf)
                {
                    U1_to_psphere(y[0],y[1],pcoord);
                    *color=FindSepColor(VFResultsK->gcf_U1,*type,y);
                }
                else
                {
                    VV1_to_psphere(y[0],y[1],pcoord);
                    if(VFResultsK->dir_vec_field==1)
                    {
                        *dir=-1; *hhi=-(*hhi);
                        *type=change_type(*type);
                    }
                    psphere_to_V1(pcoord[0],pcoord[1],pcoord[2],y);
                    *color=FindSepColor(VFResultsK->gcf_V1,*type,y);
                    *dashes=false;
                }
            }
            else
            {
                hhi0 = *hhi;
                while( 1 )
                {
                    *dashes=true;
                    *dir=1;
                    psphere_to_V1(p0,p1,p2,y);
                    rk78(eval_V1_vec_field,y,hhi,h_min,h_max,VFResults.config_tolerance);
                    if(y[1]>=0 || !VFResultsK->singinf)
                    {
                        if( ThisVF->GetVFIndex_V1( y ) == VFResults.K )
                            break;
                    }
                    else
                    {
                        if( ThisVF->GetVFIndex_UU1( y ) == VFResults.K )
                            break;
                    }
                    h_min = VFResults.config_branchhmi;
                    h_max /= 2;
                    hhi0 = fabs(hhi0*(*hhi))/2/(*hhi);
                    if( fabs(hhi0) < h_min || h_max < VFResults.config_branchhmi )
                    {
                        hhi0 = h_min;
                        if( *hhi<0 ) hhi0=-hhi0;
                        break;
                    }
                }
                *hhi = hhi0;
                if(y[1]>=0 || !VFResultsK->singinf)
                {
                    V1_to_psphere(y[0],y[1],pcoord);
                    *color=FindSepColor(VFResultsK->gcf_V1,*type,y);
                }
                else
                {
                    UU1_to_psphere(y[0],y[1],pcoord);
                    if(VFResultsK->dir_vec_field==1)
                    {
                        *dir=-1; *hhi=-(*hhi);
                        *type=change_type(*type);
                    }
                    psphere_to_U1(pcoord[0],pcoord[1],pcoord[2],y);
                    *color=FindSepColor(VFResultsK->gcf_U1,*type,y);
                    *dashes=false;
                }
            }
        }
        else
        {
            if(p1>0)
            {
                hhi0 = *hhi;
                while( 1 )
                {
                    *dashes=true;
                    *dir=1;
                    psphere_to_U2(p0,p1,p2,y);
                    rk78(eval_U2_vec_field,y,hhi,h_min,h_max,VFResults.config_tolerance);
                    if(y[1]>=0 || !VFResultsK->singinf)
                    {
                        if( ThisVF->GetVFIndex_U2( y ) == VFResults.K )
                            break;
                    }
                    else
                    {
                        if( ThisVF->GetVFIndex_VV2( y ) == VFResults.K )
                            break;
                    }
                    h_min = VFResults.config_branchhmi;
                    h_max /= 2;
                    hhi0 = fabs(hhi0*(*hhi))/2/(*hhi);
                    if( fabs(hhi0) < h_min || h_max < VFResults.config_branchhmi )
                    {
                        hhi0 = h_min;
                        if( *hhi<0 ) hhi0=-hhi0;
                        break;
                    }
                }
                *hhi = hhi0;
                if(y[1]>=0 || !VFResultsK->singinf)
                {
                    U2_to_psphere(y[0],y[1],pcoord);
                    *color=FindSepColor(VFResultsK->gcf_U2,*type,y);
                }
                else
                {
                    VV2_to_psphere(y[0],y[1],pcoord);
                    if(VFResultsK->dir_vec_field==1)
                    {
                        *dir=-1;*hhi=-(*hhi);
                        *type=change_type(*type);
                    }
                    psphere_to_V2(pcoord[0],pcoord[1],pcoord[2],y);
                    *color=FindSepColor(VFResultsK->gcf_V2,*type,y);
                    *dashes=false;
                }
            }
            else
            {
                hhi0 = *hhi;
                while( 1 )
                {
                    *dashes=true;
                    *dir=1;
                    psphere_to_V2(p0,p1,p2,y);
                    rk78(eval_V2_vec_field,y,hhi,h_min,h_max,VFResults.config_tolerance);
                    if( y[1]>=0 || !VFResultsK->singinf)
                    {
                        if( ThisVF->GetVFIndex_V2( y ) == VFResults.K )
                            break;
                    }
                    else
                    {
                        if( ThisVF->GetVFIndex_UU2( y ) == VFResults.K )
                            break;
                    }
                    h_min = VFResults.config_branchhmi;
                    h_max /= 2;
                    hhi0 = fabs(hhi0*(*hhi))/2/(*hhi);
                    if( fabs(hhi0) < h_min || h_max < VFResults.config_branchhmi )
                    {
                        hhi0 = h_min;
                        if( *hhi<0 ) hhi0=-hhi0;
                        break;
                    }
                }
                *hhi = hhi0;
                if(y[1]>=0 || !VFResultsK->singinf)
                {
                    V2_to_psphere(y[0],y[1],pcoord);
                    *color=FindSepColor(VFResultsK->gcf_V2,*type,y);
                }
                else
                {
                    UU2_to_psphere(y[0],y[1],pcoord);
                    if(VFResultsK->dir_vec_field==1)
                    {
                        *dir=-1;*hhi=-(*hhi);
                        *type=change_type(*type);
                    }
                    psphere_to_U2(pcoord[0],pcoord[1],pcoord[2],y);
                    *color=FindSepColor(VFResultsK->gcf_U2,*type,y);
                    *dashes=false;
                }
            }
        }
    }
}

// ---------------------------------------------------------------------------
//                  INTEGRATE_LYAPUNOV_SEP
// ---------------------------------------------------------------------------
//
// Integrate Separatrix on the Poincare-Lyapunov sphere
// This routine calcuates 1 integration step.

void integrate_lyapunov_sep( double p0, double p1, double p2, double * pcoord,
                            double * hhi, int * type, int * color, int * dashes,
                            int * dir, double h_min, double h_max)
{
    double y[2];
    double hhi0;

    if( p0 == 0 )
    {
        hhi0 = *hhi;
        while( 1 )
        {
            *dashes=true;
            *dir=1;
            y[0]=p1;
            y[1]=p2;
            rk78(eval_r_vec_field,y,hhi,h_min,h_max,VFResults.config_tolerance);
            if( ThisVF->GetVFIndex_R2( y ) == VFResults.K )
                break;
            h_min = VFResults.config_branchhmi;
            h_max /= 2;
            hhi0 = fabs(hhi0*(*hhi))/2/(*hhi);
            if( fabs(hhi0) < h_min || h_max < VFResults.config_branchhmi )
            {
                hhi0 = h_min;
                if( *hhi<0 ) hhi0=-hhi0;
                break;
            }
        }    
        *hhi = hhi0;
        R2_to_plsphere(y[0],y[1],pcoord);
        *color=FindSepColor(VFResultsK->gcf,*type,y);
    }
    else
    {
        hhi0 = *hhi;
        while( 1 )
        {
            *dashes=true;
            *dir=1;
            y[0]=p1;
            y[1]=p2;
            rk78(eval_vec_field_cyl,y,hhi,h_min,h_max,VFResults.config_tolerance);
            if(y[1]>=TWOPI) y[1] -= TWOPI;
            if( ThisVF->GetVFIndex_cyl( y ) == VFResults.K )
                break;
            h_min = VFResults.config_branchhmi;
            h_max /= 2;
            hhi0 = fabs(hhi0*(*hhi))/2/(*hhi);
            if( fabs(hhi0) < h_min || h_max < VFResults.config_branchhmi )
            {
                hhi0 = h_min;
                if( *hhi<0 ) hhi0=-hhi0;
                break;
            }
        }
        *hhi = hhi0;
        cylinder_to_plsphere(y[0],y[1],pcoord);
        *color=FindSepColor3(VFResultsK->gcf_C,*type,y);
    }
}

// ---------------------------------------------------------------------------
//                  INTEGRATE_SEP
// ---------------------------------------------------------------------------
//
// The last integration point is stored in *orbit, the first integration point
// is returned.  In case no vector field is represented, NULL is returned
// and *orbit will not be stored.
//
// The vector field vfK need not be prepared

struct orbits_points *integrate_sep( QWinSphere * spherewnd, double pcoord[3], double step,
                                           int dir, int type, int points_to_int, struct orbits_points **orbit )
{
    int i,d,h;
    int color,dashes;
    double hhi;
    double pcoord2[3],h_min,h_max;
    struct orbits_points *first_orbit;
    struct orbits_points *last_orbit;
    
    first_orbit = NULL;
    last_orbit = NULL;

    /* if we integrate a separatrice and use the original vector field
    then it is possible that we have to change the direction of the
    separatrice, because the separatrice is evaluate for the reduced
    vector field
    */

    if( !PrepareVfForIntegration( pcoord ) )
        return NULL;

    if( VFResults.config_kindvf == INTCONFIG_ORIGINAL && MATHFUNC(change_dir)(pcoord))
        hhi = - VFResults.config_step * dir;
    else
        hhi=(double)dir*step;

    h_min=VFResults.config_hmi;
    h_max=VFResults.config_hma;
    copy_x_into_y(pcoord,pcoord2);
    for(i=1;i<=points_to_int;++i)
    {
        MATHFUNC(integrate_sphere_sep)(pcoord[0],pcoord[1],pcoord[2],pcoord,
                        &hhi,&type,&color,&dashes,&d,h_min,h_max);

        if( (i % UPDATEFREQ_STEPSIZE) == 0 )
        {
            set_current_step(fabs(hhi));
        }
        if( last_orbit == NULL )
        {
            first_orbit = (struct orbits_points *)malloc( sizeof(struct orbits_points) );
            last_orbit = first_orbit;
            h=dir;
        }
        else
        {
            last_orbit->nextpt = (struct orbits_points *)malloc( sizeof (struct orbits_points) );
            h=last_orbit->dir;
            last_orbit=last_orbit->nextpt;
        }

        copy_x_into_y( pcoord, last_orbit->pcoord );
        last_orbit->color = color;
        last_orbit->dashes = dashes * VFResults.config_dashes;
        last_orbit->dir = d*h;
        last_orbit->type = type;
        last_orbit->nextpt = NULL;

        if( dashes * VFResults.config_dashes )
            (*plot_l)( spherewnd, pcoord, pcoord2, color );
        else
            (*plot_p)( spherewnd, pcoord, color );
            
        copy_x_into_y( pcoord, pcoord2 );
        if( !PrepareVfForIntegration( pcoord ) )
            break;
    }
    set_current_step(fabs(hhi));
    if( first_orbit != NULL )
        *orbit = last_orbit;
    return first_orbit;
}

// ---------------------------------------------------------------------------
//                  FINDINITIALSEPPOINT
// ---------------------------------------------------------------------------
//
// Given a parametrized curve (t,sep(t)) with sep(0)=0.  The parametrized curve
// is in fact a Taylor polynomial of a separatrix.
//
// Given a value of epsilon.
//
// We find a value of t such that Norm2( t,sep(t) ) is approximately given by epsilon.
// (Norm2 is the standard Euclidean norm)
//
// More precisely we find a t such that the Norm2^2 lies in a 1% - error interval from epsilon^2.
//

static double FindInitialSepPoint( struct term1 * sep, double epsilon, int dir)
{
    double t,t1,t2,r0,a,b;
    
    t = epsilon*dir;
    a = pow( epsilon-epsilon/100, 2.0 );
    b = pow( epsilon+epsilon/100, 2.0 );
    r0 = pow(t,2.0) + pow( eval_term1(sep,t), 2.0 );
    
    if( r0>a && r0<b )
        return t;
        
    if( r0 <=a )
    {
        while(r0<=a)
        {
            t*=2.;
            r0=pow(t,2.0) + pow(eval_term1(sep,t),2.0);
        }
        if(r0<b)
            t2=t;
        else
        {
            t1=t/2;
            t2=(t+t1)/2;
            for(;;)
            {
                r0=pow(t2,2.0)+pow(eval_term1(sep,t2),2.0);
                if((r0>a) && (r0<b))break;
                if(r0<=a)t1=t2;
                else t=t2;
                t2=(t+t1)/2;
            }
        }
    }
    else
    {
        while( r0>=b )
        {
            t/=2.;
            r0=pow(t,2.0)+pow(eval_term1(sep,t),2.0);
        }
        if(r0>a)
            t2=t;
        else
        {
            t1=2.*t;
            t2=(t+t1)/2;
            for(;;)
            {
                r0 = pow(t2,2.0)+pow(eval_term1(sep,t2),2.0);
                if( r0>a && r0<b )
                    break;
                if(r0<=a)
                    t=t2;
                else
                    t1=t2;
                t2=(t+t1)/2;
            }
        }
    }
    return t2;
}

// ---------------------------------------------------------------------------
//                  PLOT_SEPARATRICE
// ---------------------------------------------------------------------------
//
// Starts separatrix integration, and plots the results meanwhile to the screen
//
// The first 100 points are drawn by looking at the formal power series development
// of the separatrix, as calculated by maple.  These points are drawn in an
// epsilon-ball around the singularity, where epsilon can be changed by the user.
//
// At the end, a normal integration cycle is added.
//
// The vector field vfK needs not be prepared.

struct orbits_points * plot_separatrice( QWinSphere * spherewnd, double x0, double y0,
                            double a11, double a12, double a21, double a22,
                            double epsilon, struct sep * sep1, struct orbits_points ** orbit, int chart, int vfindex )
{
    double t, h, y;
    double pcoord[3];
    double pcoord2[3];
    double point[2];
    struct orbits_points * first_orbit;
    struct orbits_points * last_orbit;
//    struct orbits_points * sep2;
    int color, dir, type, i;
    bool dashes;
    bool ok;
    bool prepok;
    
    ok=true;
    first_orbit=NULL;
    last_orbit=NULL;
    t=0.0;

    switch( chart )
    {
        case CHART_R2: MATHFUNC(R2_to_sphere)(x0,y0,pcoord); break;
        case CHART_U1: MATHFUNC(U1_to_sphere)(x0,y0,pcoord); break;
        case CHART_V1: MATHFUNC(V1_to_sphere)(x0,y0,pcoord); break;
        case CHART_U2: MATHFUNC(U2_to_sphere)(x0,y0,pcoord); break;
        case CHART_V2: MATHFUNC(V2_to_sphere)(x0,y0,pcoord); break;
    }
    if( !PrepareVfForIntegration( pcoord ) )
        return NULL;

    // if we have a line of singularities at infinity then we have to change the chart if the chart is V1 or V2

    if( VFResultsK->singinf )
    {
        if( chart==CHART_V1 ) chart = CHART_U1;
        else if( chart==CHART_V2 ) chart = CHART_U2;
    }

    // h=(epsilon/100)*sep1->direction;
    
    h = FindInitialSepPoint( sep1->separatrice, epsilon, sep1->direction ) / 100;

    point[0] = x0;
    point[1] = y0;
    
    switch( chart )
    {
        case CHART_R2:
                MATHFUNC(R2_to_sphere)(x0,y0,pcoord);
                color=FindSepColor(VFResultsK->gcf,sep1->type,point);
                break;
        case CHART_U1:
                MATHFUNC(U1_to_sphere)(x0,y0,pcoord);
                color=FindSepColor(VFResultsK->gcf_U1,sep1->type,point);
                break;
        case CHART_V1:
                MATHFUNC(V1_to_sphere)(x0,y0,pcoord);
                if( !VFResults.plweights )
                    psphere_to_V1(pcoord[0],pcoord[1],pcoord[2],point);
                color = FindSepColor(VFResultsK->gcf_V1,sep1->type,point);
                break;
        case CHART_U2:
                MATHFUNC(U2_to_sphere)(x0,y0,pcoord);
                color=FindSepColor(VFResultsK->gcf_U2,sep1->type,point);
                break;
        case CHART_V2:
                MATHFUNC(V2_to_sphere)(x0,y0,pcoord);
                if( !VFResults.plweights )
                    psphere_to_V2(pcoord[0],pcoord[1],pcoord[2],point);
                color = FindSepColor(VFResultsK->gcf_U2,sep1->type,point);
                break;
        default:
                color=0;
                break;
    }

    first_orbit = (struct orbits_points *)malloc( sizeof(struct orbits_points) );
    last_orbit = first_orbit;
    copy_x_into_y( pcoord, last_orbit->pcoord );
    type = sep1->type;
    switch( sep1->type )
    {
        case OT_STABLE:         dir=OT_STABLE;      break;
        case OT_UNSTABLE:       dir=OT_UNSTABLE;    break;
        case OT_CENT_STABLE:    dir=OT_STABLE;      break;
        case OT_CENT_UNSTABLE:  dir=OT_UNSTABLE;    break;
        default:                dir = 0;            break;
    }

    last_orbit->color = color;
    last_orbit->dashes = 0;

/*
    P5 Addition: Check if one can start with this separatrix.
*/

    y = eval_term1(sep1->separatrice,100*h);
    if( sep1->d )
    {
        point[0]=x0+a11*y+a12*100*h;
        point[1]=y0+a21*y+a22*100*h;
    }
    else
    {
        point[0]=x0+a11*100*h+a12*y;
        point[1]=y0+a21*100*h+a22*y;
    }
    switch(chart)
    {
        case CHART_R2:
            MATHFUNC(R2_to_sphere)(point[0],point[1],pcoord2);
            break;
        case CHART_U1:
            if( point[1]>=0 || !VFResultsK->singinf )
                MATHFUNC(U1_to_sphere)(point[0],point[1],pcoord2);
            else
                VV1_to_psphere(point[0],point[1],pcoord2);
            break;
        case CHART_V1:
            MATHFUNC(V1_to_sphere)(point[0],point[1],pcoord2);
            break;
        case CHART_U2:
            if(point[1]>=0 || !VFResultsK->singinf)
                MATHFUNC(U2_to_sphere)(point[0],point[1],pcoord2);
            else
                VV2_to_psphere(point[0],point[1],pcoord2);
            break;
        case CHART_V2:
            MATHFUNC(V2_to_sphere)(point[0],point[1],pcoord2);
            break;
    }
    if( ThisVF->GetVFIndex_sphere( pcoord2 ) != vfindex )
    {
        free( first_orbit );
        return NULL;
    }

/*
    end ofP5 Addition
*/

    copy_x_into_y( pcoord, pcoord2 );
    for( i=0; i <= 99; i++ )
    {
        t = t+h;
        y = eval_term1(sep1->separatrice,t);
        if( (t*t+y*y) > epsilon*epsilon )
            break;
        
        if( sep1->d )
        {
            point[0]=x0+a11*y+a12*t;
            point[1]=y0+a21*y+a22*t;
        }
        else
        {
            point[0]=x0+a11*t+a12*y;
            point[1]=y0+a21*t+a22*y;
        }
        dashes=true;
        switch(chart)
        {
            case CHART_R2:
                MATHFUNC(R2_to_sphere)(point[0],point[1],pcoord);
                prepok=PrepareVfForIntegration( pcoord );
                if( !prepok )
                    break;
                color=FindSepColor(VFResultsK->gcf,sep1->type,point);
                break;
            case CHART_U1:
                if( point[1]>=0 || !VFResultsK->singinf )
                {
                    MATHFUNC(U1_to_sphere)(point[0],point[1],pcoord);
                    prepok=PrepareVfForIntegration( pcoord );
                    if( !prepok )
                        break;
                    if(!ok)
                    {
                        dashes=false;
                        ok=true;
                        if( VFResultsK->dir_vec_field ==1 )
                            dir*=-1;
                    }
                    type=sep1->type;
                    color=FindSepColor(VFResultsK->gcf_U1,type,point);
                }
                else
                {
                    VV1_to_psphere(point[0],point[1],pcoord);
                    prepok=PrepareVfForIntegration( pcoord );
                    if( !prepok )
                        break;
                    if(ok)
                    {
                        dashes=false;
                        ok=false;
                        if( VFResultsK->dir_vec_field == 1 )
                            dir*=-1;
                    }
                    psphere_to_V1(pcoord[0],pcoord[1],pcoord[2],point);
                    prepok=PrepareVfForIntegration( pcoord );
                    if( !prepok )
                        break;
                    if(VFResultsK->dir_vec_field==1)
                        type=change_type(sep1->type);
                    else
                        type=sep1->type;
                    color = FindSepColor(VFResultsK->gcf_V1,type,point);
                }
                break;
            case CHART_V1:
                MATHFUNC(V1_to_sphere)(point[0],point[1],pcoord);
                prepok=PrepareVfForIntegration( pcoord );
                if( !prepok )
                    break;
                if( !VFResults.plweights )
                    psphere_to_V1(pcoord[0],pcoord[1],pcoord[2],point);
                color = FindSepColor(VFResultsK->gcf_V1,sep1->type,point);
                break;
            case CHART_U2:
                if(point[1]>=0 || !VFResultsK->singinf)
                {
                    MATHFUNC(U2_to_sphere)(point[0],point[1],pcoord);
                    prepok=PrepareVfForIntegration( pcoord );
                    if( !prepok )
                        break;
                    if(!ok)
                    {
                        dashes=false;
                        ok=true;
                        if(VFResultsK->dir_vec_field==1) dir*=-1;
                    }
                    type=sep1->type;
                    color=FindSepColor(VFResultsK->gcf_U2,type,point);
                }
                else
                {
                    VV2_to_psphere(point[0],point[1],pcoord);
                    prepok=PrepareVfForIntegration( pcoord );
                    if( !prepok )
                        break;
                    if(ok)
                    {
                        dashes=false;
                        ok=false;
                        if(VFResultsK->dir_vec_field==1)dir*=-1;
                    }
                    psphere_to_V2(pcoord[0],pcoord[1],pcoord[2],point);
                    if(VFResultsK->dir_vec_field==1)
                        type=change_type(sep1->type);
                    else
                        type=sep1->type;
                    color=FindSepColor(VFResultsK->gcf_V2,type,point);
                }
                break;
            case CHART_V2:
                MATHFUNC(V2_to_sphere)(point[0],point[1],pcoord);
                prepok=PrepareVfForIntegration( pcoord );
                if( !prepok )
                    break;
                if( !VFResults.plweights )
                    psphere_to_V2(pcoord[0],pcoord[1],pcoord[2],point);
                color = FindSepColor(VFResultsK->gcf_V2,sep1->type,point);
                break;
            default:
                prepok=false;
                break;
        }

        last_orbit->nextpt = (struct orbits_points *)malloc( sizeof(struct orbits_points) );
        last_orbit = last_orbit->nextpt;
        copy_x_into_y( pcoord, last_orbit->pcoord );
        last_orbit->color = color;
        last_orbit->dashes = dashes * VFResults.config_dashes;
        last_orbit->dir = dir;
        last_orbit->type = type;
        if(last_orbit->dashes)
        (*plot_l)(spherewnd,pcoord,pcoord2,color);
        else
        (*plot_p)(spherewnd,pcoord,color);
        copy_x_into_y(pcoord,pcoord2);
        if( !prepok )
            break;
    }

    *orbit = last_orbit;
    last_orbit->nextpt = integrate_sep( spherewnd, pcoord, VFResults.config_step, last_orbit->dir, type, VFResults.config_intpoints, orbit );
    return first_orbit;
}

// ---------------------------------------------------------------------------
//                  PLOT_ALL_SEP
// ---------------------------------------------------------------------------
//
// Plots all separatrices.  If the separatrix plotting has not started yet,
// it will be started; otherwhise it will be continued.

void plot_all_sep( QWinSphere * spherewnd )
{
    int i;
    if( VFResults.vf != NULL )
    {
        for( i = 0; i < ThisVF->numvf; i++ )
        {
            plot_all_saddle_sep( spherewnd, i, VFResults.vf[i]->first_saddle_point );
            plot_all_se_sep( spherewnd, i, VFResults.vf[i]->first_se_point );
            plot_all_de_sep( spherewnd, i, VFResults.vf[i]->first_de_point );
        }
    }
}

// ---------------------------------------------------------------------------
//                  DRAW_SEP
// ---------------------------------------------------------------------------
//
// Does the plotting of a separatrix that was previously calculated.
// The separatrix is plotted in the color according to the type and stability.

void draw_sep( QWinSphere * spherewnd, struct orbits_points * sep )
{
    double pcoord[3];
    
    while( sep != NULL )
    {
        if(sep->dashes)
            (*plot_l)(spherewnd, sep->pcoord,pcoord,sep->color);
        else
            (*plot_p)(spherewnd, sep->pcoord,sep->color);

        copy_x_into_y(sep->pcoord,pcoord);
        sep=sep->nextpt;
    }
}

// ---------------------------------------------------------------------------
//                  DRAW_SELECTED_SEP
// ---------------------------------------------------------------------------
//
// Does the plotting of a separatrix that was previously calculated.
// The separatrix is plotted in a specified color.

void draw_selected_sep( QWinSphere * spherewnd, struct orbits_points * sep, int color )
{
    double pcoord[3];
    
    while( sep != NULL )
    {
        if( sep->dashes )
            (*plot_l)(spherewnd, sep->pcoord,pcoord,color);
        else
            (*plot_p)(spherewnd, sep->pcoord,color);
    
        copy_x_into_y(sep->pcoord,pcoord);
        sep=sep->nextpt;
    }    
}

