#include <qobject.h>
#include "custom.h"
#include "table.h"
#include "win_p5.h"
#include "math_p4.h"
#include "win_sphere.h"
#include "file_vf.h"

extern int PrintColorTable[NUMXFIGCOLORS];

void (* print_saddle)( double, double ) = NULL;
void (* print_stablenode)( double, double ) = NULL;
void (* print_unstablenode)( double, double ) = NULL;
void (* print_stableweakfocus)( double, double ) = NULL;
void (* print_unstableweakfocus)( double, double ) = NULL;
void (* print_weakfocus)( double, double ) = NULL;
void (* print_stablestrongfocus)( double, double ) = NULL;
void (* print_unstablestrongfocus)( double, double ) = NULL;
void (* print_sesaddle)( double, double ) = NULL;
void (* print_sesaddlenode)( double, double ) = NULL;
void (* print_sestablenode)( double, double ) = NULL;
void (* print_seunstablenode)( double, double ) = NULL;
void (* print_degen)( double, double ) = NULL;
void (* print_center)( double, double ) = NULL;
void (* print_elips)( double, double, double, double, int, bool, struct P4POLYLINES * ) = NULL;
void (* print_line)( double, double, double, double, int ) = NULL;
void (* print_point)( double, double, int ) = NULL;
void (* print_comment)( QString ) = NULL;
void (* print_virtualsaddle)( double, double ) = NULL;
void (* print_virtualstablenode)( double, double ) = NULL;
void (* print_virtualunstablenode)( double, double ) = NULL;
void (* print_virtualstableweakfocus)( double, double ) = NULL;
void (* print_virtualunstableweakfocus)( double, double ) = NULL;
void (* print_virtualweakfocus)( double, double ) = NULL;
void (* print_virtualstablestrongfocus)( double, double ) = NULL;
void (* print_virtualunstablestrongfocus)( double, double ) = NULL;
void (* print_virtualsesaddle)( double, double ) = NULL;
void (* print_virtualsesaddlenode)( double, double ) = NULL;
void (* print_virtualsestablenode)( double, double ) = NULL;
void (* print_virtualseunstablenode)( double, double ) = NULL;
void (* print_virtualdegen)( double, double ) = NULL;
void (* print_virtualcenter)( double, double ) = NULL;
void (* print_coinciding)( double, double ) = NULL;

// -----------------------------------------------------------------------
//                          PLOT SINGULAR POINTS
// -----------------------------------------------------------------------

void QWinSphere::printPoint( struct saddle * p )
{
    double pos[2];
    if( p != NULL )
    {
        getChartPos( p->chart, p->x0, p->y0, pos );
        if( pos[0] < x0 || pos[0] > x1 || pos[1] < y0 || pos[1] > y1 )
            return;

        switch( p->position )
        {
        case POSITION_VIRTUAL:  print_virtualsaddle( coWinX(pos[0]), coWinY(pos[1]) ); break;
        case POSITION_COINCIDING: break;
        case POSITION_COINCIDING_VIRTUAL: break;
        case POSITION_COINCIDING_MAIN:
                                print_coinciding( coWinX(pos[0]), coWinY(pos[1]) ); break;
        default:                print_saddle( coWinX(pos[0]), coWinY(pos[1]) ); break;
        }
    }
}

void QWinSphere::printPoint( struct node * p )
{
    double pos[2];

    if( p != NULL )
    {
        getChartPos( p->chart, p->x0, p->y0, pos );
        
        if( pos[0] < x0 || pos[0] > x1 || pos[1] < y0 || pos[1] > y1 )
            return;

        if( p->stable == -1 )
        {
            switch( p->position )
            {
            case POSITION_VIRTUAL:  print_virtualstablenode( coWinX( pos[0] ), coWinY( pos[1] ) ); break;
            case POSITION_COINCIDING: break;
            case POSITION_COINCIDING_VIRTUAL: break;
            case POSITION_COINCIDING_MAIN:
                                    print_coinciding( coWinX(pos[0]), coWinY(pos[1]) ); break;
            default:                print_stablenode( coWinX( pos[0] ), coWinY( pos[1] ) ); break;
            }
        }
        else
        {
            switch( p->position )
            {
            case POSITION_VIRTUAL:  print_virtualunstablenode( coWinX( pos[0] ), coWinY( pos[1] ) ); break;
            case POSITION_COINCIDING: break;
            case POSITION_COINCIDING_VIRTUAL: break;
            case POSITION_COINCIDING_MAIN:
                                    print_coinciding( coWinX(pos[0]), coWinY(pos[1]) ); break;
            default:                print_unstablenode( coWinX( pos[0] ), coWinY( pos[1] ) ); break;
            }
        }
    }
}

void QWinSphere::printPoint( struct weak_focus * p )
{
    double pos[2];

    if( p != NULL )
    {
        getChartPos( p->chart, p->x0, p->y0, pos );

        if( pos[0] < x0 || pos[0] > x1 || pos[1] < y0 || pos[1] > y1 )
            return;

        switch( p->type )
        {
        case SINGTYPE_STABLE:
            switch( p->position )
            {
            case POSITION_VIRTUAL:  print_virtualstableweakfocus( coWinX( pos[0] ), coWinY( pos[1] ) ); break;
            case POSITION_COINCIDING: break;
            case POSITION_COINCIDING_VIRTUAL: break;
            case POSITION_COINCIDING_MAIN:
                                    print_coinciding( coWinX(pos[0]), coWinY(pos[1]) ); break;
            default:                print_stableweakfocus( coWinX( pos[0] ), coWinY( pos[1] ) ); break;
            }            
            break;
        case SINGTYPE_UNSTABLE:
            switch( p->position )
            {
            case POSITION_VIRTUAL:  print_virtualunstableweakfocus( coWinX( pos[0] ), coWinY( pos[1] ) ); break;
            case POSITION_COINCIDING: break;
            case POSITION_COINCIDING_VIRTUAL: break;
            case POSITION_COINCIDING_MAIN:
                                    print_coinciding( coWinX(pos[0]), coWinY(pos[1]) ); break;
            default:                print_unstableweakfocus( coWinX( pos[0] ), coWinY( pos[1] ) ); break;
            }            
            break;
        case SINGTYPE_CENTER:
            switch( p->position )
            {
            case POSITION_VIRTUAL:  print_virtualcenter( coWinX( pos[0] ), coWinY( pos[1] ) ); break;
            case POSITION_COINCIDING: break;
            case POSITION_COINCIDING_VIRTUAL: break;
            case POSITION_COINCIDING_MAIN:
                                    print_coinciding( coWinX(pos[0]), coWinY(pos[1]) ); break;
            default:                print_center( coWinX( pos[0] ), coWinY( pos[1] ) ); break;
            }            
            break;
        default:
            switch( p->position )
            {
            case POSITION_VIRTUAL:  print_virtualweakfocus( coWinX( pos[0] ), coWinY( pos[1] ) ); break;
            case POSITION_COINCIDING: break;
            case POSITION_COINCIDING_VIRTUAL: break;
            case POSITION_COINCIDING_MAIN:
                                    print_coinciding( coWinX(pos[0]), coWinY(pos[1]) ); break;
            default:                print_weakfocus( coWinX( pos[0] ), coWinY( pos[1] ) ); break;
            }            
            break;
        }
    }
}

void QWinSphere::printPoint( struct strong_focus * p )
{
    double pos[2];

    if( p != NULL )
    {
        getChartPos( p->chart, p->x0, p->y0, pos );

        if( pos[0] < x0 || pos[0] > x1 || pos[1] < y0 || pos[1] > y1 )
            return;

        if( p->stable == -1 )
        { 
            switch( p->position )
            {
            case POSITION_VIRTUAL:  print_virtualstablestrongfocus( coWinX( pos[0] ), coWinY( pos[1] ) ); break;
            case POSITION_COINCIDING: break;
            case POSITION_COINCIDING_VIRTUAL: break;
            case POSITION_COINCIDING_MAIN:
                                    print_coinciding( coWinX(pos[0]), coWinY(pos[1]) ); break;
            default:                print_stablestrongfocus( coWinX( pos[0] ), coWinY( pos[1] ) ); break;
            }
        }
        else
        {
            switch( p->position )
            {
            case POSITION_VIRTUAL:  print_virtualunstablestrongfocus( coWinX( pos[0] ), coWinY( pos[1] ) ); break;
            case POSITION_COINCIDING: break;
            case POSITION_COINCIDING_VIRTUAL: break;
            case POSITION_COINCIDING_MAIN:
                                    print_coinciding( coWinX(pos[0]), coWinY(pos[1]) ); break;
            default:                print_unstablestrongfocus( coWinX( pos[0] ), coWinY( pos[1] ) ); break;
            }
        }
    }
}

void QWinSphere::printPoint( struct degenerate * p )
{
    double pos[2];

    if( p != NULL )
    {
        getChartPos( p->chart, p->x0, p->y0, pos );

        if( pos[0] < x0 || pos[0] > x1 || pos[1] < y0 || pos[1] > y1 )
            return;

        switch( p->position )
        {
        case POSITION_VIRTUAL:  print_virtualdegen( coWinX( pos[0] ), coWinY( pos[1] ) ); break;
        case POSITION_COINCIDING: break;
        case POSITION_COINCIDING_VIRTUAL: break;
        case POSITION_COINCIDING_MAIN:
                                print_coinciding( coWinX(pos[0]), coWinY(pos[1]) ); break;
        default:                print_degen( coWinX( pos[0] ), coWinY( pos[1] ) ); break;
        }
    }
}

void QWinSphere::printPoint( struct semi_elementary * p )
{
    double pos[2];

    if( p != NULL )
    {
        getChartPos( p->chart, p->x0, p->y0, pos );

        if( pos[0] < x0 || pos[0] > x1 || pos[1] < y0 || pos[1] > y1 )
            return;

        switch(p->position)
        {
        case POSITION_VIRTUAL:  
                    switch(p->type)
                    {
                    case 1: print_virtualsesaddlenode( coWinX( pos[0] ), coWinY( pos[1] ) ); break;
                    case 2: print_virtualsesaddlenode( coWinX( pos[0] ), coWinY( pos[1] ) ); break;
                    case 3: print_virtualsesaddlenode( coWinX( pos[0] ), coWinY( pos[1] ) ); break;
                    case 4: print_virtualsesaddlenode( coWinX( pos[0] ), coWinY( pos[1] ) ); break;
                    case 5: print_virtualseunstablenode( coWinX( pos[0] ), coWinY( pos[1] ) ); break;
                    case 6: print_virtualsesaddle( coWinX( pos[0] ), coWinY( pos[1] ) ); break;
                    case 7: print_virtualsesaddle( coWinX( pos[0] ), coWinY( pos[1] ) ); break;
                    case 8: print_virtualsestablenode( coWinX( pos[0] ), coWinY( pos[1] ) ); break;
                    }
                    break;
        case POSITION_COINCIDING: break;
        case POSITION_COINCIDING_VIRTUAL: break;
        case POSITION_COINCIDING_MAIN:
                    print_coinciding( coWinX(pos[0]), coWinY(pos[1]) );
                    break;
        default:
                    switch(p->type)
                    {
                    case 1: print_sesaddlenode( coWinX( pos[0] ), coWinY( pos[1] ) ); break;
                    case 2: print_sesaddlenode( coWinX( pos[0] ), coWinY( pos[1] ) ); break;
                    case 3: print_sesaddlenode( coWinX( pos[0] ), coWinY( pos[1] ) ); break;
                    case 4: print_sesaddlenode( coWinX( pos[0] ), coWinY( pos[1] ) ); break;
                    case 5: print_seunstablenode( coWinX( pos[0] ), coWinY( pos[1] ) ); break;
                    case 6: print_sesaddle( coWinX( pos[0] ), coWinY( pos[1] ) ); break;
                    case 7: print_sesaddle( coWinX( pos[0] ), coWinY( pos[1] ) ); break;
                    case 8: print_sestablenode( coWinX( pos[0] ), coWinY( pos[1] ) ); break;
                    }
                    break;
        }
    }
}

void QWinSphere::printPoints( void )
{
    struct saddle * sp;
    struct node * np;
    struct weak_focus * wfp;
    struct strong_focus * sfp;
    struct semi_elementary * sep;
    struct degenerate * dp;
    int i;

    print_comment( "Printing symbols at all singular points:" );

    for( i = 0; i < ThisVF->numvf; i++ )
    {
        for( sp = VFResults.vf[i]->first_saddle_point; sp != NULL; sp = sp->next_saddle )
            printPoint( sp );
        for( np = VFResults.vf[i]->first_node_point; np != NULL; np = np->next_node )
            printPoint( np );
        for( wfp = VFResults.vf[i]->first_wf_point; wfp != NULL; wfp = wfp->next_wf )
            printPoint( wfp );
        for( sfp = VFResults.vf[i]->first_sf_point; sfp != NULL; sfp = sfp->next_sf )
            printPoint( sfp );
        for( sep = VFResults.vf[i]->first_se_point; sep != NULL; sep = sep->next_se )
            printPoint( sep );
        for( dp = VFResults.vf[i]->first_de_point; dp != NULL; dp = dp->next_de )
            printPoint( dp );
    }
}

void QWinSphere::printPointSeparatrices( struct semi_elementary * p )
{
    struct sep * separatrice;
    
    for( separatrice = p->separatrices; separatrice != NULL; separatrice = separatrice->next_sep )
    {
        draw_sep( this, separatrice->first_sep_point );
        if( separatrice->next_sep != NULL )
        {
            print_comment( "Next separatrix of degenerate point:" );
        }
    }
}

void QWinSphere::printPointSeparatrices( struct saddle * p )
{
    struct sep * separatrice;
    
    for( separatrice = p->separatrices; separatrice != NULL; separatrice = separatrice->next_sep )
    {
        draw_sep( this, separatrice->first_sep_point );
        if( separatrice->next_sep != NULL )
        {
            print_comment( "Next separatrix of saddle point:" );
        }
    }
}

void QWinSphere::printPointSeparatrices( struct degenerate * p )
{
    struct blow_up_points *blow_up;
    for( blow_up = p->blow_up; blow_up != NULL; blow_up = blow_up->next_blow_up_point )
    {
        draw_sep( this, blow_up->first_sep_point );
        if( blow_up->next_blow_up_point != NULL )
        {
            print_comment( "Next separatrix of degenerate point:" );
        }
    }
}

void QWinSphere::printSeparatrices( void )
{
    QString comment;
    int i;
    struct saddle * sp;
    struct semi_elementary * sep;
    struct degenerate * dp;

    for( i = 0; i < ThisVF->numvf; i++ )
    {
        for( sp = VFResults.vf[i]->first_saddle_point; sp != NULL; sp = sp->next_saddle )
        {
            comment = "Printing separatrice for saddle singularity:";
            print_comment( comment );
            printPointSeparatrices( sp );
        }
        for( sep = VFResults.vf[i]->first_se_point; sep != NULL; sep = sep->next_se )
        {
            comment = "Printing separatrices for semi-hyperbolic singularity:";
            print_comment( comment );
            printPointSeparatrices( sep );
        }
        for( dp = VFResults.vf[i]->first_de_point; dp != NULL; dp = dp->next_de )
        {
            comment = "Printing separatrices for degenerate singularity:";
            print_comment( comment );
            printPointSeparatrices( dp );
        }
    }
}

void QWinSphere::printGcf( void )
{
    QString comment;
    bool isagcf;
    int i;
    
    isagcf = false;
    for( i = 0; i < ThisVF->numvf; i++ )
        if( VFResults.vf[i]->gcf_points != NULL )
            isagcf=true;

    if( isagcf )
    {
        comment = "Printing Greatest common factor:";
        print_comment( comment );
        for( i = 0; i < ThisVF->numvf; i++ )
            draw_gcf( this, VFResults.vf[i]->gcf_points, CSING, 1 );
    }
}

void QWinSphere::printSeparatingCurves( void )
{
    QString comment;
    int i;
    bool dashes;
    double pcoord[3];
    struct orbits_points * sep;

    if( ThisVF->numcurves > 0 && VFResults.curves != NULL )
    {
        comment = "Printing Separating curves:";
        print_comment( comment );
        for( i = 0; i < ThisVF->numcurves; i++ )
        {
            comment.sprintf( "Curve #%d:", i+1 );
            print_comment( comment );

            dashes=true;
            sep = VFResults.curves[i].sep_points;
            while( sep != NULL )
            {
                if( sep->color == CSEPCURVE )
                {
                    if( sep->dashes && dashes )
                        (*plot_l)(this,pcoord,sep->pcoord,sep->color);
                    else
                    {
                        if( sep->nextpt == NULL )
                            (*plot_p)(this,sep->pcoord,sep->color);
                        else
                            if( !(sep->nextpt->dashes && sep->nextpt->color == CSEPCURVE && dashes) )
                                (*plot_p)(this,sep->pcoord,sep->color); // draw nothing when the next point is a dash
                    }
                    dashes=true;
                }
                else
                {
                    dashes=false;
                }
                copy_x_into_y(sep->pcoord,pcoord);
                sep=sep->nextpt;
            }
        }
    }
}

void QWinSphere::printPoincareSphere( void )
{
    QString comment;
    struct P4POLYLINES * p;
    struct P4POLYLINES * q;

    comment = "Printing Poincare Sphere:";
    print_comment( comment );
    p = produceEllipse( 0.0, 0.0, 1.0, 1.0, false, coWinH(1.0), coWinV(1.0) );
    for( q = p; q != NULL; q = q->next )
    {
        q->x1 = coWinX(q->x1);
        q->y1 = coWinY(q->y1);
        q->x2 = coWinX(q->x2);
        q->y2 = coWinY(q->y2);
    }

    print_elips( coWinX(0), coWinY(0), coWinH(1), coWinV(1), 
        /*P5 VFResultsK->singinf ? CSING:CLINEATINFINITY */
        CLINEATINFINITY,
        false, p );
    

    while( p != NULL )
    {
        q = p;
        p = p->next;
        free( q );
    }
}

void QWinSphere::printPoincareLyapunovSphere( void )
{
    QString comment;
    struct P4POLYLINES * p;
    struct P4POLYLINES * q;

    comment = "Printing Poincare Lyapunov-Sphere (circle at infinity):";
    print_comment( comment );

    p = produceEllipse( 0.0, 0.0, 1.0, 1.0, false, coWinH(1.0), coWinV(1.0) );
    for( q = p; q != NULL; q = q->next )
    {
        q->x1 = coWinX(q->x1);
        q->y1 = coWinY(q->y1);
        q->x2 = coWinX(q->x2);
        q->y2 = coWinY(q->y2);
    }

    print_elips( coWinX(0.0), coWinY(0.0), coWinH(1.0), coWinV(1.0), CLINEATINFINITY, false, p );

    while( p != NULL )
    {
        q = p;
        p = p->next;
        free( q );
    }

    comment = "Printing Poincare Lyapunov-Sphere (circle of finite radius):";
    print_comment( comment );

    p = produceEllipse( 0.0, 0.0, RADIUS, RADIUS, true, coWinH(RADIUS), coWinV(RADIUS) );
    for( q = p; q != NULL; q = q->next )
    {
        q->x1 = coWinX(q->x1);
        q->y1 = coWinY(q->y1);
        q->x2 = coWinX(q->x2);
        q->y2 = coWinY(q->y2);
    }

    print_elips( coWinX(0.0), coWinY(0.0), coWinH(RADIUS), coWinV(RADIUS), CLINEATINFINITY, true, p );

    while( p != NULL )
    {
        q = p;
        p = p->next;
        free( q );
    }
}

void QWinSphere::printLineAtInfinity( void )
{
    switch( VFResults.typeofview )
    {
    case TYPEOFVIEW_U1:
    case TYPEOFVIEW_V1:
        if( x0 < 0.0 && x1 > 0.0 )
            print_line( coWinX(0.0), coWinY(y0), coWinX(0.0), coWinY(y1), CLINEATINFINITY );
        break;
    case TYPEOFVIEW_U2:
    case TYPEOFVIEW_V2:
        if( y0 < 0.0 && y1 > 0.0 )
            print_line( coWinX(x0), coWinY(0.0), coWinX(x1), coWinY(0.0), CLINEATINFINITY );
        break;
    }
}

// -----------------------------------------------------------------------
//                      PRINTORBITS
// -----------------------------------------------------------------------

extern void DrawOrbit( QWinSphere *, double *, struct orbits_points *, int );

void QWinSphere::printOrbits( void )
{
    // inspired by DrawOrbits, except that we put comments between

    struct orbits * orbit;
    QString s;
    int i;
    i = 1;

    for( orbit = VFResults.first_orbit; orbit != NULL; orbit = orbit->next )
    {
        s.sprintf( "Starting orbit %d", i++ );
        print_comment( s );
        DrawOrbit( this, orbit->pcoord, orbit->firstpt, orbit->color );
    }
}

void QWinSphere::printLimitCycles( void )
{
    // inspired by DrawOrbits, except that we put comments between

    struct orbits * orbit;
    QString s;
    int i;
    i = 1;

    for( orbit = VFResults.first_lim_cycle; orbit != NULL; orbit = orbit->next )
    {
        s.sprintf( "Starting Limit Cycle %d", i++ );
        print_comment( s );
        DrawOrbit( this, orbit->pcoord, orbit->firstpt, orbit->color );
    }
}
