#include <qapplication.h>
#include <qstatusbar.h>
#include <qmessagebox.h>
#include <qpushbutton.h>
#include <qlabel.h>
#include <qlineedit.h>
#include <qlayout.h>
#include <qlistwidget.h>
#include "custom.h"
#include "table.h"
#include "file_vf.h"
#include "win_p5.h"
#include "p5application.h"
#include "win_curves.h"
#include "win_find.h"
#include "win_inputsphere.h"
#include "math_p4.h"
#include "win_view.h"

extern QP5Application * p5app;

extern void SetP4WindowTitle( QWidget *, QString );
extern QString DescribeRegion( double * );

QCurvesDlg::~QCurvesDlg()
{
    GetDataFromDlg();
    delete ViewParams_Window;
    ViewParams_Window = NULL;
}

QCurvesDlg::QCurvesDlg( QFindDlg * _parent )
    : QWidget()
{
    parent = _parent;
//  setFont( QFont( FONTSTYLE, FONTSIZE ) );

    QLabel * p4title = new QLabel( "Region Separating Curves:", this );
    p4title->setFont( *(p5app->TitleFont) );
    QLabel * p4title2 = new QLabel( "Vector Field List:", this );
    p4title2->setFont( *(p5app->TitleFont) );

    btn_add = new QPushButton( "&Add", this );
    btn_del = new QPushButton( "&Del", this );
    btn_edit = new QPushButton( "E&dit", this );

    btn_zoomout = new QPushButton( "&Zoom out", this );
    btn_refresh = new QPushButton( "&Refresh", this );
    btn_eval = new QPushButton( "&Eval", this );
    btn_mark = new QPushButton( "&Mark", this );
    btn_unmark = new QPushButton( "&Unmark", this );
    btn_resetmarks = new QPushButton( "&Reset Marks", this );
    
    lbl_vf_or_curves = new QLabel( "Curves:", this );
    lbl_vf_or_curves->setFont( *(p5app->TitleFont) );
    lbl_vf_or_curves->setMinimumSize( lbl_vf_or_curves->sizeHint() );
    lbl_vf_or_curves->setMaximumSize( lbl_vf_or_curves->sizeHint() );
    lbl_vf_or_curves->setText("VF:");
    markingvf = true;
    
    btn_view = new QPushButton( "&View", this );
    lbl_status = new QLabel( "MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM", this );
    lbl_status->setMaximumSize( lbl_status->sizeHint() );
    lbl_status->setText( "Ready" );
    lbl_info = new QLabel( "Click on a region to assign the selected vector field", this );
    
    btn_mark->setCheckable(true);
    btn_unmark->setCheckable(true);
    btn_mark->setChecked(marking=true);

    lst_curves = new QListWidget(this);
    lst_vfs = new QListWidget(this);

    isphere = new QWinInputSphere(this, lbl_status);
    isphere->SetupPlot();
    
    ViewParams_Window = new QViewDlg(this,false);
    ViewParams_Window->UpdateDlgData();

    lbl_numpoints = new QLabel( "# curve points:", this );
    edt_numpoints = new QLineEdit( "", this );

    if( p4smallicon != NULL )
        setWindowIcon( *p4smallicon );

    // layout

    mainLayout = new QBoxLayout( QBoxLayout::LeftToRight, this );
    QBoxLayout * LayoutA = new QBoxLayout( QBoxLayout::TopToBottom, this );
    QBoxLayout * LayoutB = new QBoxLayout( QBoxLayout::TopToBottom, this );
    LayoutA->addWidget( p4title );
    QHBoxLayout * layout1 = new QHBoxLayout();
    layout1->addWidget( btn_add );
    layout1->addWidget( btn_del );
    QHBoxLayout * layout1a = new QHBoxLayout();
    layout1a->addWidget( lbl_numpoints );
    layout1a->addWidget( edt_numpoints );

    layout1->addWidget( btn_edit );
    LayoutA->addLayout( layout1 );
    LayoutA->addWidget( lst_curves );
    LayoutA->addLayout( layout1a );
    LayoutA->addWidget( p4title2 );
    LayoutA->addWidget( lst_vfs );

    QHBoxLayout * layout2 = new QHBoxLayout();
    layout2->addWidget( lbl_vf_or_curves );
    layout2->addWidget( btn_mark );
    layout2->addWidget( btn_unmark );
    layout2->addWidget( btn_resetmarks );
    layout2->addStretch(1);

    QHBoxLayout * layout3 = new QHBoxLayout();
    layout3->addWidget( btn_refresh );
    layout3->addWidget( btn_eval );
    layout3->addWidget( btn_view );
    layout3->addWidget( btn_zoomout );
    layout3->addStretch(1);

    LayoutB->addLayout( layout2 );
    LayoutB->addWidget( lbl_info );
    LayoutB->addLayout( layout3 );
    LayoutB->addWidget(isphere);
    LayoutB->addWidget( lbl_status );
    
    LayoutB->setStretchFactor(isphere,1);
    LayoutB->setStretchFactor(lbl_status,0);
    
    mainLayout->addLayout(LayoutA);
    mainLayout->addLayout(LayoutB);
    mainLayout->setStretchFactor(LayoutA,0);
    mainLayout->setStretchFactor(LayoutB,2);
    setLayout(mainLayout);

    // connections

    QObject::connect( btn_add,  SIGNAL(clicked()), this, SLOT(OnBtnAdd()) );
    QObject::connect( btn_del,  SIGNAL(clicked()), this, SLOT(OnBtnDel()) );
    QObject::connect( btn_edit,  SIGNAL(clicked()), this, SLOT(OnBtnEdit()) );
    QObject::connect( btn_mark,  SIGNAL(toggled(bool)), this, SLOT(OnBtnMarkToggled(bool)) );
    QObject::connect( btn_unmark,  SIGNAL(toggled(bool)), this, SLOT(OnBtnUnmarkToggled(bool)) );
    QObject::connect( btn_resetmarks,  SIGNAL(clicked()), this, SLOT(OnBtnResetMarks()) );
    QObject::connect( btn_zoomout,  SIGNAL(clicked()), this, SLOT(OnBtnZoomOut()) );
    QObject::connect( btn_refresh,  SIGNAL(clicked()), this, SLOT(OnBtnRefresh()) );
    QObject::connect( btn_eval,  SIGNAL(clicked()), this, SLOT(OnBtnEval()) );
    QObject::connect( btn_view,  SIGNAL(clicked()), this, SLOT(OnBtnView()) );

    QObject::connect( lst_curves,  SIGNAL(itemChanged(QListWidgetItem *)), this, SLOT(OnCurveChanged(QListWidgetItem *)) );
    QObject::connect( lst_curves,  SIGNAL(itemSelectionChanged()), this, SLOT(OnCurvesSelectionChanged()) );
    QObject::connect( lst_vfs,  SIGNAL(itemSelectionChanged()), this, SLOT(OnVfsSelectionChanged()) );
    QObject::connect( lst_curves,  SIGNAL(itemActivated(QListWidgetItem *)), this, SLOT(OnCurvesItemActivated(QListWidgetItem *)) );
    QObject::connect( lst_vfs,  SIGNAL(itemActivated(QListWidgetItem *)), this, SLOT(OnVfsItemActivated(QListWidgetItem *)) );

    QObject::connect( lst_curves,  SIGNAL(itemClicked(QListWidgetItem *)), this, SLOT(OnCurvesClicked(QListWidgetItem *)) );
    QObject::connect( lst_vfs,  SIGNAL(itemClicked(QListWidgetItem *)), this, SLOT(OnVfsClicked(QListWidgetItem *)) );
    QObject::connect( edt_numpoints,  SIGNAL(editingFinished()), this, SLOT(OnNumpointsEditingFinished()) );

#ifdef TOOLTIPS
    btn_add->setToolTip( "Adds a new separating curve to the list.\nEnter a Maple expression after pressing the button." );
    btn_del->setToolTip( "Deletes the selected separating curve from the list." );
    btn_edit->setToolTip( "Edits the separating curve Maple expression." );
    btn_mark->setToolTip( "If selected, clicking in the configuration window adds a marks to the clicked region." );
    btn_unmark->setToolTip( "If selected, clicking in the configuration window removes a marks to the clicked region." );
    btn_resetmarks->setToolTip( "Reset all region marks." );
    btn_zoomout->setToolTip( "Zoom out.  You can zoom in by pressing CTRL, clicking in the window upper-left corner\nand keeping the mouse pressed while going to the lower-right corner." );
    btn_refresh->setToolTip( "Refresh the configuration window." );
    btn_eval->setToolTip( "Evaluate the configuration using Maple." );
    btn_view->setToolTip( "Changes the view on the configuration window." );
    lst_curves->setToolTip( "A list of all separating curves.\nThe combination of the signs define the different regions." );
    lst_vfs->setToolTip( "A list of vector fields, numbered #1, #2, ..." );
    edt_numpoints->setToolTip( "Number of points to calculate this separating curve while evaluating." );
//    isphere->setToolTip( "Configuration window.  Click to mark/unmark a region with the selected vector field (or curve)" );

#endif
    // finishing

    SetP4WindowTitle( this, "P5 Piecewise Configuration" );

    UpdateDlgData();
    lst_vfs->setCurrentRow(0);
    lst_vfs->setFocus();
}

void QCurvesDlg::OnBtnAdd( void )
{
    QListWidgetItem * itm;

    if( ThisVF->numvfregions > 0 || ThisVF->numcurveregions > 0 )
    {
        ThisVF->ClearVfMarks();
        ThisVF->ClearCurveMarks();
        
    }
    
    itm = new QListWidgetItem( "polynomial" );
    itm->setFlags( Qt::ItemIsSelectable | Qt::ItemIsEditable | Qt::ItemIsEnabled );
    lst_curves->addItem(itm);
    lst_curves->scrollToItem(itm);
    lst_curves->setCurrentItem(itm);
    lst_curves->editItem(itm);
    ThisVF->AddSeparatingCurve();
    if( ThisVF->changed == false )
    {
        ThisVF->changed = true;
        p5app->Signal_Changed();
    }
    isphere->refresh();
}

void QCurvesDlg::OnBtnDel( void )
{
    int r;
    QListWidgetItem * itm;
    
    if( lst_curves->selectedItems().count() == 0 )
        return;

    if( ThisVF->numvfregions > 0 || ThisVF->numcurveregions > 0 )
    {
        ThisVF->ClearVfMarks();
        ThisVF->ClearCurveMarks();
    }
            
    r = lst_curves->currentRow();
    
    if( r >= 0 && r < ThisVF->numcurves )
    {
        itm = lst_curves->takeItem(r);
        delete(itm);
    }
    ThisVF->DeleteSeparatingCurve(r);
    if( ThisVF->changed == false )
    {
        ThisVF->changed = true;
        p5app->Signal_Changed();
    }
    isphere->refresh();
}

void QCurvesDlg::OnBtnEdit( void )
{
    lst_curves->editItem(lst_curves->currentItem());
}

void QCurvesDlg::GetDataFromDlg( void )
{
}

void QCurvesDlg::UpdateDlgData( void )
{
    QListWidgetItem * itm;
    int i,v;
    QString s;

    lst_curves->clear();

    for( i = 0; i < ThisVF->numcurves; i++ )
    {
        if( ThisVF->curve[i]->length() == 0 )
            itm = new QListWidgetItem( "polynomial", lst_curves );
        else
            itm = new QListWidgetItem( *(ThisVF->curve[i]), lst_curves );
        itm->setFlags( Qt::ItemIsSelectable | Qt::ItemIsEditable | Qt::ItemIsEnabled );
    }

    if( lst_curves->selectedItems().count() > 0 )
    {
        int curveindex;
        curveindex = lst_curves->row( lst_curves->selectedItems()[0] );
        if( curveindex <0 || curveindex >= ThisVF->numcurves || ThisVF->numpointscurve == NULL )
            v = -1;
        else
            v = ThisVF->numpointscurve[curveindex];
    }
    else
        v=-1;
    
    if( v < 0 )
        edt_numpoints->setText("");
    else
    {
        s.setNum(v);
        edt_numpoints->setText(s);
    }
    
    lst_vfs->clear();

    for( i = 0; i < ThisVF->numvf; i++ )
    {
        s = "x' = ";
        s = s.append( *(ThisVF->xdot[i]) );
        s = s.append( "\ny' = " );
        s = s.append( *(ThisVF->ydot[i]) );
        s = s.append( "\n" );
        itm = new QListWidgetItem( s, lst_vfs );
        itm->setFlags( Qt::ItemIsSelectable | Qt::ItemIsEnabled );
    }

    isphere->SetupCoordinateTransformations();
    isphere->refresh();
}

void QCurvesDlg::OnCurveChanged( QListWidgetItem * itm )
{
    int r;
    QString s;
    r = lst_curves->row(itm);
    if( r < 0 || r >= ThisVF->numcurves )
        return;
    s = itm->text();
    s = s.trimmed();

    if( ThisVF->curve[r]->compare( s ) )
    {
        *(ThisVF->curve[r]) = s;
        if( ThisVF->changed == false )
        {
            ThisVF->changed = true;
            p5app->Signal_Changed();
        }
    }
}

void QCurvesDlg::OnCurvesSelectionChanged( void )
{
    int curveindex;
    int v;
    QString s;
    
    if( lst_curves->selectedItems().count() > 0 )
    {
        if( markingvf )
        {
            lbl_vf_or_curves->setText( "Curves:" );
            if( marking )
                lbl_info->setText( "Click on a region where selected curve is not to be drawn" );
            else
                lbl_info->setText( "Click on a region to where selected curve is to be drawn" );
            markingvf=false;
        }

        curveindex = lst_curves->row( lst_curves->selectedItems()[0] );
        if( curveindex <0 || curveindex >= ThisVF->numcurves || ThisVF->numpointscurve == NULL )
            v = -1;
        else
            v = ThisVF->numpointscurve[curveindex];
    }
    else
        v=-1;
    
    if( v < 0 )
        s="";
    else
        s.setNum(v);
    
    edt_numpoints->setText(s);
}

void QCurvesDlg::OnCurvesItemActivated(QListWidgetItem * itm )
{
    OnCurvesSelectionChanged();
}


void QCurvesDlg::OnVfsSelectionChanged( void )
{
    if( lst_vfs->selectedItems().count() > 0 )
    {
        if( !markingvf )
        {
            lbl_vf_or_curves->setText( "VF:" );
            if( marking )
                lbl_info->setText( "Click on a region to assign the selected vector field" );
            else
                lbl_info->setText( "Click on a region to unassign the selected vector field" );
            markingvf=true;
        }
    }
}

void QCurvesDlg::OnVfsItemActivated(QListWidgetItem *)
{
    OnVfsSelectionChanged();
}

void QCurvesDlg::OnCurvesClicked(QListWidgetItem *)
{
    OnCurvesSelectionChanged();
}

void QCurvesDlg::OnVfsClicked(QListWidgetItem *)
{
    OnVfsSelectionChanged();
}

void QCurvesDlg::OnBtnUnmarkToggled( bool c )
{
    if( c )
    {
        btn_mark->setChecked(false);
        marking = false;
        if( markingvf )
            lbl_info->setText( "Click on a region to unassign the selected vector field" );
        else
            lbl_info->setText( "Click on a region to where selected curve is to be drawn" );
    }
    else
    {
        marking = true;
        if( !btn_mark->isChecked() )
            btn_mark->setChecked(true);
        if( markingvf )
            lbl_info->setText( "Click on a region to assign the selected vector field" );
        else
            lbl_info->setText( "Click on a region where selected curve is not to be drawn" );
    }
}

void QCurvesDlg::OnBtnMarkToggled( bool c )
{
    if( c )
    {
        btn_unmark->setChecked(false);
        marking = true;
        if( markingvf )
            lbl_info->setText( "Click on a region to assign the selected vector field" );
        else
            lbl_info->setText( "Click on a region where selected curve is not to be drawn" );
    }
    else
    {
        marking = false;
        if( !btn_unmark->isChecked() )
            btn_unmark->setChecked(true);
        if( markingvf )
            lbl_info->setText( "Click on a region to unassign the selected vector field" );
        else
            lbl_info->setText( "Click on a region to where selected curve is to be drawn" );
    }
}

void QCurvesDlg::OnBtnResetMarks( void )
{
    if( markingvf )
    {
        if( ThisVF->numvfregions > 0 )
        {
            ThisVF->ClearVfMarks();
            if( ThisVF->changed == false )
            {
                ThisVF->changed = true;
                p5app->Signal_Changed();
            }
            isphere->refresh();
        }
        lbl_info->setText( "All Vector field marks have been erased.  Please re-assign vector fields from scratch." );
    }
    else
    {
        if( ThisVF->numcurveregions > 0 )
        {
            ThisVF->ClearCurveMarks();
            if( ThisVF->changed == false )
            {
                ThisVF->changed = true;
                p5app->Signal_Changed();
            }
            isphere->refresh();
        }
        lbl_info->setText( "All Curve marks have been erased. Now separating curves are drawn completely." );
    }
}

void QCurvesDlg::OnBtnZoomOut( void )
{
    isphere->iszoom = false;
    isphere->SetupPlot();
    isphere->refresh();
}

void QCurvesDlg::OnBtnView( void )
{
    ViewParams_Window->show();
    ViewParams_Window->raise();
}

void QCurvesDlg::OnBtnEval( void )
{
    if( ThisVF->evaluating )
        return;

    parent->GetDataFromDlg();

    ThisVF->EvaluateCurves();
    if( ThisVF->evaluating )
        btn_eval->setEnabled(false);
    else
    {
        QMessageBox::critical( this, "P5",
            "An error occured while plotting the separating curves.\n" );
    }
}

void QCurvesDlg::OnBtnRefresh( void )
{
    if( ViewParams_Window->GetDataFromDlg() )
    {
        // true when a big change occured in the view

        VFResults.SetupCoordinateTransformations();
        isphere->SetupPlot();
    }
    isphere->refresh();
}

void QCurvesDlg::Signal_CurvesEvaluated( void )
{
    btn_eval->setEnabled(true);
    
    if( !VFResults.ReadTables( ThisVF->getbarefilename(), true, false ) )
    {
        QMessageBox::critical( this, "P5",
                "Cannot read computation results.\n"
                "Please check the input-vector field, curves and parameters!\n" );
    }
    isphere->SetupPlot();
    isphere->refresh();
}

void QCurvesDlg::MakeZoom( double x1, double y1, double x2, double y2 )
{
    // chart a zoom window.

    double swap;

    if( x1 > x2 )
    {
        swap = x1;
        x1 = x2;
        x2 = swap;
    }
    if( y1 > y2 )
    {
        swap = y1;
        y1 = y2;
        y2 = swap;
    }
    if( x1 == x2 || y1 == y2 )
        return;

    isphere->x0 = x1;
    isphere->y0 = y1;
    isphere->x1 = x2;
    isphere->y1 = y2;
    isphere->iszoom = true;
    isphere->SetupPlot();
    isphere->refresh();
}

void QCurvesDlg::OnMouseClickLeft( double x0, double y0, double z0 )
{
    int curveindex;
    int vfindex;
    int oldindex;
    double pcoord[3];
    QString s;
    
    pcoord[0]=x0;
    pcoord[1]=y0;
    pcoord[2]=z0;
    
    if( VFResults.curves == NULL )
    {
        QMessageBox::critical( this, "P5",
            "Please evaluate the piecewise configuration first,\nusing the Eval button.\n" );
        return;
    }
    
    if( markingvf )
    {
        if( lst_vfs->selectedItems().count() == 0 )
            return;
        vfindex = lst_vfs->row( lst_vfs->selectedItems()[0] );
        
        oldindex = ThisVF->GetVFIndex_sphere( pcoord );
        if( marking )
        {
            if( oldindex == -1 )
            {
                s.sprintf( "Marked region with VF #%d", vfindex+1 );
                lbl_status->setText( s );
                ThisVF->MarkVFRegion( vfindex, pcoord );
                if( ThisVF->changed == false )
                {
                    ThisVF->changed = true;
                    p5app->Signal_Changed();
                }
                isphere->refresh();
            }
            else
            {
                if( oldindex == vfindex )
                {
                    s.sprintf( "Region is already marked with VF #%d", vfindex+1 );
                    lbl_status->setText( s );
                    return;
                }
                else
                {
                    s.sprintf( "Marked region with VF #%d", vfindex+1 );
                    lbl_status->setText( s );
                    ThisVF->MarkVFRegion( vfindex, pcoord );
                    if( ThisVF->changed == false )
                    {
                        ThisVF->changed = true;
                        p5app->Signal_Changed();
                    }
                    isphere->refresh();
                }
            }
        }
        else
        {
            if( oldindex == -1 )
            {
                s.sprintf( "The region is not marked yet: no VF is assigned" );
                lbl_status->setText( s );
                return;
            }
            else
            {
                if( oldindex == vfindex )
                {
                    s.sprintf( "Region has been unmarked: no VF is assigned" );
                    lbl_status->setText( s );
                    ThisVF->UnmarkVFRegion( vfindex, pcoord );
                    if( ThisVF->changed == false )
                    {
                        ThisVF->changed = true;
                        p5app->Signal_Changed();
                    }
                    isphere->refresh();
                }
                else
                {
                    s.sprintf( "Region is marked by a different VF (#%d)!", oldindex+1 );
                    lbl_status->setText( s );
                    return;
                }
            }
        }
    }
    else
    {
        if( lst_curves->selectedItems().count() == 0 )
            return;
        curveindex = lst_curves->row( lst_curves->selectedItems()[0] );

        if( marking )
        {
            if( !ThisVF->IsCurvePointDrawn( curveindex, pcoord ) )
            {
                s.sprintf( "In this region, the plotting of the curve #%d was already disabled.", curveindex+1 );
                lbl_status->setText( s );
            }
            else
            {
                s.sprintf( "The plotting of the curve #%d is now disabled in this region.", curveindex+1 );
                lbl_status->setText( s );

                ThisVF->MarkCurveRegion( curveindex, pcoord );
                if( ThisVF->changed == false )
                {
                    ThisVF->changed = true;
                    p5app->Signal_Changed();
                }
                isphere->refresh();
            }
        }
        else
        {
            if( ThisVF->IsCurvePointDrawn( curveindex, pcoord ) )
            {
                s.sprintf( "In this region, the plotting of the curve #%d was already enabled.", curveindex+1 );
                lbl_status->setText( s );
            }
            else
            {
                s.sprintf( "The plotting of the curve #%d is now again enabled in this region.", curveindex+1 );
                lbl_status->setText( s );

                ThisVF->UnmarkCurveRegion( curveindex, pcoord );
                if( ThisVF->changed == false )
                {
                    ThisVF->changed = true;
                    p5app->Signal_Changed();
                }
                isphere->refresh();
            }
        }
    }
}

void QCurvesDlg::OnMouseClickRight( double x0, double y0, double z0 )
{
    // unused
}

void QCurvesDlg::OnNumpointsEditingFinished( void )
{
    QString s;
    bool ok;
    int v;
    int curveindex;

    if( lst_curves->selectedItems().count() == 0 )
        return;
    curveindex = lst_curves->row( lst_curves->selectedItems()[0] );
    if( curveindex <0 || curveindex >= ThisVF->numcurves )
        return;
    
    s = edt_numpoints->text();
    v = s.toInt( &ok );
    if( !ok || v <= 0 )
        return;
    
    if( v == ThisVF->numpointscurve[curveindex] )
        return;
    
    ThisVF->numpointscurve[curveindex] = v;
            
    if( ThisVF->changed == false )
    {
        ThisVF->changed = true;
        p5app->Signal_Changed();
    }
}
