#include <math.h>
#include <qlabel.h>
#include <qpaintengine.h>
#include <qtimer.h>
#include <qpixmap.h>
#include <qimage.h>
#include <qmessagebox.h>
#include <qpainter.h>
#include <qdatetime.h>
#include <qapplication.h>
#include <qevent.h>
#include "custom.h"
#include "table.h"
#include "file_vf.h"
#include "win_p5.h"
#include "math_p4.h"
#include "win_inputsphere.h"
#include "win_main.h"
#include "win_curves.h"
#include "p5application.h"
#include "win_event.h"

extern void SetP4WindowTitle( QWidget *, QString );
extern bool LineRectangleIntersect( double &, double &, double &, double &, double, double, double, double );
extern QString makechartstring( int p, int q, bool isu1v1chart, bool negchart );
extern QString DescribeRegion( double * );

QWinInputSphere::QWinInputSphere( QCurvesDlg * parent, QLabel * lbl_status )
    : QWidget(parent)
{
    PainterCache = NULL;
    iszoom = false;
    isPainterCacheDirty = true;
    AnchorMap = NULL;
    refreshTimeout = NULL;
    status = lbl_status;
    
    setAttribute( Qt::WA_PaintOnScreen );

    parentWnd = parent;
    setMinimumSize(MINWIDTHPLOTWINDOW,MINHEIGHTPLOTWINDOW);         // THIS IS THE MINIMUM SIZE
    w = width();
    h = height();
    idealh = w;

    horPixelsPerMM = ((double)w) / ((double)widthMM());
    verPixelsPerMM = ((double)h) / ((double)heightMM());

    setMouseTracking(true);

    selectingZoom = false;
    setFocusPolicy( Qt::ClickFocus );
    setWindowFlags( windowFlags() );

    CircleAtInfinity = NULL;
    PLCircle = NULL;
    SetupPlot();
}

void QWinInputSphere::keyPressEvent ( QKeyEvent * e )
{
    e->ignore();
}

void QWinInputSphere::SetupPlot( void )
{
    struct P4POLYLINES * t;
    QPalette palette;
    double idealhd;

    SetupCoordinateTransformations();

    palette.setColor(backgroundRole(), QXFIGCOLOR(spherebgcolor=CBACKGROUND) );
    setPalette(palette);

    while( CircleAtInfinity != NULL )
    {
        t = CircleAtInfinity;
        CircleAtInfinity = t->next;
        free( t );
    }
    while( PLCircle != NULL )
    {
        t = PLCircle;
        PLCircle = t->next;
        free( t );
    }

    if( !iszoom )
    {
        switch( VFResults.typeofview )
        {
        case TYPEOFVIEW_PLANE:
        case TYPEOFVIEW_U1:
        case TYPEOFVIEW_U2:
        case TYPEOFVIEW_V1:
        case TYPEOFVIEW_V2:
            x0 = VFResults.xmin;
            y0 = VFResults.ymin;
            x1 = VFResults.xmax;
            y1 = VFResults.ymax;
            break;
        case TYPEOFVIEW_SPHERE:
            x0 = -1.1;
            y0 = -1.1;
            x1 = 1.1;
            y1 = 1.1;
            break;
        }
    }
        
    dx = x1-x0;
    dy = y1-y0;    
    idealhd = w;
    idealhd = (idealhd/dx)*dy;
    idealh = (int)(idealhd+.5);

    if( VFResults.typeofview == TYPEOFVIEW_SPHERE )
    {
        CircleAtInfinity = produceEllipse( 0.0, 0.0, 1.0, 1.0, false, coWinH(1.0), coWinV(1.0) );
        if( ThisVF->p != 1 || ThisVF->q != 1 )
            PLCircle = produceEllipse( 0.0, 0.0, RADIUS, RADIUS, true, coWinH(RADIUS), coWinV(RADIUS) );
    }
    isPainterCacheDirty = true;
}

QWinInputSphere::~QWinInputSphere()
{
    struct P4POLYLINES * t;
    while( CircleAtInfinity != NULL )
    {
        t = CircleAtInfinity;
        CircleAtInfinity = t->next;
        free( t );
    }
    while( PLCircle != NULL )
    {
        t = PLCircle;
        PLCircle = t->next;
        free( t );
    }

    if( PainterCache != NULL )
    {
        delete PainterCache;
        PainterCache = NULL;
    }
    
    if( AnchorMap != NULL )
    {
        delete AnchorMap;
        AnchorMap = NULL;
    }
    
    if( refreshTimeout != NULL )
    {
        delete refreshTimeout;
        refreshTimeout = NULL;
    }
}

void QWinInputSphere::LoadAnchorMap( void )
{
     int x1,y1;
     int x2,y2;
     int aw,ah;         
     int s;
     if( selectingZoom )
     {
         x1 = zoomAnchor1.x();
         y1 = zoomAnchor1.y();
         x2 = zoomAnchor2.x();
         y2 = zoomAnchor2.y();
     }
     else
         return;
     
     if( x1 > x2 ) { s=x1; x1=x2; x2=s; }
     if( y1 > y2 ) { s=y1; y1=y2; y2=s; }
     if( x1 < 0 ) x1 = 0;
     if( y1 < 0 ) y1 = 0;
     if( x2 >= width() ) x2 = width()-1;
     if( y2 >= height() ) y2 = height()-1;
     
     aw = x2-x1+1;
     ah = y2-y1+1;
     
         
     if( AnchorMap != NULL )
     {
         if( AnchorMap->width() < aw || AnchorMap->height() < ah )
         {
             delete AnchorMap;
             AnchorMap = new QPixmap( aw,ah );
         }
     }
     else
     {
         AnchorMap = new QPixmap( aw,ah );
     }
     
     if( PainterCache == NULL )
     {
         delete AnchorMap;
         AnchorMap = NULL;
         return;
     }   

     QPainter paint( AnchorMap );
     paint.drawPixmap ( 0, 0, aw, ah, *PainterCache, x1, y1, aw, ah );
}

void QWinInputSphere::SaveAnchorMap( void )
{
     int x1,y1;
     int x2,y2;
     int aw,ah;         
     int s;

     if( AnchorMap == NULL || PainterCache == NULL || !selectingZoom )
         return;

     if( selectingZoom )
     {
         x1 = zoomAnchor1.x();
         y1 = zoomAnchor1.y();
         x2 = zoomAnchor2.x();
         y2 = zoomAnchor2.y();
     }
     
     if( x1 > x2 ) { s=x1; x1=x2; x2=s; }
     if( y1 > y2 ) { s=y1; y1=y2; y2=s; }
     if( x1 < 0 ) x1 = 0;
     if( y1 < 0 ) y1 = 0;
     if( x2 >= width() ) x2 = width()-1;
     if( y2 >= height() ) y2 = height()-1;
     
     aw = x2-x1+1;
     ah = y2-y1+1;

     QPainter paint( PainterCache );

     paint.drawPixmap ( x1, y1, aw, ah, *AnchorMap, 0, 0, aw, ah );
     update( x1,y1,aw,ah );
}

void QWinInputSphere::adjustToNewSize(void)
{
    double idealhd;
    QString buf;
    double reqratio;
    double ratio;
    struct P4POLYLINES * t;

    w=width();
    h=height();

    idealhd = w;
    idealhd = (idealhd/dx)*dy;

    idealh = (int)(idealhd+.5);

    reqratio = ( ((double)w)/horPixelsPerMM ) / ( idealh/verPixelsPerMM );
    ratio = ( ((double)w)/horPixelsPerMM ) / ( ((double)h)/verPixelsPerMM );
    buf.sprintf( "Aspect Ratio = %f\n", (float)(ratio/reqratio) );
    status->setText(buf);
    status->setMaximumWidth( w );

    while( CircleAtInfinity != NULL )
    {
        t = CircleAtInfinity;
        CircleAtInfinity = t->next;
        free( t );
    }
    while( PLCircle != NULL )
    {
        t = PLCircle;
        PLCircle = t->next;
        free( t );
    }

    if( VFResults.typeofview == TYPEOFVIEW_SPHERE )
    {
        CircleAtInfinity = produceEllipse( 0.0, 0.0, 1.0, 1.0, false, coWinH(1.0), coWinV(1.0) );
        if( ThisVF->p != 1 || ThisVF->q != 1  )
            PLCircle = produceEllipse( 0.0, 0.0, RADIUS, RADIUS, true, coWinH(RADIUS), coWinV(RADIUS) );
    }

    if( PainterCache != NULL )
    {
        delete PainterCache;
        PainterCache = new QPixmap( size() );
        isPainterCacheDirty = false;

        QPainter paint( PainterCache );
        paint.fillRect(0,0,width(),height(), QColor( QXFIGCOLOR(CBACKGROUND) ) );
        paint.setPen( QXFIGCOLOR(CLINEATINFINITY) );

        staticPainter = &paint;

        if( VFResults.typeofview != TYPEOFVIEW_PLANE )
        {
            if( VFResults.typeofview == TYPEOFVIEW_SPHERE )
            {
                if( ThisVF->p != 1 || ThisVF->q != 1 )
                    plotPoincareLyapunovSphere();
                else
                    plotPoincareSphere();
            }
            else
                plotLineAtInfinity();
        }
        // during resizing : only plot essential information

        QColor c  = QXFIGCOLOR(WHITE);
        c.setAlpha(128);
        paint.setPen(c);
        paint.drawText( 0, 0, width(), height(),
                           Qt::AlignHCenter | Qt::AlignVCenter, "Resizing ...  " );

        staticPainter = NULL;

        if( refreshTimeout != NULL )
            refreshTimeout->stop();
        else
        {
            refreshTimeout = new QTimer();
            connect( refreshTimeout, SIGNAL(timeout()), this, SLOT(refreshAfterResize()) );
        }
        refreshTimeout->start( 500 );
    }
}

void QWinInputSphere::refreshAfterResize( void )
{
     if( refreshTimeout != NULL )
     {
         delete refreshTimeout;
         refreshTimeout = NULL;
     }
    refresh();
}

void QWinInputSphere::resizeEvent( QResizeEvent * e )
{
    if( e->size() == e->oldSize() )
        return;

    adjustToNewSize( );
}

void QWinInputSphere::paintEvent( QPaintEvent * p )
{
    if( ThisVF->evaluating )
        return;

    if( PainterCache == NULL || isPainterCacheDirty )
    {
        if( PainterCache == NULL )
            PainterCache = new QPixmap( size() );
        isPainterCacheDirty = false;

        QPainter paint( PainterCache );
        paint.fillRect(0,0,width(),height(), QColor( QXFIGCOLOR(CBACKGROUND) ) );

        paint.setPen( QXFIGCOLOR(CLINEATINFINITY) );

        staticPainter = &paint;

        // Mental note: do not use prepareDrawing/FinishDrawing here,
        // since it is not good to do drawing for all spheres every time we
        // get a paint event from windows

        if( VFResults.typeofview != TYPEOFVIEW_PLANE )
        {
            if( VFResults.typeofview == TYPEOFVIEW_SPHERE )
            {
                if( ThisVF->p != 1 || ThisVF->q != 1 )
                    plotPoincareLyapunovSphere();
                else
                    plotPoincareSphere();
            }
            else
                plotLineAtInfinity();
        }
        
        plotCurves();
        
        staticPainter = NULL;
    }
    
    QPainter widgetpaint( this );
    widgetpaint.drawPixmap( 0, 0, *PainterCache );
}

void QWinInputSphere::MarkSelection( int x1, int y1, int x2, int y2 )
{
    int bx1, by1, bx2, by2;
     
    if( PainterCache == NULL )
          return;
    
    bx1 = (x1<x2) ? x1 : x2;
    bx2 = (x1<x2) ? x2 : x1;
    by1 = (y1<y2) ? y1 : y2;
    by2 = (y1<y2) ? y2 : y1;
    
    if( bx1 < 0 ) bx1 = 0;
    if( by1 < 0 ) by1 = 0;
    if( bx2 >= width() ) bx2 = width()-1;
    if( by2 >= height() ) by2 = height()-1;

    QPainter p(PainterCache); 
    QColor c;

    c = QXFIGCOLOR(WHITE);
    c.setAlpha(32);
    p.setPen(QXFIGCOLOR(WHITE));
    p.setBrush(c);
    if( bx1 == bx2 || by1 == by2 )
        p.drawLine( bx1, by1, bx2, by2 );
    else
        p.drawRect( bx1, by1, bx2-bx1, by2-by1 );
    
    update( bx1, by1, bx2-bx1+1, by2-by1+1 );
}

void QWinInputSphere::mouseMoveEvent( QMouseEvent * e )
{
    int x, y;
    int index;
    double wx, wy;
    double ucoord[2];
    QString buf;
    double pcoord[3];

    x = e->x();
    y = e->y();
    
    wx = coWorldX(x);
    wy = coWorldY(y);
    
    if( MATHFUNC(is_valid_viewcoord)( wx, wy, pcoord ) )
    {
        switch( VFResults.typeofview )
        {
        case TYPEOFVIEW_PLANE:
            if( VFResults.typeofstudy == TYPEOFSTUDY_ONE )
                buf.sprintf( "Local study   (x,y) = (%f,%f)", (float)wx, (float)wy );
            else
                buf.sprintf( "Planar view   (x,y) = (%f,%f)", (float)wx, (float)wy );
            break;
        case TYPEOFVIEW_SPHERE:
            MATHFUNC(sphere_to_R2)( pcoord[0],pcoord[1],pcoord[2], ucoord );

            if( VFResults.p == 1 && VFResults.q == 1 )
                buf.sprintf( "The Poincare sphere  (x,y) = (%f,%f)",
                    (float)ucoord[0], (float)ucoord[1] );
            else
                buf.sprintf( "The P-L sphere of type (%d,%d)  (x,y) = (%f,%f)",
                    VFResults.p, VFResults.q, (float)ucoord[0], (float)ucoord[1] );
            break;
        case TYPEOFVIEW_U1:
            MATHFUNC(sphere_to_U1)( pcoord[0],pcoord[1],pcoord[2], ucoord );
            if( ucoord[1] >= 0 )
                buf.sprintf( "The U1 chart  (z2,z1) = (%f,%f) ", (float)ucoord[1], (float)ucoord[0] );
            else
                buf.sprintf( "The V1' chart (z2,z1) = (%f,%f) ", (float)ucoord[1], (float)ucoord[0] );
            break;
        case TYPEOFVIEW_V1:
            MATHFUNC(sphere_to_V1)( pcoord[0],pcoord[1],pcoord[2], ucoord );
            if( !VFResults.plweights ) { ucoord[0] = -ucoord[0]; ucoord[1] = -ucoord[1]; }
            if( ucoord[1] >= 0 )
                buf.sprintf( "The V1 chart  (z2,z1) = (%f,%f) ", (float)ucoord[1], (float)ucoord[0] );
            else
                buf.sprintf( "The U1' chart (z2,z1) = (%f,%f) ", (float)ucoord[1], (float)ucoord[0] );
            break;
        case TYPEOFVIEW_U2:
            MATHFUNC(sphere_to_U2)( pcoord[0],pcoord[1],pcoord[2], ucoord );
            if( ucoord[1] >= 0 )
                buf.sprintf( "The U2 chart  (z1,z2) = (%f,%f) ", (float)ucoord[0], (float)ucoord[1] );
            else
                buf.sprintf( "The V2' chart (z1,z2) = (%f,%f) ", (float)ucoord[0], (float)ucoord[1] );
            break;
        case TYPEOFVIEW_V2:
            MATHFUNC(sphere_to_V2)( pcoord[0],pcoord[1],pcoord[2], ucoord );
            if( !VFResults.plweights ) { ucoord[0] = -ucoord[0]; ucoord[1] = -ucoord[1]; }
            if( ucoord[1] >= 0 )
                buf.sprintf( "The V2 chart  (z1,z2) = (%f,%f) ", (float)ucoord[0], (float)ucoord[1] );
            else
                buf.sprintf( "The U2' chart (z1,z2) = (%f,%f) ", (float)ucoord[0], (float)ucoord[1] );
            break;
        }
        if( VFResults.curves != NULL )
        {
            QString s;
            index = ThisVF->GetVFIndex_sphere( pcoord );
            if( index < 0 )
            {
                buf.append( "  VF: NONE" );
            }
            else
            {
                s.sprintf( "  VF: #%d", index+1 );
                buf.append( s );
            }
            s = DescribeRegion( pcoord );
            buf += " (";
            buf += s;
            buf += ")";
        }
    }
    else
    {
        switch( VFResults.typeofview )
        {
        case TYPEOFVIEW_PLANE:
            if( VFResults.typeofstudy == TYPEOFSTUDY_ONE )
                buf.sprintf( "Local study" );
            else
                buf.sprintf( "Planar view" );
            break;
        case TYPEOFVIEW_SPHERE:
            if( VFResults.p == 1 && VFResults.q == 1 )
                buf.sprintf( "The Poincare sphere" );
            else
                buf.sprintf( "The P-L sphere of type (%d,%d)", VFResults.p, VFResults.q );
            break;
        case TYPEOFVIEW_U1:     buf.sprintf( "The U1 chart" );  break;
        case TYPEOFVIEW_V1:     buf.sprintf( "The V1 chart" );  break;
        case TYPEOFVIEW_U2:     buf.sprintf( "The U2 chart" );  break;
        case TYPEOFVIEW_V2:     buf.sprintf( "The V2 chart" );  break;
        }
    }

    status->setText( buf );

    if( selectingZoom )
    {
        SaveAnchorMap();
        zoomAnchor2 = e->pos();
        LoadAnchorMap();
        MarkSelection( zoomAnchor1.x(), zoomAnchor1.y(), zoomAnchor2.x(), zoomAnchor2.y() );
    }
}


double QWinInputSphere::coWorldX( int x )
{
    double wx;

    wx = (double)x;
    wx /= (w-1);
    return (wx*dx + x0);
}

double QWinInputSphere::coWorldY( int y )
{
    double wy;

    wy = (double)(h-1-y);
    wy /= (h-1);
    return (wy*dy + y0);
}

int QWinInputSphere::coWinX( double x )
{
    double wx;
    int iwx;

    wx = (x-x0)/dx;
    wx *= w-1;

    iwx = (int)(wx+0.5);        // +0.5 to round upwards
    if( iwx >= w )
        iwx = w-1;

    return iwx;
}

int QWinInputSphere::coWinH( double deltax )
{
    double wx;

    wx = deltax/dx;
    wx *= w-1;
    return (int)(wx+0.5);
}

int QWinInputSphere::coWinV( double deltay )
{
    double wy;

    wy = deltay/dy;
    wy *= h-1;
    return (int)(wy+0.5);
}

int QWinInputSphere::coWinY( double y )
{
    double wy;
    int iwy;

    wy = (y-y0)/dy;
    wy *= h-1;

    iwy = (int)(wy+0.5);        // +0.5 to round upwards
    if( iwy >= h )
        iwy = h-1;

    return h-1-iwy;
}

void QWinInputSphere::mousePressEvent( QMouseEvent * e )
{
    int x, y;
    double wx, wy;
    double pcoord[3];

    if( e->modifiers() == Qt::ControlModifier && e->button() == Qt::LeftButton )
    {
        if( !selectingZoom )
        {
            selectingZoom = true;
            zoomAnchor1 = e->pos();
            zoomAnchor2 = zoomAnchor1;
            LoadAnchorMap();
            MarkSelection( zoomAnchor1.x(), zoomAnchor1.y(), zoomAnchor2.x(), zoomAnchor2.y() );
        }
        QWidget::mousePressEvent(e);
        return;
    }



    if( e->button() == Qt::LeftButton )
    {
        x = e->x();
        y = e->y();    
        wx = coWorldX(x);
        wy = coWorldY(y);    
        if( MATHFUNC(is_valid_viewcoord)( wx, wy, pcoord ) )
            parentWnd->OnMouseClickLeft( pcoord[0],pcoord[1],pcoord[2] );
        return;
    }

    if( e->button() == Qt::RightButton )
    {
        // cancel zoom window with right mouse button

        if( selectingZoom )
        {
            SaveAnchorMap();
            selectingZoom = false;
        }
        else
        {
            x = e->x();
            y = e->y();    
            wx = coWorldX(x);
            wy = coWorldY(y);    
            if( MATHFUNC(is_valid_viewcoord)( wx, wy, pcoord ) )
            {
                parentWnd->OnMouseClickRight( pcoord[0],pcoord[1],pcoord[2] );
            }
            return;
        }
        QWidget::mousePressEvent(e);
        return;
    }

    QWidget::mousePressEvent(e);
}

void QWinInputSphere::mouseReleaseEvent( QMouseEvent * e )
{
    if( e->button() == Qt::LeftButton )
    {
        // finish zoom window between zoomAnchor1 and zoomAnchor2

        if( selectingZoom )
        {
            SaveAnchorMap();
            selectingZoom = false;

            int _x1, _x2, _y1, _y2;
            
            _x1 = zoomAnchor1.x();
            _y1 = zoomAnchor1.y();
            _x2 = zoomAnchor2.x();
            _y2 = zoomAnchor2.y();

            parentWnd->MakeZoom( coWorldX(_x1), coWorldY(_y1), coWorldX(_x2), coWorldY(_y2) );
        }

        QWidget::mouseReleaseEvent(e);
        return;
    }

    QWidget::mouseReleaseEvent(e);
}

void QWinInputSphere::drawLine( double _x1, double _y1, double _x2, double _y2, int color )
{
    int wx1, wy1, wx2, wy2;

    if( staticPainter != NULL )
    {
        if( _x1 >= x0 && _x1 <= x1 && _y1 >= y0 && _y1 <= y1 )
        {
            wx1 = coWinX(_x1);
            wy1 = coWinY(_y1);

            if( _x2 >= x0 && _x2 <= x1 && _y2 >= y0 && _y2 <= y1 )
            {
                // both points are visible in the window

                wx2 = coWinX(_x2);
                wy2 = coWinY(_y2);

                if( paintedXMin > wx1 ) paintedXMin = wx1;
                if( paintedXMax < wx1 ) paintedXMax = wx1;
                if( paintedYMin > wy1 ) paintedYMin = wy1;
                if( paintedYMax < wy1 ) paintedYMax = wy1;
                if( paintedXMin > wx2 ) paintedXMin = wx2;
                if( paintedXMax < wx2 ) paintedXMax = wx2;
                if( paintedYMin > wy2 ) paintedYMin = wy2;
                if( paintedYMax < wy2 ) paintedYMax = wy2;

                staticPainter->setPen( QXFIGCOLOR(color) );
                staticPainter->drawLine( wx1, wy1, wx2, wy2 );
            }
            else
            {
                // only (_x2,_y2) is not visible

                if( LineRectangleIntersect( _x1, _y1, _x2, _y2, x0, x1, y0, y1 ) )
                {
                    wx1 = coWinX(_x1);
                    wy1 = coWinY(_y1);
                    wx2 = coWinX(_x2);
                    wy2 = coWinY(_y2);
                    staticPainter->setPen( QXFIGCOLOR(color) );

                    if( paintedXMin > wx1 ) paintedXMin = wx1;
                    if( paintedXMax < wx1 ) paintedXMax = wx1;
                    if( paintedYMin > wy1 ) paintedYMin = wy1;
                    if( paintedYMax < wy1 ) paintedYMax = wy1;
                    if( paintedXMin > wx2 ) paintedXMin = wx2;
                    if( paintedXMax < wx2 ) paintedXMax = wx2;
                    if( paintedYMin > wy2 ) paintedYMin = wy2;
                    if( paintedYMax < wy2 ) paintedYMax = wy2;

                    staticPainter->drawLine( wx1, wy1, wx2, wy2 );
                }
            }
        }
        else
        {
            if( _x2 >= x0 && _x2 <= x1 && _y2 >= y0 && _y2 <= y1 )
            {
                // only (_x2,_y2) is visible

                if( LineRectangleIntersect( _x1, _y1, _x2, _y2, x0, x1, y0, y1 ) )
                {
                    wx1 = coWinX(_x1);
                    wy1 = coWinY(_y1);
                    wx2 = coWinX(_x2);
                    wy2 = coWinY(_y2);
                    staticPainter->setPen( QXFIGCOLOR(color) );

                    if( paintedXMin > wx1 ) paintedXMin = wx1;
                    if( paintedXMax < wx1 ) paintedXMax = wx1;
                    if( paintedYMin > wy1 ) paintedYMin = wy1;
                    if( paintedYMax < wy1 ) paintedYMax = wy1;
                    if( paintedXMin > wx2 ) paintedXMin = wx2;
                    if( paintedXMax < wx2 ) paintedXMax = wx2;
                    if( paintedYMin > wy2 ) paintedYMin = wy2;
                    if( paintedYMax < wy2 ) paintedYMax = wy2;

                    staticPainter->drawLine( wx1, wy1, wx2, wy2 );
                }
            }
            else
            {
                // both end points are invisible

                if( LineRectangleIntersect( _x1, _y1, _x2, _y2, x0, x1, y0, y1 ) )
                {
                    wx1 = coWinX(_x1);
                    wy1 = coWinY(_y1);
                    wx2 = coWinX(_x2);
                    wy2 = coWinY(_y2);
                    staticPainter->setPen( QXFIGCOLOR(color) );

                    if( paintedXMin > wx1 ) paintedXMin = wx1;
                    if( paintedXMax < wx1 ) paintedXMax = wx1;
                    if( paintedYMin > wy1 ) paintedYMin = wy1;
                    if( paintedYMax < wy1 ) paintedYMax = wy1;
                    if( paintedXMin > wx2 ) paintedXMin = wx2;
                    if( paintedXMax < wx2 ) paintedXMax = wx2;
                    if( paintedYMin > wy2 ) paintedYMin = wy2;
                    if( paintedYMax < wy2 ) paintedYMax = wy2;
                    
                    staticPainter->drawLine( wx1, wy1, wx2, wy2 );
                }
            }
        }
    }
}

void QWinInputSphere::drawPoint( double x, double y, int color )
{
    int _x, _y;
    if( staticPainter != NULL )
    {
        if( x < x0 || x > x1 || y < y0 || y > y1 )
            return;
        staticPainter->setPen( QXFIGCOLOR(color) );
        _x=coWinX(x);
        _y=coWinY(y);

        if( paintedXMin > _x ) paintedXMin = _x;
        if( paintedXMax < _x ) paintedXMax = _x;
        if( paintedYMin > _y ) paintedYMin = _y;
        if( paintedYMax < _y ) paintedYMax = _y;

        staticPainter->drawPoint( _x, _y );
    }
}

void QWinInputSphere::prepareDrawing()
{
    if( PainterCache == NULL )
    {
        isPainterCacheDirty = true;
        PainterCache = new QPixmap( size() );
    }
    staticPainter = new QPainter(PainterCache);
    
    paintedXMin = width()-1;
    paintedYMin = height()-1;
    paintedXMax = 0;
    paintedYMax = 0;
}

void QWinInputSphere::finishDrawing()
{
    if( staticPainter != NULL )
    {
        staticPainter->end();
        delete staticPainter;
        staticPainter = NULL;

        if( paintedXMin < 0 ) paintedXMin = 0;
        if( paintedXMax >= width() ) paintedXMax = width()-1;
        if( paintedYMin < 0 ) paintedYMin = 0;
        if( paintedYMax >= height() ) paintedYMax = height()-1;

        if( paintedYMax >= paintedYMin && paintedXMax >= paintedXMin )
            update( paintedXMin, paintedYMin, paintedXMax-paintedXMin+1, paintedYMax-paintedYMin+1);
    }
}

void QWinInputSphere::refresh( void )
{
    isPainterCacheDirty = true;
    update();
}

void QWinInputSphere::plotPoincareSphere( void )
{
    int color;
    struct P4POLYLINES * p;

    p = CircleAtInfinity;
    color = CLINEATINFINITY;

    staticPainter->setPen( QXFIGCOLOR(color) );
    while( p != NULL )
    {
        staticPainter->drawLine( coWinX(p->x1), coWinY(p->y1), coWinX(p->x2), coWinY(p->y2) );
        p = p->next;
    }
}

void QWinInputSphere::plotPoincareLyapunovSphere( void )
{
    int color;
    struct P4POLYLINES * p;

    p = CircleAtInfinity;
    color = CLINEATINFINITY;

    staticPainter->setPen( QXFIGCOLOR(color) );
    while( p != NULL )
    {
        staticPainter->drawLine( coWinX(p->x1), coWinY(p->y1), coWinX(p->x2), coWinY(p->y2) );
        p = p->next;
    }

    p = PLCircle;
    color = CLINEATINFINITY;

    staticPainter->setPen( QXFIGCOLOR(color) );
    while( p != NULL )
    {
        staticPainter->drawLine( coWinX(p->x1), coWinY(p->y1), coWinX(p->x2), coWinY(p->y2) );
        p = p->next;
    }
    return;
}

void QWinInputSphere::plotLineAtInfinity( void )
{
    switch( VFResults.typeofview )
    {
    case TYPEOFVIEW_U1:
    case TYPEOFVIEW_V1:
        if( x0 < 0.0 && x1 > 0.0 )
        {
            staticPainter->setPen( QXFIGCOLOR(CLINEATINFINITY) );
            staticPainter->drawLine( coWinX(0.0), 0, coWinX(0.0), h-1 );
        }
        break;
    case TYPEOFVIEW_U2:
    case TYPEOFVIEW_V2:
        if( y0 < 0.0 && y1 > 0.0 )
        {
            staticPainter->setPen( QXFIGCOLOR(CLINEATINFINITY) );
            staticPainter->drawLine( 0, coWinY(0.0), w-1, coWinY(0.0) );
        }

        break;
    }
}

struct P4POLYLINES * QWinInputSphere::produceEllipse( double cx, double cy, double a, double b, bool dotted, double resa, double resb )
{
    // this is an exact copy of the plotEllipse routine, except that output is stored
    // in a list of points that is dynamically allocated.

    double theta, t1, t2, e, R, x, y, c, prevx, prevy;
    bool d;
    bool doton;
    int dotcount;
    struct P4POLYLINES * first;
    struct P4POLYLINES * last;

    prevx = prevy = 0;
    dotcount=0;
    first = NULL;
    last = NULL;
    
    R = (resa < resb) ? resa : resb;
    if( R < 1.0 ) R = 1.0; // protection
    e = 2*acos(1.0-0.5/R);
    if( R*sin(e) > 1.0 )
        e = asin(1.0/R);

    theta = 0;
    
    d = false;
    doton = true;

//  FILE * fp;
//  fp = fopen( "C:\\test.txt", "wt" );

    while( theta < TWOPI )
    {
//        fprintf( fp, "%f\n", (float)theta );
//        fflush(fp);
        c = (x0-cx)/a;
        if( c > -1.0 && c < 1.0 )
        {
            t1 = acos(c);
            t2 = TWOPI - t1;
            if( theta >= t1 && theta < t2 )
            {
//                fprintf( fp, "A EXCL [%f %f]\n", (float)t1, (float)t2 );
                theta = t2+e/4; d = false; continue;
            }
        }
        c = (x1-cx)/a;
        if( c > -1.0 && c < 1.0 )
        {
            t1 = acos(c);
            t2 = TWOPI - t1;
            if( theta < t1 )
            {
//                fprintf( fp, "B EXCL [-infinity %f]\n", (float)t1 );
                theta = t1+e/4; d = false; continue;
            }
            if( theta >= t2 )
            {
//                fprintf( fp, "C EXCL [%f, infinity]\n",  (float)t2 );
                theta = TWOPI+e/4; d = false; break;
            }
        }
        c = (y0-cy)/b;
        if( c > -1.0 && c < 1.0 )
        {
            t1 = asin(c);
            t2 = PI-t1;
            if( t1 < 0 )
            {
                t2 = t1+TWOPI;
                t1 = PI-t1;
                if( theta >= t1 && theta < t2 )
                {
//                    fprintf( fp, "D EXCL [%f %f]\n", (float)t1, (float)t2 );
                    theta = t2+e/4; d = false; continue;
                }
            }
            else
            {
                if( theta < t1 )
                {
//                    fprintf( fp, "E EXCL [-infinity %f]\n", (float)t1 );
                    theta = t1+e/4; d = false; continue;
                }
                if( theta >= t2 )
                {
//                    fprintf( fp, "F EXCL [%f, infinity]\n",  (float)t2 );
                    theta = TWOPI+e/4; d = false; break;
                }
            }
        }
        c = (y1-cy)/b;
        if( c > -1.0 && c < 1.0 )
        {
            t1 = asin(c);
            t2 = PI-t1;
            if( t1 < 0 )
            {
                t2 = t1+TWOPI;
                t1 = PI-t1;
                if( theta < t1 )
                {
//                    fprintf( fp, "G EXCL [-infinity %f]\n", (float)t1 );
                    theta = t1+e/4; d = false; continue;
                }
                if( theta >= t2 )
                {
//                    fprintf( fp, "H EXCL [%f, infinity]\n",  (float)t2 );
                    theta = TWOPI; d = false; break;
                }
            }
            else
            {
                if( theta >= t1 && theta < t2 )
                {
//                    fprintf( fp, "I EXCL [%f %f]\n", (float)t1, (float)t2 );
                    theta = t2+e/4; d = false; continue;
                }
            }
        }

        x = cx + a*cos(theta);
        y = cy + b*sin(theta);
        theta += e;
        
        // (x,y) is inside view

        if( !d )
        {
            if( doton )
            {
                d = true;
                prevx = x;
                prevy = y;
                dotcount = 0;
                doton = true;
            }
            else
            {
                if( ++dotcount > 7 && dotted )
                {
                    d = false;
                    doton = (doton) ? false:true;
                    dotcount = 0;
                }
            }
        }
        else
        {
            if( doton )
            {
                if( first == NULL )
                {
                    last = first = (struct P4POLYLINES *)malloc( sizeof(struct P4POLYLINES) );
                    last->next = NULL;
                }
                else
                {
                    last->next = (struct P4POLYLINES *)malloc( sizeof(struct P4POLYLINES) );
                    last = last->next;
                    last->next = NULL;
                }

                last->x1 = prevx;
                last->y1 = prevy;
                last->x2 = x;
                last->y2 = y;

                prevx = x;
                prevy = y;
            }
            if( ++dotcount > 7 && dotted )
            {
                d = false;
                doton = (doton) ? false:true;
                dotcount = 0;
            }
        }
    }
//  fclose(fp);
    return first;
}

void QWinInputSphere::plotCurves( void )
{
    int r;
    if( VFResults.curves != NULL )
        for( r = 0; r < ThisVF->numcurves; r++ )
            plotCurve( VFResults.curves+r, r );
}

void QWinInputSphere::plotCurve( CURVERESULT * crv, int index )
{
    double pcoord[3];
    struct orbits_points * sep;

    sep = crv->sep_points;
    
    while( sep != NULL )
    {
        if(sep->dashes)
            PlotLine(pcoord,sep->pcoord, sep->color);
        else
            PlotPoint(sep->pcoord, sep->color);
            
        copy_x_into_y(sep->pcoord,pcoord);

        sep=sep->nextpt;
    }
}

void QWinInputSphere::PlotLine( double * p1, double * p2, int color )
{
    double ucoord1[2];
    double ucoord2[2];
    double ucoord3[2];
    double ucoord4[2];

    if( MATHFUNC(sphere_to_viewcoordpair)(p1,p2,ucoord1,ucoord2,ucoord3,ucoord4) )
    {
        drawLine( ucoord1[0], ucoord1[1], ucoord2[0], ucoord2[1], color );
    }
    else
    {
        drawLine( ucoord1[0], ucoord1[1], ucoord2[0], ucoord2[1], color );
        drawLine( ucoord3[0], ucoord3[1], ucoord4[0], ucoord4[1], color );
    }
}

void QWinInputSphere::PlotPoint( double * p, int color )
{
    double ucoord[2];

    MATHFUNC(sphere_to_viewcoord)(p[0],p[1],p[2],ucoord);

    drawPoint( ucoord[0], ucoord[1], color );
}

void QWinInputSphere::SetupCoordinateTransformations( void )
{
    VFResults.SetupCoordinateTransformations();
}
