#include <qwidget.h>
#ifndef TABLE_H
#include "table.h"
#endif

class QStatusBar;
class QPoint;
class QTimer;
class QCurvesDlg;
class QLabel;

class QWinInputSphere : public QWidget
{
    Q_OBJECT

public:
    QWinInputSphere( QCurvesDlg *, QLabel * );
    ~QWinInputSphere();
    void paintEvent( QPaintEvent * );

    int coWinX( double x );     // coordinate changes: from world to windows coordinates
    int coWinY( double y );
    double coWorldX( int x );   // from windows to world coordinates
    double coWorldY( int y );
    int coWinV( double );
    int coWinH( double );

    double x0, y0;              // world-coordinates of upper-left corner
    double x1, y1;              // world-coordinates of upper-right corner
    double dx;                  // x1-x0
    double dy;                  // y1-y0
    bool iszoom;

    double horPixelsPerMM;
    double verPixelsPerMM;

    QPixmap * PainterCache;
    QLabel * status;
    bool isPainterCacheDirty;
    int paintedXMin;            // to know the update rectangle after painting
    int paintedXMax;            // we keep to smallest rectangle enclosing
    int paintedYMin;            // all painted objects.
    int paintedYMax;

private:
    QPainter * staticPainter;
    QCurvesDlg * parentWnd;
    
    struct P4POLYLINES * CircleAtInfinity;
    struct P4POLYLINES * PLCircle;
    QTimer * refreshTimeout;

public:
    int spherebgcolor;

    void adjustToNewSize( void );

    void plotPoincareSphere( void );
    void plotPoincareLyapunovSphere( void );
    void plotLineAtInfinity( void );

    struct P4POLYLINES * produceEllipse( double cx, double cy, double a, double b, bool dotted, double resa, double resb );

    void prepareDrawing( void );
    void drawPoint( double x, double y, int color );
    void drawLine( double x1, double y1, double x2, double y2, int color );
    void finishDrawing( void );
    
    void SaveAnchorMap( void );
    void LoadAnchorMap( void );

    void SetupPlot( void );
    void refresh( void );
    void CalculateHeightFromWidth( int * width, int * height, int maxheight, double aspectratio );
    void updatePointSelection( void );
    void MarkSelection( int x1, int y1, int x2, int y2 );

    void plotCurves( void );
    void plotCurve( CURVERESULT *, int );
    void PlotLine( double *, double *, int );
    void PlotPoint( double *, int );
    void SetupCoordinateTransformations( void );

public slots:
    void resizeEvent( QResizeEvent * e );
    void mouseMoveEvent( QMouseEvent * e );
    void mousePressEvent( QMouseEvent * e );
    void mouseReleaseEvent( QMouseEvent * e );
    void keyPressEvent ( QKeyEvent * e );
    void refreshAfterResize( void );

private:
    bool selectingZoom;
    QPoint zoomAnchor1;
    QPoint zoomAnchor2;
    QPixmap * AnchorMap;

public:
    int w;                          // width of window
    int h;                          // height of window
    int idealh;                     // ideal height of window to get good aspect ratio
};
