#include <qapplication.h>
#include <qbuttongroup.h>
#include <qlabel.h>
#include <qradiobutton.h>
#include <qspinbox.h>
#include <qlineedit.h>
#include <qlayout.h>
#include "custom.h"
#include "table.h"
#include "file_vf.h"
#include "win_p5.h"
#include "p5application.h"
#include "win_params.h"
#include "win_find.h"

extern QP5Application * p5app;
extern bool action_OnlyPrepareFile;

extern void SetP4WindowTitle( QWidget *, QString );

QParamsDlg::~QParamsDlg()
{
    GetDataFromDlg();
}

QParamsDlg::QParamsDlg( QFindDlg * finddlg )
#ifdef DOCK_PARAMSWINDOW
    : QWidget( finddlg )
#else
    : QWidget()
#endif
{
    parent = finddlg;
//  setFont( QFont( FONTSTYLE, FONTSIZE ) );

    QLabel * p4title = new QLabel( "Find Singular Points Parameters", this );
    p4title->setFont( *(p5app->TitleFont) );

    QLabel * symlabel = new QLabel( "Calculations: ", this );
    symlabel->setFont( *(p5app->BoldFont) );
    btn_alg = new QRadioButton( "Algebraic", this );
    btn_num = new QRadioButton( "Numeric", this );

    QLabel * seplabel = new QLabel( "Test Separatrices:  ", this );
    seplabel->setFont( *(p5app->BoldFont) );
    btn_sepyes = new QRadioButton( "Yes", this );
    btn_sepno = new QRadioButton( "No", this );

    QLabel * lbl_precis = new QLabel( "Precision:", this );
    lbl_precis->setFont( *(p5app->BoldFont) );
    spin_precis = new QSpinBox(this);
    spin_precis->setMinimum(MINPRECISION);
    spin_precis->setMaximum(MAXPRECISION);

    QLabel * lbl_epsilon = new QLabel( "Epsilon:", this );
    lbl_epsilon->setFont( *(p5app->BoldFont) );
    edt_epsilon = new QLineEdit( "", this );

    QLabel * lbl_level = new QLabel( "Level of Approximation:", this );
    lbl_level->setFont( *(p5app->BoldFont) );
    spin_level = new QSpinBox(this);
    spin_level->setMinimum(MINLEVEL);
    spin_level->setMaximum(MAXLEVEL);

    QLabel * lbl_numlevel = new QLabel( "Numeric level:", this );
    lbl_numlevel->setFont( *(p5app->BoldFont) );
    spin_numlevel = new QSpinBox(this);
    spin_numlevel->setMinimum(MINNUMLEVEL);
    spin_numlevel->setMaximum(MAXNUMLEVEL);

    QLabel * lbl_maxlevel = new QLabel( "Maximum level:", this );
    lbl_maxlevel->setFont( *(p5app->BoldFont) );
    spin_maxlevel = new QSpinBox(this);
    spin_maxlevel->setMinimum(MINMAXLEVEL);
    spin_maxlevel->setMaximum(MAXMAXLEVEL);

    QLabel * lbl_weakness = new QLabel( "Maximum level of weakness:", this );
    lbl_weakness->setFont( *(p5app->BoldFont) );
    spin_weakness = new QSpinBox(this);
    spin_weakness->setMinimum(MINWEAKNESS);
    spin_weakness->setMaximum(MAXWEAKNESS);

    lbl_p = new QLabel( "p:", this );
    lbl_p->setFont( *(p5app->BoldFont) );
    spin_p = new QSpinBox(this);
    spin_p->setMinimum(MINP);
    spin_p->setMaximum(MAXP);

    lbl_q = new QLabel( "q:", this );
    lbl_q->setFont( *(p5app->BoldFont) );
    spin_q = new QSpinBox(this);
    spin_q->setMinimum(MINP);
    spin_q->setMaximum(MAXP);

    lbl_x0 = new QLabel( "x0:", this );
    lbl_x0->setFont( *(p5app->BoldFont) );
    edt_x0 = new QLineEdit( "", this );

    lbl_y0 = new QLabel( "y0:", this );
    lbl_y0->setFont( *(p5app->BoldFont) );
    edt_y0 = new QLineEdit( "", this );

#ifdef TOOLTIPS
    btn_alg->setToolTip( "Let the symbolic manipulator work in algebraic mode.\n"
                            "The symbolic package will automatically switch over to\n"
                            "numeric calculations whenever a symbolic study is impossible." );
    btn_num->setToolTip( "Let the symbolic manipulator work in numeric mode" );
    btn_sepyes->setToolTip( "Enable numeric testing of the Taylor developments for the separatrices" );
    btn_sepno->setToolTip( "Disable separatrice testing" );
    spin_precis->setToolTip( "Number of digits of numeric precision" );
    edt_epsilon->setToolTip( "Default epsilon for the separatrices.\n"
                    "This may be changed later when plotting individual separatrices" );
    spin_level->setToolTip( "Order of Taylor approximation\nThis order is increased as long as the separatrix test fails." );
    spin_numlevel->setToolTip( "Start working numerically when order of Taylor approximation reaches this level" );
    spin_maxlevel->setToolTip( "Maximum order of Taylor approximation.  Stop testing separatrices at this order." );
    spin_weakness->setToolTip( "Maximum order of Lyapunov constants to be calculated for weak foci" );
    spin_p->setToolTip( "Poincare-Lyapunov weights" );
    spin_q->setToolTip( "Poincare-Lyapunov weights" );
    edt_x0->setToolTip( "Coordinates of the singular point in case of study of ONE singularity" );
    edt_y0->setToolTip( "Coordinates of the singular point in case of study of ONE singularity" );
#endif

    // layout

    mainLayout = new QBoxLayout( QBoxLayout::TopToBottom, this );

    mainLayout->addWidget( p4title );

    QHBoxLayout * symLayout = new QHBoxLayout();
    symLayout->addWidget( symlabel );
    symLayout->addWidget( btn_alg );
    symLayout->addWidget( btn_num );
    mainLayout->addLayout( symLayout );

    QHBoxLayout * sepLayout = new QHBoxLayout();
    sepLayout->addWidget( seplabel );
    sepLayout->addWidget( btn_sepyes );
    sepLayout->addWidget( btn_sepno );
    mainLayout->addLayout( sepLayout );

    QHBoxLayout * layout1 = new QHBoxLayout();
    layout1->addWidget( lbl_precis );
    layout1->addWidget( spin_precis );
    layout1->addStretch(0);

    QHBoxLayout * layout2 = new QHBoxLayout();
    layout2->addWidget( lbl_epsilon );
    layout2->addWidget( edt_epsilon );
    layout2->addStretch(0);

    QHBoxLayout * layout3 = new QHBoxLayout();
    layout3->addWidget( lbl_level );
    layout3->addWidget( spin_level );
    layout3->addStretch(0);

    QHBoxLayout * layout4 = new QHBoxLayout();
    layout4->addWidget( lbl_numlevel );
    layout4->addWidget( spin_numlevel );
    layout4->addStretch(0);

    QHBoxLayout * layout5 = new QHBoxLayout();
    layout5->addWidget( lbl_maxlevel );
    layout5->addWidget( spin_maxlevel );
    layout5->addStretch(0);

    QHBoxLayout * layout6 = new QHBoxLayout();
    layout6->addWidget( lbl_weakness );
    layout6->addWidget( spin_weakness );
    layout6->addStretch(0);

    QHBoxLayout * layout7 = new QHBoxLayout();
    layout7->addWidget( lbl_p );
    layout7->addWidget( spin_p );
    layout7->addWidget( lbl_q );
    layout7->addWidget( spin_q );
    layout7->addStretch(0);

    QHBoxLayout * layout8 = new QHBoxLayout();
    layout8->addWidget( lbl_x0 );
    layout8->addWidget( edt_x0 );
    layout8->addStretch(0);

    QHBoxLayout * layout9 = new QHBoxLayout();
    layout9->addWidget( lbl_y0 );
    layout9->addWidget( edt_y0 );
    layout9->addStretch(0);

    mainLayout->addLayout( layout1 );
    mainLayout->addLayout( layout2 );
    mainLayout->addLayout( layout3 );
    mainLayout->addLayout( layout4 );
    mainLayout->addLayout( layout5 );
    mainLayout->addLayout( layout6 );
    mainLayout->addLayout( layout7 );
    mainLayout->addLayout( layout8 );
    mainLayout->addLayout( layout9 );
//  mainLayout->addStretch(0);

    setLayout(mainLayout);

    btngrp_numeric = new QButtonGroup( this );
    btngrp_numeric->addButton( btn_alg );
    btngrp_numeric->addButton( btn_num );

    btngrp_testsep = new QButtonGroup( this );
    btngrp_testsep->addButton( btn_sepyes );
    btngrp_testsep->addButton( btn_sepno );

    // connections

    QObject::connect( btn_alg,  SIGNAL(toggled(bool)), this, SLOT(btn_alg_toggled(bool)) );
    QObject::connect( btn_num,  SIGNAL(toggled(bool)), this, SLOT(btn_num_toggled(bool)) );
    QObject::connect( spin_level, SIGNAL(valueChanged(int)), this, SLOT(OnLevelChange(int)) );
    QObject::connect( btn_sepno, SIGNAL(toggled(bool)), this, SLOT( btn_sepno_toggled(bool)) );
    QObject::connect( btn_sepyes, SIGNAL(toggled(bool)), this, SLOT( btn_sepyes_toggled(bool)) );

    // finishing

    UpdateDlgData();
#ifndef DOCK_PARAMSWINDOW
    SetP4WindowTitle( this, "Parameters" );
#endif
}

void QParamsDlg::OnLevelChange( int value )
{
}

void QParamsDlg::btn_alg_toggled( bool on )
{
    if( on ) 
    {
        if( !ThisVF->HasCommonBool( ThisVF->numeric ) || ThisVF->CommonBoolean( ThisVF->numeric ) != false )
        {
            ThisVF->SetCommonBool( ThisVF->numeric, false );
            if( ThisVF->changed == false )
            {
                ThisVF->changed = true;
                p5app->Signal_Changed();
            }
        }
    }
}

void QParamsDlg::btn_num_toggled( bool on )
{
    if( on ) 
    {
        if( !ThisVF->HasCommonBool( ThisVF->numeric ) || ThisVF->CommonBoolean( ThisVF->numeric ) != true )
        {
            ThisVF->SetCommonBool( ThisVF->numeric, true );
            if( ThisVF->changed == false )
            {
                ThisVF->changed = true;
                p5app->Signal_Changed();
            }
        }
    }
}

void QParamsDlg::btn_sepyes_toggled( bool on )
{
    if( on ) 
    {
        if( !ThisVF->HasCommonBool( ThisVF->testsep ) || ThisVF->CommonBoolean( ThisVF->testsep ) != true )
        {
            ThisVF->SetCommonBool( ThisVF->testsep, true );
            if( ThisVF->changed == false )
            {
                ThisVF->changed = true;
                p5app->Signal_Changed();
            }
        }
    }
}

void QParamsDlg::btn_sepno_toggled( bool on )
{
    if( on ) 
    {
        if( !ThisVF->HasCommonBool( ThisVF->testsep ) || ThisVF->CommonBoolean( ThisVF->testsep ) != false )
        {
            ThisVF->SetCommonBool( ThisVF->testsep, false );
            if( ThisVF->changed == false )
            {
                ThisVF->changed = true;
                p5app->Signal_Changed();
            }
        }
    }
}

void QParamsDlg::ExclusiveToggle( bool on, QRadioButton * first, ... )
{
     first->setChecked(on);
}

void QParamsDlg::SetSpinBoxCommonValue( QSpinBox * sb, int * val, int minval )
{
    if( ThisVF->HasCommonInt( val ) )
    {
        sb->setSpecialValueText("");
        sb->setMinimum( minval );
        sb->setValue( ThisVF->CommonInt(val) );
    }
    else
    {
        sb->setMinimum( minval-1 );
        sb->setSpecialValueText("##");
        sb->setValue( minval-1 );
    }
}

bool QParamsDlg::GetSpinBoxCommonValue( QSpinBox * sb, int * newval )
{
    int val;
    val = sb->value();
    if( val == sb->minimum() )
    {
        if( sb->specialValueText().length() > 0 )
        {
            // special value: ## --> ignore change
            
            return false;
        }
    }
    
    if( ThisVF->HasCommonInt( newval ) )
        if( val == ThisVF->CommonInt( newval ) )
            return false;
    
    ThisVF->SetCommonInt( newval, val );
    return true;
}

void QParamsDlg::SetLineEditCommonValue( QLineEdit * le, QString * * val )
{
    if( ThisVF->HasCommonString(val) )
        le->setText( ThisVF->CommonString(val) );
    else
        le->setText( "##" );
}

bool QParamsDlg::GetLineEditCommonValue( QLineEdit * le, QString * * newval )
{
    QString val;
    int i;
    
    val = le->text();
    val = val.trimmed();
    
    if( !ThisVF->HasCommonString(newval) )
    {
        for( i = val.length()-1; i >= 0; i-- )
            if( val[i] != '#' )
                break;
    
        if( i < 0 )
            return false;
    
        if( !val.compare( ThisVF->CommonString(newval) ) )
            return false;
    }
    
    if( !val.compare( ThisVF->CommonString(newval) ) )
        return false;
    ThisVF->SetCommonString( newval, val );
    return true;
}

void QParamsDlg::GetDataFromDlg( void )
{
    QString epsilon;
    QString x0;
    QString y0;
    bool changed;
    
    changed = false;

    if( spin_p->value() != ThisVF->p ||
        spin_q->value() != ThisVF->q  )
    {
        ThisVF->p = spin_p->value();
        ThisVF->q = spin_q->value();
        changed=true;
    }
    
    x0 = edt_x0->text();
    x0 = x0.trimmed();

    y0 = edt_y0->text();
    y0 = y0.trimmed();

    if( x0.compare( ThisVF->x0 ) != 0 || y0.compare( ThisVF->y0 ) != 0 )
    {
        ThisVF->x0 = x0;
        ThisVF->y0 = y0;
        changed=true;
    }    

    changed |= GetSpinBoxCommonValue( spin_level, ThisVF->taylorlevel );
    changed |= GetSpinBoxCommonValue( spin_numlevel, ThisVF->numericlevel );
    changed |= GetSpinBoxCommonValue( spin_maxlevel, ThisVF->maxlevel );
    changed |= GetSpinBoxCommonValue( spin_weakness, ThisVF->weakness );
    changed |= GetSpinBoxCommonValue( spin_precis, ThisVF->precision );
    changed |= GetLineEditCommonValue( edt_epsilon, ThisVF->epsilon );

    if( changed )
    {
        if( ThisVF->changed == false )
        {
            ThisVF->changed = true;
            p5app->Signal_Changed();
        }
    }
}

void QParamsDlg::UpdateDlgData( void )
{
    if( ThisVF->HasCommonBool( ThisVF->numeric ) )
    {
        if( ThisVF->CommonBoolean( ThisVF->numeric ) )
            ExclusiveToggle( true, btn_num, btn_alg, NULL );
        else
            ExclusiveToggle( true, btn_alg, btn_num, NULL );
    }
    else
    {
        btngrp_numeric->setExclusive(false);
        btn_num->setChecked(false);
        btn_alg->setChecked(false);
        btngrp_numeric->setExclusive(true);
    }
    
    if( ThisVF->HasCommonBool( ThisVF->testsep ) )
    {
        if( ThisVF->CommonBoolean( ThisVF->testsep ) )
            ExclusiveToggle( true, btn_sepyes, btn_sepno, NULL );
        else
            ExclusiveToggle( true, btn_sepno, btn_sepyes, NULL );
    }
    else
    {
        btngrp_testsep->setExclusive(false);
        btn_sepyes->setChecked(false);
        btn_sepno->setChecked(false);
        btngrp_testsep->setExclusive(true);
    }

    SetSpinBoxCommonValue( spin_level, ThisVF->taylorlevel, MINLEVEL );
    SetSpinBoxCommonValue( spin_numlevel, ThisVF->numericlevel, MINNUMLEVEL );
    SetSpinBoxCommonValue( spin_maxlevel, ThisVF->maxlevel, MINMAXLEVEL );
    SetSpinBoxCommonValue( spin_weakness, ThisVF->weakness, MINWEAKNESS );
    SetLineEditCommonValue( edt_epsilon, ThisVF->epsilon );
    SetSpinBoxCommonValue( spin_precis, ThisVF->precision, MINPRECISION );
    
    if( ThisVF->typeofstudy == TYPEOFSTUDY_ONE )
    {
        edt_x0->setText( ThisVF->x0 );
        edt_y0->setText( ThisVF->y0 );
        spin_p->setValue( ThisVF->p );
        spin_q->setValue( ThisVF->q );

        spin_p->setEnabled(false);
        spin_q->setEnabled(false);
        edt_x0->setEnabled(true);
        edt_y0->setEnabled(true);
        
        lbl_x0->setFont( *(p5app->BoldFont) );
        lbl_y0->setFont( *(p5app->BoldFont) );
        lbl_p->setFont( *(p5app->StandardFont) );
        lbl_q->setFont( *(p5app->StandardFont) );
    }
    else
    {
        spin_p->setValue( ThisVF->p );
        spin_q->setValue( ThisVF->q );
        edt_x0->setText( ThisVF->x0 );
        edt_y0->setText( ThisVF->x0 );

        spin_p->setEnabled(true);
        spin_q->setEnabled(true);
        edt_x0->setEnabled(false);
        edt_y0->setEnabled(false);
        
        lbl_p->setFont( *(p5app->BoldFont) );
        lbl_q->setFont( *(p5app->BoldFont) );
        lbl_x0->setFont( *(p5app->StandardFont) );
        lbl_y0->setFont( *(p5app->StandardFont) );
    }
}
