#include <qtooltip.h>
#include <qapplication.h>
#include <qlabel.h>
#include <qlayout.h>
#include <qlineedit.h>
#include <qspinbox.h>
#include <qscrollbar.h>

#include "custom.h"
#include "table.h"
#include "file_vf.h"
#include "win_p5.h"
#include "p5application.h"
#include "win_vf.h"
#include "win_find.h"
#include "win_vfselect.h"
#include "win_curves.h"

extern QP5Application * p5app;

extern void SetP4WindowTitle( QWidget *, QString );

QVectorFieldDlg::QVectorFieldDlg( QFindDlg * finddlg )
#ifdef DOCK_VFWINDOW
    : QWidget( finddlg )
#else
    : QWidget( )
#endif
{
    parent = finddlg;
//  setFont( QFont( FONTSTYLE, FONTSIZE ) );

#ifdef DOCK_VFWINDOW
    QLabel * p4title = new QLabel( "Specify the vector field:", this );
    p4title->setFont( *(p5app->TitleFont) );
#endif

    edt_xprime = new QLineEdit( "", this );
    QLabel * xlabel = new QLabel( "&x' = ", this );
    xlabel->setFont( *(p5app->BoldFont) );
    xlabel->setBuddy( edt_xprime );

    edt_yprime = new QLineEdit( "", this );
    QLabel * ylabel = new QLabel( "&y' = ", this );
    ylabel->setFont( *(p5app->BoldFont) );
    ylabel->setBuddy( edt_yprime );

    edt_gcf = new QLineEdit( "", this );
    QLabel * glabel = new QLabel( "&Gcf: ", this );
    glabel->setFont( *(p5app->BoldFont) );
    glabel->setBuddy( edt_gcf );

    QLabel * plabel = new QLabel( "Number of Parameters: ", this );
    plabel->setFont( *(p5app->BoldFont) );
    
    spin_numparams = new QSpinBox( this );
    spin_numparams->setMinimum(0);
    spin_numparams->setMaximum(MAXNUMPARAMS);

#ifdef TOOLTIPS
    edt_xprime->setToolTip( "Enter your differential equations here.\n"
                        "Use syntax conform to the symbolic manipulator." );
    edt_yprime->setToolTip( "Enter your differential equations here.\n"
                        "Use syntax conform to the symbolic manipulator." );
    edt_gcf->setToolTip( "Enter greatest common factor here.\n"
                        "Enter \"1\" if you are sure there is no GCF\n"
                        "Enter \"0\" if you want P5 to search for a GCF" );
    spin_numparams->setToolTip( "If your vector field contains parameters,\n"
                        "specify the number of parameters here." );
#endif
    // layout

    mainLayout = new QBoxLayout( QBoxLayout::TopToBottom, this );

#ifdef DOCK_VFWINDOW
    mainLayout->addWidget( p4title );
#endif

    QHBoxLayout * layout1 = new QHBoxLayout();
    layout1->addWidget( xlabel );
    layout1->addWidget( edt_xprime );
    
    QHBoxLayout * layout2 = new QHBoxLayout();
    layout2->addWidget( ylabel );
    layout2->addWidget( edt_yprime );

    QHBoxLayout * layout22 = new QHBoxLayout();
    layout22->addWidget( glabel );
    layout22->addWidget( edt_gcf );

    QHBoxLayout * layout3 = new QHBoxLayout();
    layout3->addWidget( plabel );
    layout3->addWidget( spin_numparams );
    layout3->addStretch(0);

    paramLayout = new QHBoxLayout();

    mainLayout->addLayout( layout1 );
    mainLayout->addLayout( layout2 );
    mainLayout->addLayout( layout22 );
    mainLayout->addLayout( layout3 );
    mainLayout->addLayout( paramLayout );

    mainLayout->setSizeConstraint(QLayout::SetFixedSize);
    setLayout(mainLayout);

    sb_params = NULL;
    params = NULL;

    if( ThisVF->numparams != 0 )
    {
        if( ThisVF->numparams > MAXNUMPARAMSSHOWN )
        {
            sb_params = new QScrollBar( Qt::Vertical, this );
            sb_params->setRange( 0, ThisVF->numparams - MAXNUMPARAMSSHOWN );
            sb_params->setSingleStep(1);
            sb_params->setPageStep( MAXNUMPARAMSSHOWN );
        }
        else
            sb_params = NULL;

        params = new QVFParams(this, sb_params);
        paramLayout->addWidget( params );
        if( ThisVF->numparams > MAXNUMPARAMSSHOWN )
        {   
            paramLayout->addWidget( sb_params );
            QObject::connect( sb_params, SIGNAL(valueChanged(int)), params, SLOT(paramsSliderChanged(int)) );
            QObject::connect( sb_params, SIGNAL(sliderMoved(int)), params, SLOT(paramsSliderChanged(int)) );
        }
        params->show();
        if( ThisVF->numparams > MAXNUMPARAMSSHOWN )
            sb_params->show();
        
        spin_numparams->setValue(ThisVF->numparams);
    }

    // connections

    QObject::connect( spin_numparams, SIGNAL(valueChanged(int)), this, SLOT(numParamsChanged(int)) );
    QObject::connect( edt_xprime, SIGNAL(editingFinished()), this, SLOT(xdot_editingFinished()) );
    QObject::connect( edt_yprime, SIGNAL(editingFinished()), this, SLOT(ydot_editingFinished()) );
    QObject::connect( edt_gcf, SIGNAL(editingFinished()), this, SLOT(gcf_editingFinished()) );

#ifndef DOCK_VFWINDOW
    SetP4WindowTitle( this, "Vector Field" );
#endif

    UpdateDlgData();
}

void QVectorFieldDlg::numParamsChanged( int val )
{
    if( val < 0 || val > MAXNUMPARAMS || val == ThisVF->numparams )
        return;

    delete params;
    params = NULL;

    if( sb_params != NULL )
    {
        delete sb_params;
        sb_params = NULL;
    }

    if( ThisVF->changed == false )
    {
        ThisVF->changed = true;
        p5app->Signal_Changed();
    }
    ThisVF->numparams = val;

    if( val == 0 )
        return;

    if( val > MAXNUMPARAMSSHOWN )
    {
        sb_params = new QScrollBar( Qt::Vertical, this );
        sb_params->setRange( 0, ThisVF->numparams - MAXNUMPARAMSSHOWN );
        sb_params->setSingleStep(1);
        sb_params->setPageStep( MAXNUMPARAMSSHOWN );
    }
    else
        sb_params = NULL;

    params = new QVFParams(this, sb_params );
    paramLayout->addWidget( params );
    if( val > MAXNUMPARAMSSHOWN )
    {   
        paramLayout->addWidget( sb_params );
        QObject::connect( sb_params, SIGNAL(valueChanged(int)), params, SLOT(paramsSliderChanged(int)) );
        QObject::connect( sb_params, SIGNAL(sliderMoved(int)), params, SLOT(paramsSliderChanged(int)) );
    }
    params->show();
    if( val > MAXNUMPARAMSSHOWN )
        sb_params->show();
}


QVectorFieldDlg::~QVectorFieldDlg()
{
    if( params != NULL )
    {
        delete params;
        params = NULL;
    }
    GetDataFromDlg();
}

void QVectorFieldDlg::GetDataFromDlg( void )
{
    bool changed;
    
    changed = false;

    changed |= GetLineEditCommonValue( edt_xprime, ThisVF->xdot );
    changed |= GetLineEditCommonValue( edt_yprime, ThisVF->ydot );
    changed |= GetLineEditCommonValue( edt_gcf, ThisVF->gcf );

    if( changed )
    {
        if( ThisVF->changed == false )
        {
            ThisVF->changed = true;
            p5app->Signal_Changed();
        }
    }
    if( params != NULL )
        params->GetDataFromDlg();
}

void QVectorFieldDlg::UpdateDlgData( void )
{
    SetLineEditCommonValue( edt_xprime, ThisVF->xdot );
    SetLineEditCommonValue( edt_yprime, ThisVF->ydot );
    SetLineEditCommonValue( edt_gcf, ThisVF->gcf );
    spin_numparams->setValue( ThisVF->numparams );

    if( params != NULL )
    {
        if( !params->UpdateDlgData() )
        {
            delete params;
            params = NULL;
        }
    }
    if( ThisVF->numparams == 0 )
    {
        if( sb_params != NULL )
        {
            delete sb_params;
            sb_params = NULL;
        }
    }
    
    if( ThisVF->numparams != 0 && params == NULL )
    {
        if( ThisVF->numparams > MAXNUMPARAMSSHOWN )
        {
            sb_params = new QScrollBar( Qt::Vertical, this );
            sb_params->setRange( 0, ThisVF->numparams - MAXNUMPARAMSSHOWN );
            sb_params->setSingleStep(1);
            sb_params->setPageStep( MAXNUMPARAMSSHOWN );
        }
        else
            sb_params = NULL;

        params = new QVFParams( this, sb_params );
        paramLayout->addWidget( params );
        if( ThisVF->numparams > MAXNUMPARAMSSHOWN )
        {   
            paramLayout->addWidget( sb_params );
            QObject::connect( sb_params, SIGNAL(valueChanged(int)), params, SLOT(paramsSliderChanged(int)) );
            QObject::connect( sb_params, SIGNAL(sliderMoved(int)), params, SLOT(paramsSliderChanged(int)) );
        }
        params->show();
        if( ThisVF->numparams > MAXNUMPARAMSSHOWN )
            sb_params->show();
    }
}


void QVectorFieldDlg::SetLineEditCommonValue( QLineEdit * le, QString * * val )
{
    if( ThisVF->HasCommonString(val) )
        le->setText( ThisVF->CommonString(val) );
    else
        le->setText( "########" );
}

void QVectorFieldDlg::xdot_editingFinished( void )
{
    GetDataFromDlg();
    if( parent->VfSelect_Window != NULL )
        if( parent->VfSelect_Window->win_curves != NULL )
            parent->VfSelect_Window->win_curves->UpdateDlgData();
}

void QVectorFieldDlg::ydot_editingFinished( void )
{
    GetDataFromDlg();
    if( parent->VfSelect_Window != NULL )
        if( parent->VfSelect_Window->win_curves != NULL )
            parent->VfSelect_Window->win_curves->UpdateDlgData();
}

void QVectorFieldDlg::gcf_editingFinished( void )
{
    GetDataFromDlg();
    if( parent->VfSelect_Window != NULL )
        if( parent->VfSelect_Window->win_curves != NULL )
            parent->VfSelect_Window->win_curves->UpdateDlgData();
}

bool QVectorFieldDlg::GetLineEditCommonValue( QLineEdit * le, QString * * newval )
{
    QString val;
    int i;
    
    val = le->text();
    val = val.trimmed();
    
    if( !ThisVF->HasCommonString(newval) )
    {
        for( i = val.length()-1; i >= 0; i-- )
            if( val[i] != '#' )
                break;
    
        if( i < 0 )
            return false;
    
        if( !val.compare( ThisVF->CommonString(newval) ) )
            return false;
    }
    if( !val.compare( ThisVF->CommonString(newval) ) )
        return false;
    
    ThisVF->SetCommonString( newval, val );
    return true;
}






QVFParams::QVFParams( QVectorFieldDlg * parent, QScrollBar * sb )
    : QWidget( parent )
{
    int i;

    QLabel * label0 = new QLabel( "Enter values for all parameters" );
    label0->setFont( *(p5app->TitleFont) );

    sb_params = sb;
    currentnumparams = ThisVF->numparams;
    currentshownparams = (currentnumparams < MAXNUMPARAMSSHOWN ) ? currentnumparams : MAXNUMPARAMSSHOWN;
    currentpageindex = 0;

    for( i = 0; i < currentshownparams; i++ )
    {
        edt_parlabel[i] = new QLineEdit( ThisVF->parlabel[i], this );
        lbl_parequal[i] = new QLabel( " = ", this );
        edt_parvalue[i] = new QLineEdit( "", this );
#ifdef TOOLTIPS
        edt_parlabel[i]->setToolTip( "Enter the name of the parameter" );
        edt_parvalue[i]->setToolTip( "Enter the value of this parameter" );
#endif
    }

    mainLayout = new QBoxLayout( QBoxLayout::TopToBottom );
    mainLayout->addWidget( label0 );
    for( i = 0; i < currentshownparams; i++ )
    {
        lay_parvalue[i] = new QHBoxLayout();
        lay_parvalue[i]->addWidget( edt_parlabel[i] );
        lay_parvalue[i]->addWidget( lbl_parequal[i] );
        lay_parvalue[i]->addWidget( edt_parvalue[i] );
        mainLayout->addLayout( lay_parvalue[i] );

        QObject::connect( edt_parlabel[i], SIGNAL(editingFinished()), this, SLOT(params_editingFinished()) );
        QObject::connect( edt_parvalue[i], SIGNAL(editingFinished()), this, SLOT(params_editingFinished()) );

    }
    
    setLayout(mainLayout);
    UpdateDlgData();
}

QVFParams::~QVFParams()
{
}

bool QVFParams::focusNextPrevChild( bool next ) // subclassed from QWidget
{
    QWidget * p;
    p = focusWidget();
    if( p == edt_parlabel[0] && currentpageindex != 0 && !next )
    {
        sb_params->setValue( currentpageindex-1 );
        return focusNextPrevChild( true );
    }
    if( p == edt_parvalue[currentshownparams-1] && currentpageindex < currentnumparams - MAXNUMPARAMSSHOWN && next )
    {
        sb_params->setValue( currentpageindex+1 );
        return focusNextPrevChild( false );
    }
    return QWidget::focusNextPrevChild( next );
}

void QVFParams::params_editingFinished( void )
{
    int i;
    bool changed;
    QString temp;
    QString lbl;

    if( ThisVF == NULL )
        return;

    if( ThisVF->numparams != currentnumparams )
        return;

    changed = false;
    for( i = 0; i < currentshownparams; i++ )
    {
        lbl = edt_parlabel[i]->text();
        lbl = lbl.trimmed();        
        if( lbl.compare( ThisVF->parlabel[i+currentpageindex] ) )
        {
            ThisVF->parlabel[i+currentpageindex] = lbl;
            changed = true;
        }
        changed |= GetLineEditCommonParvalue( edt_parvalue[i], i+currentpageindex );
    }

    if( changed )
    {
        if( ThisVF->changed == false )
        {
            ThisVF->changed = true;
            p5app->Signal_Changed();
        }
    }
    return;
}


bool QVFParams::GetDataFromDlg( void )
{
    return true;
}

bool QVFParams::UpdateDlgData( void )
{
    int i;

    if( ThisVF->numparams != currentnumparams )
        return false;

    for( i = 0; i < currentshownparams; i++ )
    {
        edt_parlabel[i]->setText( ThisVF->parlabel[i+currentpageindex] );
        SetLineEditCommonParvalue( edt_parvalue[i], i+currentpageindex );
    }

    return true;
}

void QVFParams::paramsSliderChanged( int value )
{
    int i;
    int newindex;

    newindex = value;
    if( newindex < 0 )
        newindex = 0;
    if( newindex > currentnumparams - MAXNUMPARAMSSHOWN )
        newindex = currentnumparams - MAXNUMPARAMSSHOWN;

    if( newindex != currentpageindex )
    {
        currentpageindex = newindex;

        for( i = 0; i < currentshownparams; i++ )
        {
            edt_parlabel[i]->setText( ThisVF->parlabel[i+currentpageindex] );
            SetLineEditCommonParvalue( edt_parvalue[i], i+currentpageindex );
        }
    }
}

void QVFParams::SetLineEditCommonParvalue( QLineEdit * le, int index )
{
    if( ThisVF->HasCommonParvalue(index) )
        le->setText( ThisVF->CommonParvalue(index) );
    else
        le->setText( "########" );
}

bool QVFParams::GetLineEditCommonParvalue( QLineEdit * le, int index )
{
    QString val;
    int i;
    
    val = le->text();
    val = val.trimmed();
    
    if( !ThisVF->HasCommonParvalue(index) )
    {
        for( i = val.length()-1; i >= 0; i-- )
            if( val[i] != '#' )
                break;
    
        if( i < 0 )
            return false;
    
        if( !val.compare( ThisVF->CommonParvalue(index) ) )
            return false;
    }
    if( !val.compare( ThisVF->CommonParvalue(index) ) )
        return false;
    
    ThisVF->SetCommonParvalue( index, val );
    return true;
}
