#include <qtooltip.h>
#include <qmessagebox.h>
#include <qapplication.h>
#include <qlabel.h>
#include <qlayout.h>
#include <qcombobox.h>
#include <qpushbutton.h>
#include <qlineedit.h>
#include "custom.h"
#include "table.h"
#include "file_vf.h"
#include "win_p5.h"
#include "p5application.h"
#include "win_plot.h"
#include "win_vfselect.h"
#include "win_find.h"
#include "win_curves.h"
#include "win_main.h"
#include "math_p4.h"

extern QP5Application * p5app;

extern void SetP4WindowTitle( QWidget *, QString );

QVFSelectDlg::QVFSelectDlg( QFindDlg * finddlg )
#ifdef DOCK_VFSELECTWINDOW
    : QWidget( finddlg )
#else
    : QWidget( )
#endif
{
    parent = finddlg;
    win_curves = NULL;
//  setFont( QFont( FONTSTYLE, FONTSIZE ) );

#ifdef DOCK_VFSELECTWINDOW
    QLabel * p4title = new QLabel( "Vector Fields:", this );
    p4title->setFont( *(p5app->TitleFont) );
#endif

    cbb_vfselect = new QComboBox( this );
    btn_add = new QPushButton( "Add", this );
    btn_del = new QPushButton( "Delete", this );
    btn_prev = new QPushButton( "<", this );
    btn_next = new QPushButton( ">", this );
    btn_p5config = new QPushButton( "Piecewise &Configuration...", this );

#ifdef TOOLTIPS
    btn_add->setToolTip( "Adds a new vector field" );
    btn_del->setToolTip( "Deletes the current vector field" );
    btn_prev->setToolTip( "Selects previous vector field" );
    btn_next->setToolTip( "Selects next vector field" );
    btn_p5config->setToolTip( "Opens the piecewise-configuration window" );
#endif
    // layout

    mainLayout = new QBoxLayout( QBoxLayout::TopToBottom, this );

    QHBoxLayout * L1 = new QHBoxLayout();
    L1->addWidget( p4title );
    L1->addWidget( cbb_vfselect );
    mainLayout->addLayout( L1 );

    QHBoxLayout * L2 = new QHBoxLayout();
    L2->addWidget( btn_prev );
    L2->addWidget( btn_next );
    mainLayout->addLayout( L2 );

    QHBoxLayout * L3 = new QHBoxLayout();
    L3->addWidget( btn_add );
    L3->addWidget( btn_del );
    mainLayout->addLayout( L3 );

    mainLayout->addWidget( btn_p5config );

    // connections

    QObject::connect( btn_prev,  SIGNAL(clicked()), this, SLOT(OnBtnPrev()) );
    QObject::connect( btn_next,  SIGNAL(clicked()), this, SLOT(OnBtnNext()) );
    QObject::connect( btn_add,  SIGNAL(clicked()), this, SLOT(OnBtnAdd()) );
    QObject::connect( btn_del,  SIGNAL(clicked()), this, SLOT(OnBtnDel()) );
    QObject::connect( btn_p5config,  SIGNAL(clicked()), this, SLOT(OnBtnP5Config()) );
    QObject::connect( cbb_vfselect, SIGNAL(currentIndexChanged(int)), this, SLOT(OnVfSelectionChanged(int)) );

#ifndef DOCK_VFSELECTWINDOW
    SetP4WindowTitle( this, "P5 Vector Field Selection" );
#endif

    UpdateDlgData();
}

QVFSelectDlg::~QVFSelectDlg()
{
    if( win_curves != NULL )
    {
        delete win_curves;
        win_curves=NULL;
    }
    GetDataFromDlg();
}

void QVFSelectDlg::GetDataFromDlg( void )
{
    if( win_curves != NULL )
        win_curves->GetDataFromDlg();
}

void QVFSelectDlg::UpdateDlgData( void )
{
    int k;
    QString s;
    
    for( k = 0; k < ThisVF->numvf; k++ )
    {
        s.sprintf( "%d", k+1 );
        if( cbb_vfselect->count() > k )
            cbb_vfselect->setItemText( k, s );
        else
            cbb_vfselect->addItem( s );
    }
    if( ThisVF->numvf > 1 )
    {
        s = "Select All";
        if( cbb_vfselect->count() > k )
            cbb_vfselect->setItemText( k, s );
        else
            cbb_vfselect->addItem( s );
        while( cbb_vfselect->count() > ThisVF->numvf+1 )
            cbb_vfselect->removeItem( ThisVF->numvf+1 );
    }    
    else
    {
        while( cbb_vfselect->count() > ThisVF->numvf+1 )
            cbb_vfselect->removeItem( ThisVF->numvf+1 );
    }
    if( ThisVF->numselected == 1 )
        cbb_vfselect->setCurrentIndex( ThisVF->selected[0] );
    else
        cbb_vfselect->setCurrentIndex( ThisVF->numselected );
    
    if( ThisVF->numvf > 1 )
    {
        if( ThisVF->selected[0]==0 && ThisVF->numselected == 1 )
            btn_prev->setEnabled(false);
        else
            btn_prev->setEnabled(true);

        if( ThisVF->selected[0]==ThisVF->numvf-1 && ThisVF->numselected == 1 )
            btn_next->setEnabled(false);
        else
            btn_next->setEnabled(true);
    }
    else
    {
        btn_prev->setEnabled(false);
        btn_next->setEnabled(false);
    }
    
    if( win_curves != NULL )
        win_curves->UpdateDlgData();
}

void QVFSelectDlg::OnBtnAdd( void )
{
    parent->GetDataFromDlg();
    if( !CheckPlotWindowClosed() )
        return;
        
    ThisVF->AddVectorField();
    if( ThisVF->numselected > 1 )
    {
        ThisVF->selected = (int *)realloc( ThisVF->selected, sizeof(int) );
        ThisVF->numselected = 1;
    }
    ThisVF->selected[0] = ThisVF->numvf - 1;
    parent->UpdateDlgData();
    if( ThisVF->changed == false )
    {
        ThisVF->changed = true;
        p5app->Signal_Changed();
    }
}

void QVFSelectDlg::OnBtnDel( void )
{
    parent->GetDataFromDlg();
    if( !CheckPlotWindowClosed() )
        return;

    while( ThisVF->numselected > 1 )
        ThisVF->DeleteVectorField( ThisVF->selected[ThisVF->numselected - 1] );
    ThisVF->DeleteVectorField( ThisVF->selected[0] );    
    parent->UpdateDlgData();
    if( ThisVF->changed == false )
    {
        ThisVF->changed = true;
        p5app->Signal_Changed();
    }
}

void QVFSelectDlg::OnBtnPrev( void )
{
    int j;
    parent->GetDataFromDlg();
    j = ThisVF->selected[0];
    if( ThisVF->numselected > 1 )
    {
        ThisVF->selected = (int *)realloc( ThisVF->selected, sizeof(int) );
        ThisVF->numselected = 1;
    }    
    if( --j < 0 ) j = 0;
    ThisVF->selected[0] = j;
    parent->UpdateDlgData();
}

void QVFSelectDlg::OnBtnNext( void )
{
    int j;
    parent->GetDataFromDlg();
    j = ThisVF->selected[0];
    if( ThisVF->numselected > 1 )
    {
        ThisVF->selected = (int *)realloc( ThisVF->selected, sizeof(int) );
        ThisVF->numselected = 1;
    }
    if( ++j >= ThisVF->numvf ) j = ThisVF->numvf-1;
    ThisVF->selected[0] = j;
    parent->UpdateDlgData();
}

void QVFSelectDlg::OnVfSelectionChanged( int index )
{
    int k;
        
    if( index < ThisVF->numvf )
    {
        if( index != ThisVF->selected[0] || ThisVF->numselected > 1 )
        {
            parent->GetDataFromDlg();
            ThisVF->selected = (int *)realloc( ThisVF->selected, sizeof(int) );
            ThisVF->selected[0] = index;
            ThisVF->numselected = 1;
            parent->UpdateDlgData();
        }
    }
    else if( index >= ThisVF->numvf )
    {
        // selected all

        if( ThisVF->numselected != ThisVF->numvf )
        {
            parent->GetDataFromDlg();
            ThisVF->selected = (int *)realloc( ThisVF->selected, sizeof(int) * ThisVF->numvf );
            for( k = 0; k < ThisVF->numvf; k++ )
                ThisVF->selected[k] = k;
            ThisVF->numselected = ThisVF->numvf;
            parent->UpdateDlgData();
        }
    }
}

bool QVFSelectDlg::CheckPlotWindowClosed( void )
{
    if( parent->parent->Plot_Window == NULL )
        return true;
    if( parent->parent->Plot_Window->isVisible() == false )
    {
        parent->parent->ClosePlotWindow();
        return true;
    }
    QMessageBox::critical( this, "P5",
            "Please close the plot window before changing the piecewise configuration!\n" );
    
    return false;
}

void QVFSelectDlg::OnBtnP5Config( void )
{
    if( !CheckPlotWindowClosed() )
        return;

    if( win_curves == NULL )
    {
        VFResults.ReadTables( ThisVF->getbarefilename(), true, true );
        
        win_curves = new QCurvesDlg(parent);
        win_curves->show();
    }
    else
    {
        win_curves->show();
        win_curves->raise();
    }
}

void QVFSelectDlg::CloseConfigWindow( void )
{
    if( win_curves != NULL )
    {
        delete win_curves;
        win_curves = NULL;
    }
}
